#!/bin/bash

. "$PWD/.GuestInstaller.source"

function cleanup()
{
	echo "Gathering debug data before removing pkg.zip file and rpms directory..."

	set -x
	cksum pkg.zip
	ls rpms
	cksum rpms/*.rpm

	# remove the file and directory
	rm -rf pkg.zip rpms
	set +x
}

VM_ID=$_vmid
CONSOLE_URL=$_consoleurl

last_status=""
TASK_STATUS_INPROGRESS=3
TASK_STATUS_SUCCESS=4
TASK_STATUS_FAILED=5

installation_result_file="Installation_Result.xml"
PRODUCT_NAME_LONG="Symantec HA Wizards Agent Pack"

# find installer PID
# Timeout = 30 secs
timeout=30
current_count=0
while /bin/true; do
	pid=`cat /var/tmp/cpi.pid`
	if [ -n "$pid" ]; then
		pid=`echo $pid | sed 's/^.*://'`
		break
	fi
	sleep 1
	current_count=`expr $current_count + 1`
	if [ $current_count -gt $timeout ]; then
		# Indicate failure case to console, because CPI could not start within the timeout period
		description="Installation of $PRODUCT_NAME_LONG failed."
		echo "Updating $installation_result_file for VM=$VM_ID, TaskStatus=$TASK_STATUS_FAILED..."

		echo "<Status>" > $installation_result_file
		echo "      <TaskStatus>$TASK_STATUS_FAILED</TaskStatus>" >> $installation_result_file
		#echo "      <TaskProgress>100</TaskProgress>" >> $installation_result_file
		echo "      <TaskDesc>$description</TaskDesc>" >> $installation_result_file
		echo "</Status>" >> $installation_result_file

		echo "Calling command [curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_FAILED -d TaskDesc=\"$description\" ..."
		curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_FAILED -d TaskDesc="$description"
		echo
		echo "-------------------------------------------"
		cleanup
		exit 1
	fi
done

while ps -p $pid; do

	# Progress statement looks like below in CPI log files
	#   "0 21:05:46 progress(Installing Symantec ApplicationHA: 93 15-16)"
	new_status=`grep "^[0-9][0-9]* [0-9][0-9]*:[0-9][0-9]*:[0-9][0-9]* progress" vii_logs/*/*.log* | sed -e 's/^[^:]*://' -e  's/[0-9][0-9]* [0-9][0-9]*:[0-9][0-9]*:[0-9][0-9]* progress(\(.*\)ing .*: \([0-9][0-9]*\) [0-9][0-9]*-[0-9][0-9]*)/\1 \2/' | tail -1`

	if [ "$last_status" != "$new_status" ]; then
		echo $new_status
		operation=`echo $new_status | sed 's/ .*//'`
		percentage=`echo $new_status | sed 's/.* //'`
		if [ "$operation" = "Install" ]; then

			# Description statement looks like below in CPI log files
			#   "1 21:05:26 Installing VRTSvcsvmw on vmguest1"
			description=`grep "[0-9: ]* Installing [^ ]* " vii_logs/*/*.log* | sed -e 's/.* \(Installing [^ ]* \)/\1/' | tail -1`

			# Postinstall tasks take a while to complete and description
			# is not updated in UI. We need to allow some time for postinstall tasks
			# and update description to reflect it.
			if [ $percentage = '100' ]; then
				percentage=99
				description='Performing postinstall tasks'
			fi

			echo "Updating $installation_result_file for VM=$VM_ID, TaskStatus=$TASK_STATUS_INPROGRESS, TaskProgress=$percentage ..."

			echo "<Status>" > $installation_result_file
			echo "      <TaskStatus>$TASK_STATUS_INPROGRESS</TaskStatus>" >> $installation_result_file
			echo "      <TaskProgress>$percentage</TaskProgress>" >> $installation_result_file
			echo "      <TaskDesc>$description</TaskDesc>" >> $installation_result_file
			echo "</Status>" >> $installation_result_file

			echo "Calling command [curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_INPROGRESS -d TaskProgress=$percentage -d TaskDesc=\"$description\"] ..."
			curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_INPROGRESS -d TaskProgress=$percentage -d TaskDesc="$description"
			description=""
			echo
			echo "-------------------------------------------"
		fi
		last_status=$new_status
	fi

	# do check for updates every 2 seconds
	# TBD: Make it configurable
	sleep 2 
done

installation_successful=0
percentage=100
install_log="installagpack.log"

#if grep "Symantec ApplicationHA Guest Components Install completed successfully" applicationha/installapplicationha.log; then
if grep "Install completed successfully" $install_log; then
	installation_successful=1
fi

if [ $installation_successful -eq 1 ]; then

	description="Installation of $PRODUCT_NAME_LONG is complete."
	echo "Updating $installation_result_file for VM=$VM_ID, TaskStatus=$TASK_STATUS_SUCCESS..."

	echo "<Status>" > $installation_result_file
	echo "      <TaskStatus>$TASK_STATUS_SUCCESS</TaskStatus>" >> $installation_result_file
	echo "      <TaskProgress>$percentage</TaskProgress>" >> $installation_result_file
	echo "      <TaskDesc>$description</TaskDesc>" >> $installation_result_file
	echo "</Status>" >> $installation_result_file

	echo "Calling command [curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_SUCCESS -d TaskDesc=\"$description\" ..."
	curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_SUCCESS -d TaskProgress=$percentage -d TaskDesc="$description"
	echo
	echo "-------------------------------------------"

else
	# Indicate failure case to console
	description="Installation of $PRODUCT_NAME_LONG failed."
	echo "Updating $installation_result_file for VM=$VM_ID, TaskStatus=$TASK_STATUS_FAILED..."

	echo "<Status>" > $installation_result_file
	echo "      <TaskStatus>$TASK_STATUS_FAILED</TaskStatus>" >> $installation_result_file
	echo "      <TaskProgress>$percentage</TaskProgress>" >> $installation_result_file
	echo "      <TaskDesc>$description</TaskDesc>" >> $installation_result_file
	echo "</Status>" >> $installation_result_file

	echo "Calling command [curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_FAILED -d TaskDesc=\"$description\" ..."
	curl -k $CONSOLE_URL -d VmID=$VM_ID -d TaskStatus=$TASK_STATUS_FAILED -d TaskProgress=$percentage -d TaskDesc="$description"
	echo
	echo "-------------------------------------------"

fi

cleanup

exit 0
