
#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Id: Oracle.pm,v 1.35 2014/07/18 07:24:13 bdas Exp $
#

package Oracle;

use UnixApplication;
use UnixOracleCodes;
use UnixNetwork;

@ISA = ("UnixApplication");  # required for Inheritance



use base qw(Application);

use Scalar::Util 'blessed';
use strict;
use warnings;
use Data::Dumper;
use Constants;
use Utils;
use PlatUtils;
use Storage;

use strict;
use warnings;

#-------------
# Gloabals
#-------------
my $OraclePackage = "VRTSvcsea";
my $OracleResType = "Oracle";
my $LsnrResType = "Netlsnr";
my $vcsencrypt = "/opt/VRTSvcs/bin/vcsencrypt";

#--------------------
# subroutines section
#--------------------
sub getInstances
{
	my $_self = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $applicable = "Yes";
	my $iterativeDiscovery = "No";
	my @instances = ();

	my %_hInstances = _discover_sid();
	@instances = keys(%_hInstances);
	return ($applicable, $iterativeDiscovery, \@instances);
}

sub vxGetInstancesXML
{
    my $self = shift;
    my $applicable = shift;
    my $iterativeDiscovery = shift;
    my $instancesRef = shift;
    my $ret = CONST_FAILURE;

    my $sMyName = ( caller 0 )[3];
    #---------
    # Announce
    #---------
    my $logStr = "Subroutine $sMyName called";
    Utils::vxlog("INFO", $logStr);

    my $str = "<DiscoveryData>";

    if ( defined $instancesRef ) {
        foreach my $ins (@$instancesRef) {
            $str .= "<Instance><Name>$ins</Name></Instance>";
            $ret = CONST_SUCCESS;
        }
    }

    $str .= "</DiscoveryData>";

END:
    if ($ret == CONST_FAILURE) {
        $str = "No instances installed";
    }

    return ($ret, $str);
}

sub discover
{
	my $_self = shift;
	my $_ahrParams = shift;
	my $sMyName = ( caller 0 )[3];
	my $msg = "";
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my ($retcode, $retxml) = (CONST_SUCCESS, "");
	my ($_xml, %_hInstances, $_tmpfile, $_simple_xml, $_xmltree, @_instances);
	my ($_discovery_complete, $_instance_name, $_ref, $_instance);
	my ($_home, $_owner, $_pfile, $_ownerUID);
	my $SysDiskIDdata = ();
	#----------------
	# Parse Arguments
	#----------------
	my @allStorageData = ();
	my ($applicable, $iterativeDiscovery, $instancesRef) = ();

	######### New function to discover existing config
	# In case of error, $resRef should be set to undef or "".
	my $resRef = $_self->discoverExistingAppConfig();
	($applicable, $iterativeDiscovery, $instancesRef) = $_self->getInstances();
	($retcode, $_xml) = $_self->vxGetInstancesXML($applicable, $iterativeDiscovery, $instancesRef);

	if ($retcode != CONST_SUCCESS) {
		$logStr = "No running Oracle instances found.";
		Utils::vxlog("ERROR", $logStr);
		return (APPCONTROL_ORACLE_DISCOVER_EMPTYXMLPROVIDED, "No running Oracle instances found.", "");
	}
	$_xmltree = Utils::convertXMLTextToHash($_xml);
	if (not defined $_xmltree or $_xmltree eq "") {
		$logStr = "No running Oracle instances found.";
		Utils::vxlog("ERROR", $logStr);
		return (APPCONTROL_ORACLE_DISCOVER_EMPTYXMLPROVIDED, "No running Oracle instances found.", "");
	}

	$logStr = Dumper($_xmltree);
	Utils::vxlog("INFO", $logStr);
	#--------------
	# parse Oratab
	#--------------
	Utils::vxlog("INFO", "Parsing Oratab");
	%_hInstances = _discover_sid();
	#------------------------------
	# discovering Oracle in details
	#------------------------------
	@_instances = ();
	if (not defined $_xmltree->{'Instance'}) {
		return (APPCONTROL_ORACLE_DISCOVER_NOINSTANCEPROVIDED,	"'Instance' tag is not provided in XML", "");
	}

	$_ref = ref ($_xmltree->{'Instance'});

	Utils::vxlog("INFO", "Instance in tree is [$_ref]");
	if ( $_ref eq "ARRAY" ) {
		@_instances = @{$_xmltree->{'Instance'}};
	} else {
		push (@_instances, $_xmltree->{'Instance'});
	}

	#--------------------------------------------------------------------------------------------------
	# Net Listener Information!
	# currently we do not co-relate listener and oracle sid, show all running listerners on the system
	#--------------------------------------------------------------------------------------------------
	my $_netListener = $_self->_get_listener($resRef);
	# Remove already configured Netlsnr Instances.	
	my @instData = ();
	foreach $_instance ( @_instances ) {
		my ($arApplicationPath, $arApplicationMounts) = ();
		my (@instanceStorageData) = ();
		if (not defined $_instance) {
			return (APPCONTROL_ORACLE_DISCOVER_NOINSTANCEPROVIDED, "'Instance' tag is not provided in XML", "");
		}
		if ( $_instance eq "" ) {
			return (APPCONTROL_ORACLE_DISCOVER_EMPTYINSTANCEPROVIDED, "Empty Instance tag is provided", "");
		}

		my $data = ();

		#----------------
		# Oracle Resource
		#----------------
		if (not defined $_instance->{'Name'}) {
			return (APPCONTROL_ORACLE_DISCOVER_INSTANCENAMENOTPROVIDED, "'Name' tag is not provided in XML", "");
		}

		$_instance_name = $_instance->{'Name'};
		Utils::vxlog("INFO", "Discovering Instance : [$_instance_name]");
		$data -> { Instance } = $_instance_name;

		# We should not show the already configured resources.
		if(CONST_SUCCESS == $_self->isOracleInstancePresent($_instance_name, $resRef)) {
			$logStr = "Skipping the already configured Oracle instance: $_instance_name";
			Utils::vxlog("INFO", $logStr);
			next;
		}
		if (not defined $_hInstances{$_instance_name}) {
			return (APPCONTROL_ORACLE_DISCOVER_INSTANCENOTAVAILABLE, "Instance named [$_instance_name] not found in /etc/oratab", "");
		}

		# Home attribute
		$_home = $_hInstances{$_instance_name}{'home'};
		$data -> { Home } = $_home;

		# Owner attribute
		my $cmd = "ls -ld $_home/bin/oracle | awk '{print \$3}' 2>&1";
		my $CmdRet = CONST_SUCCESS;
		($CmdRet, $_owner) = Utils::runCmd($cmd);
		if ( $CmdRet != CONST_SUCCESS ) {
			return ( APPCONTROL_ORACLE_DISCOVER_USERDISCOVERYFAILURE , "Failed to discover Owner!", "");
		}
		$cmd = "/usr/bin/getent passwd $_owner";
		$CmdRet = CONST_SUCCESS;
		($CmdRet, $_ownerUID) = Utils::runCmd($cmd);
		if ( $CmdRet != CONST_SUCCESS ) {
			return ( APPCONTROL_ORACLE_DISCOVER_USERDISCOVERYFAILURE , "Failed to discover Owner!", "");
		}
		$data -> { Owner } = $_owner;
		$data -> { Pfile } = "";
		$data -> { EnvFile } = "";

		# Owner UID is required for validating the shallow disovery on other nodes in the cluster.
		$_ownerUID =~ m/^.*?:.*?:(\d+)/;
		$data -> { OwnerUID } = $1;

		my @mntlist = ();
		my $oradata_path = _get_oradata_path($_home, $_instance->{'Name'}, $_owner);

		push (@$arApplicationPath, $_home);
		if (@$oradata_path) {
			push (@$arApplicationPath, @$oradata_path);
		} else {
			$logStr = "Skipping the Oracle instance: $_instance_name, because wizard could not get the Oracle data path for the instance. Check if the database is mounted.";
			Utils::vxlog("WARNING", $logStr);
			next;
		}
		my ($hrStorageDetails) = ();
		my ($sDiskGroup, $sVolumeGroup) = ();
		my $arDiskIDs;
		my $iCnt =1;
		foreach my $sApplicationPath ( @$arApplicationPath) {
			my $Storage_ret = CONST_SUCCESS;
			($Storage_ret, $hrStorageDetails) = $_self->GetStorageDetails ($sApplicationPath);
			if ( CONST_SUCCESS == $Storage_ret ) {
				if (grep { $_ eq $hrStorageDetails->{MountPoint} } @$arApplicationMounts) {
					next;
				}
				push(@$arApplicationMounts,$hrStorageDetails->{MountPoint});
				$hrStorageDetails -> { Instance } = "$_instance_name" . "_" . "$iCnt";
				if ( defined $hrStorageDetails -> { DiskGroup } ) {
					$sDiskGroup = $hrStorageDetails -> { DiskGroup };
					# DiskID is of the form [Controller:Port] 
					($Storage_ret, $msg, $arDiskIDs) = getVxVMDgDisks($sDiskGroup);
					if ($Storage_ret != CONST_SUCCESS) {
						$retcode = $Storage_ret;
						goto END;
					}
				}
				if ( defined $hrStorageDetails -> { VolumeGroup } ) {
					$sVolumeGroup = $hrStorageDetails -> { VolumeGroup };
					($Storage_ret, $msg, $arDiskIDs) = getLVMVgDisks($sVolumeGroup);
					if ($Storage_ret != CONST_SUCCESS) {
						$retcode = $Storage_ret;
						goto END;
					}
				}
				$hrStorageDetails -> { DiskID } = $arDiskIDs;
				foreach my $DiskID ( @$arDiskIDs ) {
					$SysDiskIDdata -> { $DiskID } = 1;
				}
				push ( @instanceStorageData, $hrStorageDetails);
			} else {
				Utils::vxlog("INFO", "Could not discover storage for [$sApplicationPath]");
			}
			$iCnt++;
		}
		$data -> { MountPoint } = $arApplicationMounts;
		my $lsnrRef = _get_Listeners_for_sid($_instance_name, $_netListener);
		$data -> { Listener } = $lsnrRef;
		# We should not allow the new database containing same storage dependency
		if (CONST_SUCCESS == $_self->isStorageConfigured($resRef, \@instanceStorageData)) {
			$logStr = "Skipping the Oracle instance: $_instance_name, because the storage is already configured for an application running under VCS control.";
			Utils::vxlog("INFO", $logStr);
			next;
		}
		@allStorageData = (@allStorageData, @instanceStorageData);
		push(@instData, $data);
	}

	my $lsnrxml = get_listener_xml($_netListener);
	my $_ex_listeners = $_self->get_existing_listeners($resRef);
	if (scalar (@instData) >= 1) {
		$retxml = "<DiscoveryData>";
		$retxml .= $_self -> createApplicationXML( \@instData, "Instance");
		$retxml .= $lsnrxml ;
		if (scalar (@$_ex_listeners) >= 1) {
			$retxml .= "<ExListeners>";
			$retxml .= $_self->createApplicationXML( \@$_ex_listeners, "Listener" );
			$retxml .= "</ExListeners>";
		}
		$retxml .= "<StorageList>";
		$retxml .= $_self->createApplicationXML( \@allStorageData, "Storage" );
		$retxml .= "</StorageList>";
		$retxml .= "</DiscoveryData>";
	} else {
		return (APPCONTROL_ORACLE_DISCOVER_NEWDISCOVERYFAILURE, "No new Oracle instances found. All instances are either already configured or use the storage which is already configured for an application running under VCS control.", "");
	}
	chomp $retxml;
END:
	return ($retcode, $msg, $retxml);
}
# Shallow discovery is called by validate from the node where wizard is running.
# shallow discovery happens on all the other nodes where wizard is not running.
sub shallowDiscover
{
	my $self = shift;
	my $params = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);
	my $CmdRet = CONST_SUCCESS;
	my $_ownerUID = "";
	my $ret = CONST_SUCCESS;
	my $msg = "";
	my $xml = "<xml></xml>";
	if (!(defined $params) or ($params eq "")) {
		$msg = "No input to perform validation of Oracle instances.";
		goto END;
	}
	my $tree = Utils::convertXMLTextToHash($params);
	my @instances = ();
	if (defined $tree -> {Instance}) {
		# configuration wizard
		@instances = Utils::convertHashValuesToArray($tree -> { Instance });
	} else {
		# Add node wizard
		if ((defined $tree -> {ApplicationInfo}) and
			(ref ($tree -> {ApplicationInfo}) =~ /HASH/) and
			(defined $tree -> {ApplicationInfo}{ValidationInfo}) and
			(ref ($tree -> {ApplicationInfo}{ValidationInfo}) =~ /HASH/)) {
			@instances = Utils::convertHashValuesToArray($tree ->{ ApplicationInfo } { ValidationInfo }{ Instance });
		}
	}
	# Validatation on inputs
	foreach my $ins ( @instances ) {
		if (!defined $ins->{ Selected } or $ins->{ Selected } !~ /true/i ) {
			next;
		}
		# match Oracle Home information.
		my $inOraHome = $ins->{ Home }{ Value };

		# Check if the Home is shared or Local.
		my @MountPointList = ();
		my $sharedHome = "false";
		if (( defined ($ins->{MountPointList})) and ($ins -> {MountPointList} ne "") and
		    ( defined ($ins->{MountPointList}{MountPoint})) ) {
			@MountPointList = Utils::convertHashValuesToArray ($ins ->{MountPointList}{MountPoint});
			foreach my $sMountPoint (@MountPointList) {
				if (defined $sMountPoint->{Selected} and  $sMountPoint->{Selected} =~ /true/) {
					if ($inOraHome =~ /^$sMountPoint->{Value}/) {
						$sharedHome = "true";
					}
				}
			}
		}
		if (($sharedHome !~ /true/) and (! -e "$inOraHome/bin/oracle")) {
			return ( APPCONTROL_ORACLE_SHALLOWDISCOVERY_HOMENOTFOUND, "Oracle home directory, $inOraHome, is neither present on the local storage nor on the shared storage.", "");
		}
		# match owner information.
		my $inOwnerUID = $ins->{ OwnerUID }{ Value };
		my $_owner = $ins->{ Owner }{ Value };
		my $cmd = "/usr/bin/getent passwd $_owner";
		($CmdRet, $_ownerUID) = Utils::runCmd($cmd);
		# User not present
		if ( $CmdRet != CONST_SUCCESS ) {
			return ( APPCONTROL_ORACLE_SHALLOWDISCOVERY_OWNERNOTFOUND, "Failed to get the information about $_owner, Oracle owner.", "");
		}
		$_ownerUID =~ m/^.*?:.*?:(\d+)/;
		$_ownerUID = $1;

		# Flag error if UIDs do not match.
		if ( $_ownerUID != $inOwnerUID ) {
			return ( APPCONTROL_ORACLE_SHALLOWDISCOVERY_UIDMISMATCH, "Failed to match the user ID of $_owner, Oracle Database owner because the discovered user ID ($_ownerUID) and expected user ID ($inOwnerUID) did not match.", "");
		}
	}
END:
	return ($ret, $msg, $xml);
}

sub _get_listener
{
	my $_self = shift;
	my $sMyName = ( caller 0 )[3];
	my $resRef = shift;
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $cmd = "ps -ef | grep -i tnslsnr | grep -v grep";
	my ($CmdRet, $op) = Utils::runCmd($cmd);

	$op = "" if (!defined $op);
	my $ref = ();
	foreach ( split (/\n/, $op)) {
		Utils::vxlog("INFO", "processing [$_]");
		s/^\s+//;

		my @fields = split(/\s+/);

		my $lsnr  = $fields[-2];
		my $user  = $fields[0];
		my $pid   = $fields[1];
		my $home = read_process_env( $pid, "ORACLE_HOME" );

		$cmd = "export ORACLE_HOME=$home; su $user -c '$home/bin/lsnrctl status $lsnr'";
		($CmdRet, $op) = Utils::runCmd($cmd);
		my $lsnrctl = $op;
		my @llines = split(/\n/, $op);
		my $ip = "";
		my $TnsAdmin = "";
		foreach my $lline ( @llines ) {
			if ( $lline =~ /^Listener Parameter File/ ) {
				my @cols = split(/\s+/, $lline);
				my $listenerFile = $cols[-1];
				@cols = split(/\//, $listenerFile);
				$cols[-1] = "";
				$TnsAdmin = join("/", @cols);
				#last;
			}
			if ( $lline =~ /\(DESCRIPTION=/ ) {
				$lline =~ s/.*HOST=//g;
				$lline =~ s/\).*//g;
				$ip = $lline;
			}
		}
		if(CONST_SUCCESS == $_self->isListenerPresent($lsnr, $resRef)) {
			$logStr = "Skipping the already configured Oracle Listener: $lsnr";
			Utils::vxlog("INFO", $logStr);

			next;
		}
		$ref -> { $lsnr } { Pid	 	} = $pid;
		$ref -> { $lsnr } { Home 	} = $home;
		$ref -> { $lsnr } { Owner 	} = $user;
		$ref -> { $lsnr } { TnsAdmin 	} = $TnsAdmin;
		$ref -> { $lsnr } { EnvFile  	} = "";
		$ref -> { $lsnr } { LsnrPwd  	} = "";
		$ref -> { $lsnr } { Lsnrctl  	} = $lsnrctl;
		$ref -> { $lsnr } { IP 		} = $ip;
	}
	Utils::vxlog("INFO", Dumper($ref));
	return $ref;
}


sub _get_Listeners_for_sid
{
	my $sid = shift;
	my $lsnrctlopRef = shift;

	my @result = ();
	foreach my $lsnr  ( keys %$lsnrctlopRef ) {
		if ( $lsnrctlopRef -> { $lsnr } { Lsnrctl } =~ /Instance \"$sid\"/) {
			push( @result, $lsnr );
		}
	}
	
END:
	return \@result;
}

sub get_listener_xml
{
	my $ref = shift;
	my $xmlstr = "<ListenerDetails>";
	foreach my $lsnr ( keys %$ref ) {

		my $home     = $ref -> { $lsnr } { Home };
		my $owner    = $ref -> { $lsnr } { Owner };
		my $EnvFile  = $ref -> { $lsnr } { EnvFile };
		my $TnsAdmin = $ref -> { $lsnr } { TnsAdmin };
		my $LsnrPwd  = $ref -> { $lsnr } { LsnrPwd };
		my $IP = $ref -> { $lsnr } { IP };

		$xmlstr .= "<Listener>";
		$xmlstr .= "<Name>$lsnr</Name>";
		$xmlstr .= "<Home><Value>$home</Value></Home>";
		$xmlstr .= "<Owner><Value>$owner</Value></Owner>";
		$xmlstr .= "<EnvFile><Value>$EnvFile</Value></EnvFile>";
		$xmlstr .= "<TnsAdmin><Value>$TnsAdmin</Value></TnsAdmin>";
		$xmlstr .= "<LsnrPwd><Value>$LsnrPwd</Value></LsnrPwd>";
		$xmlstr .= "<IP><Value>$IP</Value></IP>";
		$xmlstr .= "</Listener>";
	}
	$xmlstr .= "</ListenerDetails>";
}

sub _discover_sid
{
	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my ($_instance, $_orahome, $_autostart);
	my (%ret);
	# Skip the discovery of offline Oracle instances
	# Because, we can not discover the storage dependency of an offline Oracle instance
	goto CHECK_RUNNING_SIDS;

	my $_oratab;
	if ($^O =~ /solaris/i ) {
		$_oratab = "/var/opt/oracle/oratab";
	} else {
		$_oratab = "/etc/oratab";
	}

	if ( ! -f $_oratab ) {
		# ERROR: File $_oratab does not exist. Discovery not possible.
		goto CHECK_RUNNING_SIDS;
	}

	open (ORATAB, "<$_oratab") or do {
		# ERROR: File $_oratab cannot be read. Discovery not possible.
		goto CHECK_RUNNING_SIDS;
	};

	while (<ORATAB>) {
		# Skip blank lines or lines containing comments
		next if m/^\s*$/ or m/^\s*#/;

		# Remove comment part if exists
		s/\s*#.*$//;

		# Remove leading/trailing spaces if exist
		s/^\s*//;
		s/\s*$//;

		# Determine Owner attribute (i.e. user of $_orahome/bin/oracle file)
		($_instance, $_orahome, $_autostart) = split (/:/);

		$ret{$_instance}{'home'} = $_orahome;
		$ret{$_instance}{'autostart'} = $_autostart;
	}
	close ORATAB;


CHECK_RUNNING_SIDS:
	Utils::vxlog("INFO", "Discovering running SIDs ...");
	my $cmd = "ps -ef | grep ora_pmon_* | grep -v grep";
	my ($CmdRet, $op) = Utils::runCmd( $cmd );
	if ( $CmdRet != CONST_SUCCESS ) {
		goto END;
	}

	my @lines = split(/\n/, $op);
	foreach ( @lines ) {
		Utils::vxlog("INFO", "Processing [$_]");
		s/^\s+//;

		my @fields = split(/\s+/);
		my $pname  = $fields[-1];
		my $pid    = $fields[1];

		my $sid = "";
		if ( $pname =~ /ora_pmon_(.*)/ ) {
			$sid = $1;
		}

		Utils::vxlog("INFO", "SID=[$sid]");
		if (not defined $ret { $sid } ) {
			Utils::vxlog("INFO", "discovering home ...");
			my $home = read_process_env($pid, "ORACLE_HOME");
			$ret { $sid } = ();
			$ret { $sid } { 'home' } = $home;
		}
	}

END:

	return %ret;
}


sub read_process_env
{
	my $pid = shift;
	my $var = shift;

	my $value = "";
	my ($CmdRet, $op) = ();
	if ( $^O =~ /linux/i ) {
		($CmdRet, $op) = Utils::runCmd("(cat /proc/$pid/environ; echo) | tr \"\\000\" \"\n\"");
	} elsif( $^O =~ /aix/i ) {
		($CmdRet, $op) = Utils::runCmd("(ps eww $pid; echo) | tr \" \" \"\n\"");
	} elsif( $^O =~ /solaris/i ) {
		($CmdRet, $op) = Utils::runCmd("(/usr/ucb/ps -eww $pid; echo) | tr \" \" \"\n\"");
	}
	my @lines = split(/\n/, $op);
	foreach my $line ( @lines ) {
		if ( $line =~ /^$var=/) {
			my @fields = split(/=/, $line);
			$value = $fields[1] || "";
			goto END;
		}
	}

END:
	Utils::vxlog("INFO", "var=[$var], value=[$value]");
	return $value;	
}


sub createCommitSGDetails 
{
	my $self = shift;
	my $xmlstr =  shift;
	my $sMyName = ( caller 0 )[3];
	my $Name;
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);
	my ($insxml, $insResName) = ("","");
	my ( $hrInsResXMLs, $hrInsResNames ) = ();
	my $xmlstrCopy = $xmlstr;
	$xmlstrCopy =~ s/<LsnrPwd>.*<\/LsnrPwd>/<LsnrPwd>XXX<\/LsnrPwd>/gi;
	my $tree = Utils::convertXMLTextToHash($xmlstrCopy);
	Utils::vxlog("INFO", Dumper( $tree ));

	# use the normal xml
	$tree = Utils::convertXMLTextToHash($xmlstr);

	my $sgxml = "";
	my $ret = CONST_SUCCESS;
	my $msg = "";

	if ($tree eq "" or not defined $tree -> { Instance }) {
		$ret = APPCONTROL_ORACLE_COMMIT_INSTANCETAGNOTPROVIDED;
		$msg = "'Instance' tag not specified in the XML";
		goto END;
	}

	if (not defined $tree -> { ListenerDetails }) {
		$ret = APPCONTROL_ORACLE_COMMIT_LISTENERDETAILSTAGNOTPROVIDED;
		$msg = "'ListenerDetails' tag not specified in the XML";
		goto END;
	}

	my $str = "";
	my $ref = $tree -> { Instance };

	my @instances = ();
	if ( "$ref" =~ /^ARRAY/ ) {
		@instances = @$ref;
	} elsif ( "$ref" =~ /HASH/ ) {
		push( @instances, $ref );
	}

	# Validatation on inputs
	foreach my $ins ( @instances ) {

		Utils::vxlog("INFO", Dumper($ins));

		if ($ins eq "" or not defined $ins -> { Name }) {
			$ret = APPCONTROL_ORACLE_COMMIT_INSTANCENAMENOTPROVIDED ;
			$msg = "Name tag in Instance is not provided";
			goto END;
		}
		$Name = $ins -> { Name };
		if (not defined $ins -> { ListenerList }) {
			$ret = APPCONTROL_ORACLE_COMMIT_LISTENERTAGNOTPROVIDED;
			$msg = "ListenerList tag is not provided";
			goto END;
		}

		foreach my $attr ( qw[Owner Home Pfile EnvFile] ) {
			if (not defined $ins -> { $attr }) {
				$ret = APPCONTROL_ORACLE_COMMIT_ATTRNOTPROVIDED;
				$msg = "$attr tag is not provided";
				goto END;
			}
			if ( $ins -> { $attr } eq "" or not defined $ins -> { $attr } { Value } ) {
				$ret = APPCONTROL_ORACLE_COMMIT_ATTRVALUENOTPROVIDED;
				$msg = "Value tag for $attr tag is not provided";
				goto END;
			}
		}

		my $Home = $ins -> { Home } { Value };
	        $Home = Utils::strTrim( $Home );

        	if ( $Home ne "" and not -d $Home ) {
			$ret = APPCONTROL_ORACLE_COMMIT_ORACLEHOMENOTPRESENT;
			$msg = "specified Oracle Home directory [$Home] doesn't exists. Specify correct home path and retry.";
			goto END;
       		}

		my $Pfile = $ins -> { Pfile } { Value };
		$Pfile = Utils::strTrim( $Pfile );

		if ( $Pfile ne "" and not -e $Pfile ) {
			$ret = APPCONTROL_ORACLE_COMMIT_PFILENOTPRESENT;
			$msg = "specified pfile [$Pfile] doesn't exists. Specify correct pfile path and retry.";
			goto END;
		}

		my $EnvFile = $ins -> { EnvFile } { Value };
		$EnvFile = Utils::strTrim( $EnvFile );

		if ( $EnvFile ne "" and not -e $EnvFile ) {
			$ret = APPCONTROL_ORACLE_COMMIT_ORACLEENVFILENOTPRESENT;
			$msg = "specified Oracle environment file [$EnvFile] doesn't exists. Specify correct environment file path and retry.";
			goto END;
		}
		if (( defined ($ins->{MountPointList})) and ($ins -> {MountPointList} ne "") and
		    ( defined ($ins->{MountPointList}{MountPoint})) ) {
			my @sMountPoints = Utils::convertHashValuesToArray ($ins ->{MountPointList}{MountPoint});
			foreach my $sMountPoint (@sMountPoints) {
				if ($sMountPoint->{Selected} =~ /true/) {
					$self->setMountDependency($Name, $sMountPoint->{Value});
				}
			}
		}

		$self->addInstance($Name);
	}

	$ref = "";
	if ( defined $tree -> { ListenerDetails } and $tree -> { ListenerDetails } ne "" and $tree -> { ListenerDetails } { Listener } ) {
		$ref = $tree -> { ListenerDetails } { Listener } || "";
	}

	my @lsnrs = ();
	if ( "$ref" =~ /^ARRAY/ ) {
		@lsnrs = @$ref;
	} elsif ( "$ref" =~ /HASH/ ) {
		push( @lsnrs, $ref );
	}

	# Validatation on inputs
	foreach my $lsnr ( @lsnrs ) {

		if ($lsnr eq "" or not defined $lsnr -> { Name }) {
			$ret = APPCONTROL_ORACLE_COMMIT_LISTENERNAMENOTPROVIDED ;
			$msg = "Name tag in Listener is not provided";
			goto END;
		}
		$Name = $lsnr -> { Name };

		foreach my $attr ( qw[Owner Home EnvFile TnsAdmin LsnrPwd] ) {
			if (not defined $lsnr -> { $attr }) {
				$ret = APPCONTROL_ORACLE_COMMIT_ATTRNOTPROVIDED;
				$msg = "$attr tag is not provided";
				goto END;
			}
			if ( $lsnr -> { $attr } eq "" or not defined $lsnr -> { $attr } { Value } ) {
				$ret = APPCONTROL_ORACLE_COMMIT_ATTRVALUENOTPROVIDED;
				$msg = "Value tag for $attr tag is not provided";
				goto END;
			}
		}

		my $Home = $lsnr -> { Home } { Value };
	        $Home = Utils::strTrim( $Home );

        	if ( $Home ne "" and not -d $Home ) {
			$ret = APPCONTROL_ORACLE_COMMIT_LISTENERHOMENOTPRESENT;
			$msg = "specified Listener Home directory [$Home] doesn't exists. Specify correct home path and retry.";
			goto END;
        	}

		my $TnsAdmin = $lsnr -> { TnsAdmin } { Value };
		$TnsAdmin = Utils::strTrim( $TnsAdmin );

		if ( $TnsAdmin ne "" and not -d $TnsAdmin ) {
			$ret = APPCONTROL_ORACLE_COMMIT_TNSADMINNOTPRESENT;
			$msg = "specified TNS ADMIN directory [$TnsAdmin] doesn't exists. Specify correct TNS ADMIN path and retry.";
			goto END;
		}

		my $EnvFile = $lsnr -> { EnvFile } { Value };
		$EnvFile = Utils::strTrim( $EnvFile );

		if ( $EnvFile ne "" and not -e $EnvFile ) {
			$ret = APPCONTROL_ORACLE_COMMIT_LISTENERENVFILENOTPRESENT;
			$msg = "specified Listener environment file [$EnvFile] doesn't exists. Specify correct environment file path and retry.";
			goto END;
		}

		my $LsnrPwd = $lsnr -> { LsnrPwd } { Value };
		if ( $LsnrPwd ne "" ) {
			my $doNotLogCmd = 1;
			my ($CmdRet, $op ) = Utils::encrypt_password ($LsnrPwd);
			if ($CmdRet != CONST_SUCCESS ) {
				$ret = $CmdRet;
				$msg = "failed while executing of vcsencrypt";
				goto END;
			}
			$LsnrPwd = $op;
		}

		$lsnr -> { LsnrPwd } { Value } = $LsnrPwd;

		Utils::vxlog("INFO", Dumper($lsnr));
	}

	my ($SettingFound, $AppRestartLimit) = ();
	$SettingFound = 0;
	$AppRestartLimit = 1;
	Utils::vxlog("INFO", "SettingFound=[$SettingFound], AppRestartLimit=[$AppRestartLimit]");

	my $vcsobj = new VCS();
	#my $VCSHostName = VCS::getVCSNodeName();
	my $AttrScope = "global";

	$sgxml .= "<xml>";
	# TODO: need to change the service group logic for multiple service groups.
	my $sgName = "Oracle_SG";
	$sgxml .= "<Name>$sgName</Name>";

	$sgxml .= "<Attributes>";
	#$sgxml .= "<AutoStartList><$AttrScope>$VCSHostName</$AttrScope></AutoStartList>";
	if($tree->{SystemList})
	{
		my $refSys = $tree->{SystemList};
		my $uniqSys = ();
		my @systemList = ();
		if ( ref($refSys) =~ /^ARRAY/ ) {
		    @systemList = @$refSys;
		} else {
		    push( @systemList, $refSys );
		}
		foreach my $sl(@systemList) {
		    if ( ref ($sl) !~ /^HASH/ || not defined $sl->{System} ) {
			next;
		    }
		    my $refSystems = $sl->{System};
		    my @systems = ();
		    if ( ref($refSystems) =~ /^ARRAY/ ) {
		    	@systems = @$refSystems;
		    } else {
			    push( @systems, $refSystems );
		    }
		    $sgxml .= "<SystemList>";
		    foreach my $sys ( @systems ) {
			    if (defined $uniqSys -> {$sys->{Name}} and 
					$uniqSys -> {$sys->{Name}} == 1) {
				    next;
			    }
			    $uniqSys -> {$sys->{Name}} = 1;
			    $sgxml .= "<global>$sys->{Name}</global>";
			    $sgxml .= "<global>$sys->{Priority}</global>";
		    }
		    $sgxml .= "</SystemList>";
		    last;
		}
	  }
	 
	$sgxml .= "</Attributes>";

	Utils::vxlog("INFO", "Setting SG Name as [$sgName]");


	Utils::vxlog("INFO", "Processing for Netlsnr resources");
	

	# new xml will not require above information
	$sgxml = "";
	#---------------------------
	# Create Listener resources
	#---------------------------
	foreach my $lsnr ( @lsnrs ) {
		
		my $Listener = $lsnr -> { Name };
		$Name = $lsnr -> { Name };

		my $Home     = $lsnr -> { Home } { Value };
		my $Owner    = $lsnr -> { Owner } { Value };
		my $TnsAdmin = $lsnr -> { TnsAdmin } { Value };
		my $EnvFile  = $lsnr -> { EnvFile } { Value };
		my $LsnrPwd  = $lsnr -> { LsnrPwd } { Value };

		my $lsnrxml = "<Resource>";

		my $ResName = "LSNR_" . $Listener . "_res";
		$ResName =  $self->GetUniqueResName($ResName);
		Utils::vxlog("INFO", "Listener Res=[$ResName]");
		$lsnrxml .= "<Name>$ResName</Name>";
		$lsnrxml .= "<Type>$LsnrResType</Type>";
		$lsnrxml .= "<SG>$sgName</SG>";
		
		$lsnrxml .= "<Attributes>";
		$lsnrxml .= "<Home><$AttrScope>$Home</$AttrScope></Home>";
		$lsnrxml .= "<Owner><$AttrScope>$Owner</$AttrScope></Owner>";
		$lsnrxml .= "<Listener><$AttrScope>$Listener</$AttrScope></Listener>";
		$lsnrxml .= "<TnsAdmin><$AttrScope>$TnsAdmin</$AttrScope></TnsAdmin>";
		$lsnrxml .= "<EnvFile><$AttrScope>$EnvFile</$AttrScope></EnvFile>";
		$lsnrxml .= "<LsnrPwd><$AttrScope>$LsnrPwd</$AttrScope></LsnrPwd>";
		$lsnrxml .= "</Attributes>";

		$lsnrxml .= "</Resource>";

		$hrInsResXMLs -> { $Name } = $lsnrxml;
		$hrInsResNames -> { $Name } = $ResName;
	}

	#-----------------
	# Oracle Resources
	#-----------------
	foreach my $ins ( @instances ) {

		my $InstName = $ins -> { Name };
		my $Owner	 = $ins -> { Owner	 } { Value };
		my $Home	 = $ins -> { Home	 } { Value };
		my $Pfile	 = $ins -> { Pfile	 } { Value };
		my $EnvFile	 = $ins -> { EnvFile } { Value };

		$Name = $ins -> { Name };
		# Oracle Resource
		#$sgxml .= "<Resource>";
		my $oraxml = "<Resource>";
		my $resName = "Oracle_" . $InstName . "_res";;
		$resName =  $self->GetUniqueResName($resName);
		Utils::vxlog("INFO", "Oracle Res=[$resName]");

		$oraxml .= "<Name>$resName</Name>";
		$oraxml .= "<Type>$OracleResType</Type>";
		$oraxml .= "<SG>$sgName</SG>";

		$oraxml .= "<Attributes>";
		$oraxml .= "<Sid><$AttrScope>$InstName</$AttrScope></Sid>";
		$oraxml .= "<Owner><$AttrScope>$Owner</$AttrScope></Owner>";
		$oraxml .= "<Home><$AttrScope>$Home</$AttrScope></Home>";
		$oraxml .= "<Pfile><$AttrScope>$Pfile</$AttrScope></Pfile>";
		$oraxml .= "<EnvFile><$AttrScope>$EnvFile</$AttrScope></EnvFile>";
		$oraxml .= "</Attributes>";

		$oraxml .= "</Resource>";
			$hrInsResXMLs -> { $Name } = $oraxml;
			$hrInsResNames -> { $Name } = $resName;
	}

	Utils::vxlog("INFO", "Creating resource dependencies (if any)");
	foreach my $ins ( @instances ) {

		my $oraName = $ins -> { Name };

		Utils::vxlog("INFO", "InstanceName=[$oraName]");

		my $ref = "";
		if ( $ins -> { ListenerList } ne "" and defined $ins -> { ListenerList } { Listener } ) {
			$ref = $ins -> { ListenerList } { Listener } || "";
		}
		my @sid_lsnrs = ();
		if ( "$ref" =~ /ARRAY/ ) {
			@sid_lsnrs = @$ref;
		} elsif ( "$ref" =~ /HASH/ ) {
			push(@sid_lsnrs, $ref);
		}

		foreach my $lsnr ( @sid_lsnrs ) {
			my $lsnrName     = $lsnr -> { Value };
			my $selected = $lsnr -> { Selected };

			if ( $selected !~ /true/i ) {
				next;
			}

			$self->setInstanceDependency($lsnrName, $oraName);
		}
	}

	$sgxml .= "</xml>";
	my $sSystemNICSGxml = "";
	($ret, $sSystemNICSGxml) = $self->getNetworkXML($tree);
	# TODO: check $ret
	my $treeXML = $self->createSGXML ($tree, $hrInsResXMLs, $hrInsResNames, "ORA", "Oracle", "Oracle Application");

	$sgxml = "<xml>";
	$sgxml .= $sSystemNICSGxml;
	$sgxml .= $treeXML;
	$sgxml .= "</xml>";
END:

	Utils::vxlog("INFO", $sgxml);
	return ({ 'ret' => $ret, 'msg' => $msg, 'xml' => $sgxml});
}

sub checkAndAddApplicationResTypes
{
	my $self = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $ret = CONST_SUCCESS;
	my $msg = "";

	my $vcsobj = new VCS();
	$vcsobj -> makeConfigWritable();

	# check if passed type is present or not
#	my $CmdRet = $vcsobj -> isResTypeAvailable ( $OracleResType );
#	if ( $CmdRet != APPCONTROL_YES ) {
#		$CmdRet = $self -> addResType( $OracleResType );
#		if ( $CmdRet != CONST_SUCCESS ) {
	#		$ret = APPCONTROL_ORACLE_COMMIT_ADDRESTYPEFAILURE;
	#		$msg = "Failed to add [$OracleResType] type";
	#		goto EXIT_LABEL;
	#	}
	#}

	$vcsobj -> dumpAndMakeConfigReadOnly();

EXIT_LABEL:
	return ($ret, $msg);
}


sub addResType
{
	my $self = shift;
	my $resType = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $hacf = "/opt/VRTSvcs/bin/hacf";

	my $CmdRet = CONST_SUCCESS;
	my $op = "";

	my $cmd = "mkdir -m 1700 -p /tmp/OracleTypes";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ( $CmdRet != CONST_SUCCESS ) {
		return $CmdRet;
	}

	$cmd = "cp /etc/VRTSagents/ha/conf/Oracle/OracleTypes.cf /tmp/OracleTypes/main.cf";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ( $CmdRet != CONST_SUCCESS ) {
		return $CmdRet;
	}

	$cmd = "$hacf -cftocmd /tmp/OracleTypes";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ( $CmdRet != CONST_SUCCESS ) {
		return $CmdRet;
	}

	$cmd = "sed s/main.cf/OracleTypes.cf/g /tmp/OracleTypes/main.cmd > /tmp/OracleTypes/OracleTypes.cmd";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ( $CmdRet != CONST_SUCCESS ) {
		return $CmdRet;
	}

	$cmd = "/tmp/OracleTypes/OracleTypes.cmd";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ( $CmdRet != CONST_SUCCESS ) {
		Utils::vxlog("INFO", "file run failed [$CmdRet] [$op]");
		return $CmdRet;
	}

	$cmd = "rm -rf /tmp/OracleTypes";
	($CmdRet, $op) = Utils::runCmd($cmd);
	# no need to check return code! harmless error (if any)

	return CONST_SUCCESS;
}

# args: ORACLE_HOME and SID
sub _get_oradata_path()
{
	my $home = shift;
	my $sid = shift;
	my $user = shift;
	my $sMyName = ( caller 0 )[3];
	my (@orapaths)= ();
	my (@archivepath)= ();
	my (@_datapath)= ();
	my $env = "export ORACLE_HOME=$home; export ORACLE_SID=$sid; export LD_LIBRARY_PATH=$home/lib:\\\$LD_LIBRARY_PATH";
	my $cmd = "su - $user -c '$env; $home/bin/sqlplus  -s /nolog' << !
connect / as sysdba
select NAME from v\\\$datafile; 
select NAME from v\\\$controlfile; 
select NAME from v\\\$tempfile; 
select MEMBER from v\\\$logfile; 
!";
	my ($CmdRet, $op) = Utils::runCmd($cmd);
	if ($CmdRet == CONST_SUCCESS) {
		@_datapath = keys %{{map { s/\/[^\/]*$//; $_ => undef; } (grep (/(\/.*)\//, (split (/\n/, $op))))}};
	}

	$cmd = "su - $user -c '$env; $home/bin/sqlplus  -s /nolog' << !
connect / as sysdba
archive log list; 
!";
	($CmdRet, $op) = Utils::runCmd($cmd);
	if ($CmdRet == CONST_SUCCESS) {
		my @lines = split(/\n/, $op);
		foreach my $line (@lines) {
			if ($line =~ /^Archive destination\s+(\/\S+)$/) {
				push (@archivepath,$1);
			}
		}
	}

	@orapaths = (@_datapath,@archivepath);

	return \@orapaths;
}
# placeholder for actual function which is going to be used for path to mountpoint conversion
sub _get_mnt_from_path()
{
	my $home = shift;
	return $home;
}
sub getIPParentInstance
{
    my $self = shift;
    my $resRef = shift;
    my $lsnrResName= shift;
	if (defined $resRef-> { $lsnrResName } {Listener} {global}) {
		return $resRef-> { $lsnrResName } {Listener} {global};
	} else {
		return "";
	}
}

# This function would be called discoverExistingNetworkConfig to populate
# the instances to be shown on Network page
sub getDiscoveredInstances
{
    my $self = shift;
    my $resRef = shift;
    my @resList = ();
    foreach my $resName (keys %$resRef) { 
	if ($resRef->{$resName}{Type}{global} !~ /^Netlsnr$/) {
            next;
        }
	if (defined $resRef->{ $resName }{Listener}) {
   	    push (@resList, $resRef->{ $resName }{Listener}{global});
        }
    }

    return \@resList;

}

sub isOracleInstancePresent
{
    my $self = shift;
	my $sMyName = ( caller 0 )[3];
    my $sid = shift;
    my $resRef = shift;
    my $ret = CONST_FAILURE;

	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

    if ((ref $resRef) !~ /HASH/) {
		return $ret;
    }
    foreach my $resName (keys %$resRef) { 
		if ($resRef->{$resName}{Type}{global} !~ /^Oracle$/) {
            next;
        }
		if ((defined $resRef->{ $resName }{Sid}) and 
			($resRef->{ $resName }{Sid}{global} eq $sid)) {
			$logStr = "Resource for Oracle instance" . $sid . "is already created.";
			Utils::vxlog("INFO", $logStr);
		   	$ret = CONST_SUCCESS;
   	 		return $ret;
        }
    }
    return $ret;
}
sub isListenerPresent
{
    my $self = shift;
	my $sMyName = ( caller 0 )[3];
    my $lsnrName = shift;
    my $resRef = shift;
    my $ret = CONST_FAILURE;

	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

    if ((ref $resRef) !~ /HASH/) {
		return $ret;
    }
    foreach my $resName (keys %$resRef) { 
		if ($resRef->{$resName}{Type}{global} !~ /^Netlsnr$/) {
            next;
        }
		if ((defined $resRef->{ $resName }{Listener}) and 
			($resRef->{ $resName }{Listener}{global} eq $lsnrName)) {
			$logStr = "Resource for Oracle instance" . $lsnrName . "is already created.";
			Utils::vxlog("INFO", $logStr);
			$ret = CONST_SUCCESS;
			return $ret;
		}
    }
    return $ret;
}

# Get existing listeners
# We should not allow users to add these from GUI.
sub get_existing_listeners
{
    my $self = shift;
	my $sMyName = ( caller 0 )[3];
    my @lsnrList = ();
    my $resRef = shift;
    my $ret = CONST_FAILURE;

	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

    if ((ref $resRef) !~ /HASH/) {
		return \@lsnrList;
    }
    foreach my $resName (keys %$resRef) { 
		if ($resRef->{$resName}{Type}{global} !~ /^Netlsnr$/) {
            next;
        }
		if (defined $resRef->{ $resName }{Listener}) {
			my $rec = ();
			$rec -> {Home} =$resRef->{ $resName }{Home}{global};
			$rec -> {Instance} = $resRef->{ $resName }{Listener}{global};

			push (@lsnrList, $rec);
		}
    }
    return \@lsnrList;
}
# This function is called by discoverExistingNetowrkConfig during add node wizard.
# it returns the application specific information to the wizard.
# The format of data is similar to the discovery wizard
# We also simulate selection of instances so that there is no changes required in 
# the shallowDiscovery subroutine.

sub getValidationInput
{
    my $self = shift;
	my $sMyName = ( caller 0 )[3];
    my $resRef = shift;
    my $clusRef = shift;
    my $ret = CONST_SUCCESS;
	my ($sid, $oraHome, $owner, $ownerUID)  = ("", "", "", -1);
	my $outXML = "";
	my @instData = ();
	$outXML = "<ValidationInfo>";
    foreach my $resName (keys %$resRef) { 
		my $data = ();
		if ($resRef->{$resName}{Type}{global} !~ /^Oracle$/) {
            next;
        }
		if (defined $resRef->{ $resName }{Sid}) {
			$sid = $resRef->{ $resName }{Sid}{global};
        }
		if (defined $resRef->{ $resName }{Home}) {
			$oraHome = $resRef->{ $resName }{Home}{global};
        }
		if (defined $resRef->{ $resName }{Owner}) {
			$owner = $resRef->{ $resName }{Owner}{global};
			my $cmd = "/usr/bin/getent passwd $owner";
			my $CmdRet = CONST_SUCCESS;
			($CmdRet, $ownerUID) = Utils::runCmd($cmd);
			if ( $CmdRet != CONST_SUCCESS ) {
				return ( APPCONTROL_ORACLE_DISCOVER_USERDISCOVERYFAILURE , "Failed to discover Owner!", "");
			}
			# Owner UID is required for validating the shallow disovery on other nodes in the cluster.
			$ownerUID =~ m/^.*?:.*?:(\d+)/;
			$ownerUID = $1;
        }
		$data->{Sid} = $sid;
		$data->{Instance} = $sid;
		$data->{Home} = $oraHome;
		$data->{Owner} = $owner;
		$data->{OwnerUID} = $ownerUID;
		my $arMountPoint = ();
		(my $CmdRet, $arMountPoint) = $self->getDependentMountPointList($resName, $resRef, $clusRef);
		if ($CmdRet == CONST_SUCCESS) {
			$data -> { MountPoint } = $arMountPoint;
		}
		push(@instData, $data);
    }
	$outXML .= $self -> createApplicationXML( \@instData, "Instance");
	my $selectXML = "<Selected>true</Selected>";
	# Selected tags needs to be added in the XML for each instance
	$outXML =~ s/<Instance>/<Instance>$selectXML/g;
	# adding selected tag for each MountPoint
	$outXML =~ s/<MountPoint>/<MountPoint>$selectXML/g;
	$outXML .= "</ValidationInfo>";
    return ($ret, "", $outXML);
}


# This function is called to get description for service groups
# for displaying on UI
sub getSGDescription
{
	my $self = shift;
	my $sg_name = shift;
	my $resources = shift;
	my $msg = "Returning description for service group $sg_name : ";
	my $sg_desc = "";
	my $instances = "";
	foreach my $res (@$resources) {
		my $res_xml = $self -> getResourceXML ( $res );
		my $res_details = Utils::convertXMLTextToHash( $res_xml );

		if ( defined ($res_details->{Type}) && $res_details->{Type} eq "Oracle" &&
			defined $res_details->{Attributes}{Sid}{global} ) {
			$instances .=  $res_details->{Attributes}{Sid}{global} . ", " ;
			Utils::vxlog("DEBUG", "Instance name: $res_details->{Attributes}{Sid}{global}");
		}
	}
	if ( $instances ne "" ) {
		$instances =~ s/(.*)\, /$1/;
		$sg_desc = "This service group contains resources that relate to the following Oracle instance(s): " . $instances . ".";
	}
	$msg .= $sg_desc;
	Utils::vxlog("DEBUG", $msg);
	return $sg_desc;
}

1; # so that perl doesn't complain!
