
#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Id: SAP.pm,v 1.36 2014/06/23 14:32:30 avaidya Exp $
#

package SAP;
use UnixApplication;

@ISA = ("UnixApplication");  # required for Inheritance

use strict;
use warnings;

use Scalar::Util 'blessed';
use UnixSAPCodes;
use UnixNetwork;
use Data::Dumper qw(Dumper);
use Constants;
use Utils;
use PlatUtils;
use Storage;
use XML::Simple;


#------------------------
# Globals ...
#------------------------
my ($ls, $SAPResType,$mkdir,$ln) = ();

$ls = "/bin/ls";
$mkdir = "/bin/mkdir";
$ln = "/bin/ln";
$SAPResType = "SAPWebAS71";
#--------------------
# Sub routines
#--------------------
sub getInstances
{
	my $self = shift;
	my $doNotHideInstances = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	my @instances = ();

	# /sapmnt has SAP systems (SIDs), e.g. ER1
	my $cmd = "$ls -1 /sapmnt/ 2>&1";
	my ($ret, $lsOp) = Utils::runCmd ( $cmd );
	if ($ret != CONST_SUCCESS) {
		goto EXIT_LABEL;
	}

	my @SAPSystems = split(/\n/, $lsOp);
	foreach my $sapSys ( @SAPSystems ) {
		$sapSys = Utils::strTrim($sapSys);

		# for each SAP system, there can be multiple instances available
		my $profileDirName = "/sapmnt/$sapSys/profile";
		if (not -d $profileDirName ) {
			next;
		}

		my $fileName = "$profileDirName/$sapSys" . "_*";
		$cmd = "$ls -1 $fileName";
		($ret, $lsOp)  = Utils::runCmd ( $cmd );
		if ($ret != CONST_SUCCESS) {
			next;
		}

		my @lines = split(/\n/, $lsOp);
		foreach my $entry ( @lines ) {
			chomp $entry;

			my $instanceName = "";
			if ( $entry =~ m/\/sapmnt\/$sapSys\/profile\/$sapSys\_([A-Z]+[0-9]+)_/ ) {
				$instanceName = $1;
			}

			if ( defined $instanceName and $instanceName ne "" ) {
				# Skipping those instances which are not monitored by SAPWebAS71 agent 
				if ( $instanceName !~ /^AREP|^ERS|^REP|^ASCS|^SCS|^D|^J|^SMDA/i ) {
					Utils::vxlog("INFO", "Skipping [$instanceName]");
					next;
				}
				if ( not grep /$instanceName/, @instances ) {
					$instanceName = $instanceName . " ($sapSys)";
					Utils::vxlog("INFO", "Adding [$instanceName]");
					push(@instances, $instanceName);
				}
			}
		}
	}
EXIT_LABEL:
	return \@instances;
}

sub discover
{
	my $self = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);
	my $ret = CONST_SUCCESS;
	my $xml = "";
	my $msg = "Discovery of existing instances and monitoring configurations is successful.";
	my $packageName = "" ;

  $ret = PlatUtils::isAcclibPackageInstalled("SAP");

  if ( $ret != PlatUtils::PACKAGE_INSTALLED ) {
    $msg = "Package [VRTSacclib] is not installed for SAP.";
    Utils::vxlog("INFO", $msg);
    return (APPCONTROL_SAP_DISCOVER_ACCLIBNOTINSTALLED, $msg, $xml);
  }
  else
  {
    $logStr = "Package [VRTSacclib] is installed for SAP.";
    Utils::vxlog("INFO", $logStr);
  } 

	($ret,$packageName) = PlatUtils::isAgentPackageInstalled("SAP");
	
	if ( $ret != PlatUtils::PACKAGE_INSTALLED ) {
		$msg = "Package [$packageName]is not installed for SAP.";
		Utils::vxlog("INFO", $msg);
		return (APPCONTROL_SAP_DISCOVER_SAPAGENTNOTINSTALLED, $msg, $xml);
	}	
	else
	{
		$logStr = "Package [$packageName]is installed for SAP.";
		Utils::vxlog("INFO", $logStr);
	}
	
	($ret,$msg, $xml) = $self->doDiscovery();

	 #$xml = "<DiscoveryData><Instance><Name>DVEBMGS01</Name><InstProfile><Value>/sapmnt/VB5/profile/VB5_DVEBMGS01</Value></InstProfile><SAPHome><Value>/usr/sap/VB5/DVEBMGS01/exe</Value></SAPHome><SystemType><Value>ABAP</Value></SystemType><ProcMon><Value>en ms</Value></ProcMon><SAPSID><Value>VB5</Value></SAPSID><SAPAdmin><Value>vb5adm</Value></SAPAdmin><InstType><Value>APPSERV</Value></InstType><InstName><Value>DVEBMGS01</Value></InstName><EnvFile><Value>/home/vb5adm/.login</Value></EnvFile></Instance><Instance><Name>DVEBMGS00</Name><InstProfile><Value>/sapmnt/VBS/profile/VBS_DVEBMGS00</Value></InstProfile><SAPHome><Value>/usr/sap/VBS/DVEBMGS00/exe</Value></SAPHome><SystemType><Value>ABAP</Value></SystemType><ProcMon><Value>en ms</Value></ProcMon><SAPSID><Value>VBS</Value></SAPSID><SAPAdmin><Value>vbsadm</Value></SAPAdmin><InstType><Value>APPSERV</Value></InstType><InstName><Value>DVEBMGS00</Value></InstName><EnvFile><Value>/home/vbsadm/.login</Value></EnvFile></Instance> <Instance><Name>ASCS01</Name><InstProfile><Value>/sapmnt/VB5/profile/VB5_ASCS01_sapvm1</Value></InstProfile><SAPHome><Value>/usr/sap/VB5/ASCS01/exe</Value></SAPHome><SystemType><Value>ABAP</Value></SystemType><ProcMon><Value>en ms</Value></ProcMon><SAPSID><Value>VB5</Value></SAPSID><SAPAdmin><Value>vb5adm</Value></SAPAdmin><InstType><Value>ENQUEUE</Value></InstType><InstName><Value>ASCS01</Value></InstName><EnvFile><Value>/home/vb5adm/.login</Value></EnvFile></Instance><Instance><Name>ASCS00</Name><InstProfile><Value>/sapmnt/VBS/profile/VBS_ASCS00_sapvm1</Value></InstProfile><SAPHome><Value>/usr/sap/VBS/ASCS00/exe</Value></SAPHome><SystemType><Value>ABAP</Value></SystemType><ProcMon><Value>en ms</Value></ProcMon><SAPSID><Value>VBS</Value></SAPSID><SAPAdmin><Value>vbsadm</Value></SAPAdmin><InstType><Value>ENQUEUE</Value></InstType><InstName><Value>ASCS00</Value></InstName><EnvFile><Value>/home/vbsadm/.login</Value></EnvFile></Instance></DiscoveryData>";
        Utils::vxlog("INFO", "Discovered SAP monitoring configuration XML.");
        Utils::vxlog("INFO", $xml);

        Utils::vxlog("INFO", "returned [$ret]");

	return ($ret, $msg, $xml);
}

sub doDiscovery
{
	my $self = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	my $ret = CONST_SUCCESS;
	my $xmlstr = "";
	my $msg = "Discovery of existing instances and monitoring configurations is successful.";
	
	my @insdata = ();
	my @Storageinsdata = ();
	my @SysDiskIDs = ();
	my $SysDiskIDdata = ();

	my $doNotHideInstances = 1;
	my $instances = $self -> getInstances( $doNotHideInstances );

  	if( (scalar @$instances ) == 0 )
  	{
    		$ret = CONST_FAILURE;
		my $msg = "Failed to discover SAP Instances.";
    		return ($ret,$msg,$xmlstr);
  	}

        my $clusDiscoveryInputRef   = ();
        $clusDiscoveryInputRef->{Resources}{Types}          = [ "SAPWebAS71" ] ;
        my $instExistingList = ();
        my $SGDependency = ();
        my $instSidResourceNameHash = ();
        ($instExistingList,$SGDependency,$instSidResourceNameHash) = $self->getExistingConfigurations();
        my @TempInstExistingList = ();
	
	if(defined $instExistingList)
	{
		foreach my $data ( @$instExistingList ) {
			my $insName = $data -> { InstName } ;
			my $insSAPSID = $data -> { SAPSID } ;
			my $tempString = $insSAPSID . "_" . $insName ;
			push (@TempInstExistingList, $tempString);
		}
	}



	if ( grep (/^ERS./, @$instances) )
	{
	        my $preonlineDirName = "/opt/VRTSvcs/.sapwebas71/preonline/";

		if (not -d $preonlineDirName ) {

			my $cmd = "$mkdir -p $preonlineDirName";
			my ($cmdRet, $lsOp)  = Utils::runCmd ( $cmd );
			if ($cmdRet != CONST_SUCCESS) {
				$logStr = "Directory $preonlineDirName for preonline trigger could not be created. Verify that write permissions are enabled for the /opt/VRTSvcs directory";
				$msg = "Directory $preonlineDirName for preonline trigger could not be created. Verify that write permissions are enabled for the /opt/VRTSvcs directory";
				Utils::vxlog("INFO", $logStr);
				return ($cmdRet,$msg,$xmlstr);
			}

                }
		
		my $preonlineFileName = "/opt/VRTSvcs/.sapwebas71/preonline/T1preonline";

		if ( ( not -f $preonlineFileName )  && ( not -l $preonlineFileName ) ) {
			my $cmd = "$ln -s /etc/VRTSagents/ha/conf/SAPWebAS71/preonline $preonlineFileName";
			my ($cmdRet, $lsOp)  = Utils::runCmd ( $cmd );

			if ($cmdRet != CONST_SUCCESS) {
				$logStr = "File $preonlineFileName for preonline trigger could not be created. Verify that write permissions are enabled for the $preonlineDirName directory";
				$msg = "File $preonlineFileName for preonline trigger could not be created. Verify that write permissions are enabled for the $preonlineDirName directory";
				Utils::vxlog("INFO", $logStr);
				return ($cmdRet,$msg,$xmlstr);
			}
		}

	}

	foreach my $insRef ( @$instances ) {
		my $data = ();
		my ($arApplicationPath, $arApplicationMounts) = ();
		
		my $ins = "";
		my $sapSys = "";
		$ret = CONST_SUCCESS;

		if ( $insRef =~ m/([a-zA-Z0-9]+) \((.*)\)/ ) {
			$ins = $1;
			$sapSys = $2;
		}

		if ($ins eq "" or $sapSys eq "") {
			return (APPCONTROL_SAP_DISCOVER_INSTANCENOTPROVIDED , "Instance Name is not correctly provided",$xmlstr);
		}


		my $logStr = "SAP Instance = [$ins], SAPSID = [$sapSys]";
		Utils::vxlog("INFO", $logStr);

		#---------------------------
		# SAP resource information
		#---------------------------
		my @attrs = qw( SystemType
						InstName
						InstType
						EnvFile
						ProcMon
						SAPAdmin
						SAPHome
						SAPSID
						IsLocalInst
						InstProfile);

		my $pushInDiscoveryArray = 1;
		
		foreach my $attr ( @attrs ) {
			my $Value = "";
			my $function = "discover_$attr";
			($ret, $Value) = $self -> $function ( $ins,$sapSys);
			if ( $ret != CONST_SUCCESS ) {
				$xmlstr = "unable to discover '$attr' for Instance $ins inside $sapSys";
				goto EXIT_LABEL;
			}
        # Set the value always,if needed values can be overridden 
				$data -> { $attr } = $Value;

			if ( $attr =~ /IsLocalInst/ ) {
				my $insName = $data -> { InstName };
				my $insSAPSID = $data -> { SAPSID };
				my $tempString = $insSAPSID . "_" . $insName ;
				
				if( grep /$tempString/, @TempInstExistingList ){
					$data -> { $attr } = "The instance is already configured.";
				}
			}

			if ( $attr =~ /InstName/ ) {
				if ( $Value =~ /^W/i ) {
					Utils::vxlog("INFO", "Skipping ". $data -> { InstName });
					$pushInDiscoveryArray = 0;
					last;
				}
			}
		}
		
		if ( $pushInDiscoveryArray == 0 ) {
			next;
		}

		#------------------
		# basic information
		#------------------
		$data -> { Instance } = $ins;
		my $sDataPath = $data -> { SAPHome } ;
		push (@$arApplicationPath,$sDataPath);

	        # Storage discovery for the instance
                my ($hrStorageDetails) = ();
                my ($sMountPoint, $sBlockDevice, $sFSType, $sDiskGroup, $sVolume, $sVolumeGroup, $sLogicalVolume) = ();
                Utils::vxlog("INFO", "Discovered storage for SAP Instance [$ins] with DataPath [$sDataPath]");


    		my $Storage_ret = CONST_SUCCESS;
    		my $arDiskIDs;
    		my $iCnt =1;
    		foreach my $sApplicationPath ( @$arApplicationPath) { 
      			($Storage_ret, $hrStorageDetails) = $self->GetStorageDetails ($sApplicationPath);
      			if ( CONST_SUCCESS == $Storage_ret ) {
       				 push(@$arApplicationMounts,$hrStorageDetails->{MountPoint});
       				 $hrStorageDetails -> { Instance } = "$ins" . "_" . "$iCnt";

       				 if ( defined $hrStorageDetails -> { DiskGroup } ) {
       						$sDiskGroup = $hrStorageDetails -> { DiskGroup };
							($Storage_ret, $msg, $arDiskIDs) = getVxVMDgDisks($sDiskGroup);
							if ($Storage_ret != CONST_SUCCESS) {
								$ret = $Storage_ret;
								goto EXIT_LABEL;
							}
       				 }
       				 if ( defined $hrStorageDetails -> { VolumeGroup } ) {
       					   	$sVolumeGroup = $hrStorageDetails -> { VolumeGroup };
							($Storage_ret, $msg, $arDiskIDs) = getLVMVgDisks($sVolumeGroup);
							if ($Storage_ret != CONST_SUCCESS) {
								$ret = $Storage_ret;
								goto EXIT_LABEL;
							}
       				 }

       				 $hrStorageDetails -> { DiskID }  = $arDiskIDs;
       				 # Generate uniq DiskIDs
       				 foreach my $DiskID ( @$arDiskIDs ) {
       					   $SysDiskIDdata -> { $DiskID } = 1;
       				 }
			        push ( @Storageinsdata, $hrStorageDetails);
     			 } else {
       				 Utils::vxlog("INFO", "Could not discover storage for [$sDataPath]");
      			}
      			$iCnt++;
    		} # for loop for all application mounts
     		$data -> { MountPoint }  = $arApplicationMounts;
		push ( @insdata, $data );
	}

    $xmlstr = "<DiscoveryData>";
    $xmlstr .= $self->createApplicationXML( \@insdata, "Instance" );
    chomp $xmlstr;
    $xmlstr .= "<StorageList>";
    $xmlstr .= $self->createApplicationXML( \@Storageinsdata, "Storage" );
    chomp $xmlstr;
    $xmlstr .= "</StorageList>";
    $xmlstr .= "</DiscoveryData>";
	
EXIT_LABEL:

	return ($ret,$msg,$xmlstr);
}


sub discover_SystemType
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	my $SAPSys = $sapSys ;
	my $SystemType = "";

	my $doNotHideInstances = 1;
	my ($instances) = $self -> getInstances( $doNotHideInstances );
	# if ASCS and SCS both instances are present, then SystemType is AddIn
	if (grep (/^ASCS.*\($SAPSys\)/, @$instances) and 
		  grep (/^SCS.*\($SAPSys\)/, @$instances) and
		 (grep (/^DVEBMGS.*\($SAPSys\)/, @$instances) or grep (/^D.*\($SAPSys\)/, @$instances) or grep (/^SMDA.*\($SAPSys\)/, @$instances) ) ) {
		$SystemType = "AddIn";
	} elsif (grep (/^SCS.*\($SAPSys\)/, @$instances) and
			  (grep (/^DVEBMGS.*\($SAPSys\)/, @$instances) or  grep (/^D.*\($SAPSys\)/, @$instances) or grep (/^SMDA.*\($SAPSys\)/, @$instances) ) ) {
		# ASCS is embedded into DVEBMGS
		$SystemType = "AddIn-Embed-ms";
	} elsif (grep (/^ASCS.*\($SAPSys\)/, @$instances) and 
			 (grep (/^DVEBMGS.*\($SAPSys\)/, @$instances) or grep (/^D.*\($SAPSys\)/, @$instances) or grep (/^SMDA.*\($SAPSys\)/, @$instances) ) ) {
		$SystemType = "ABAP";
	} elsif (grep (/^DVEBMGS.*\($SAPSys\)/, @$instances) or grep (/^D.*\($SAPSys\)/, @$instances) or grep (/^SMDA.*\($SAPSys\)/, @$instances) ) {
		# ASCS is embedded into DVEBMGS
		$SystemType = "ABAP-Embed-ms";
	} elsif (grep (/^SCS.*\($SAPSys\)/, @$instances) and 
			 (grep (/^JC.*\($SAPSys\)/, @$instances) or grep (/^J.*\($SAPSys\)/, @$instances) or grep (/^SMDA.*\($SAPSys\)/, @$instances) )) {
		$SystemType = "Java";
	} elsif (grep (/^ASCS.*\($SAPSys\)/, @$instances) ) {
		$SystemType = "ABAP";
	} elsif (grep (/^SCS.*\($SAPSys\)/, @$instances) ) {
		$SystemType = "Java";
	} else {
		# Not a supported configuration
		return ( APPCONTROL_SAP_DISCOVER_SAPSYSTEMUNDETECTABLE,  "Unsupported InstType found.");
	}


	return ( CONST_SUCCESS, $SystemType );
}


sub discover_InstName
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	return ( CONST_SUCCESS, $ins);
}


sub discover_InstType
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	my $InstName = $ins ;
	my $InstType = ();
	my $SystemType = $self -> discover_SystemType($ins,$sapSys);

	if ( $InstName =~ /^AREP/ ) {
		$InstType = "ENQREP";
	} elsif ( $InstName =~ /^ERS/ ) {
		$InstType = "ENQREP";
	} elsif ( $InstName =~ /^REP/ ) {
		$InstType = "ENQREP";
	} elsif ( $InstName =~ /^ASCS/ ) {
		if ( $SystemType eq "AddIn" ) {
			$InstType = "ENQUEUE";
		} elsif ( $SystemType  eq "ABAP" ) {
			$InstType = "ENQUEUE";
		}
	} elsif ( $InstName =~ /^SCS/ ) {
		if ( $SystemType eq "AddIn" ) {
			$InstType = "ENQUEUE";
		} elsif ( $SystemType eq "Java" ) {
			$InstType = "ENQUEUE";
		}
	} elsif ( $InstName =~ /^DV/ ) {
		$InstType = "APPSERV";
	} elsif ( $InstName =~ /^D/ ) {
		$InstType = "APPSERV";
	} elsif ( $InstName =~ /^JC/ ) {
		$InstType = "APPSERV";
	} elsif ( $InstName =~ /^(JD|J)/ ) {
		$InstType = "APPSERV";
  } elsif ( $InstName =~ /^SMDA/ ) {
              $InstType = "SMDAgent";
	} else {
		Utils::vxlog("INFO", "Instance Name is not in SAP-compatible format");
		Utils::vxlog("INFO", "returning from getInstType");
		return ( APPCONTROL_SAP_DISCOVER_INVALIDINSTANCENAME,  "Instance Name is not in SAP-compatible format");
	}

	if (not defined $InstType) {
		Utils::vxlog("INFO", "Unable to discover SAP system type");
		return (APPCONTROL_SAP_DISCOVER_INVALIDINSTANCENAME, "Unable to discover SAP system type");
	}

	return ( CONST_SUCCESS, $InstType );
}

sub discover_EnvFile
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	my $SAPAdmin = $self -> discover_SAPAdmin($ins,$sapSys);
	my $EnvFile = "";
	my ($name, $passwd, $uid, $gid, $quota,
		$comment, $gcos, $dir, $shell) = getpwnam($SAPAdmin);

	my $SAPHome = $dir;

	# if SAPAdmin shell is cshell, then env file to be used is .login
	# otherwise use .profile (files are searched in SAPHome)
	if ( $shell =~ /csh/ ) {
		$EnvFile = "$SAPHome/.login";
	} else {
		$EnvFile = "$SAPHome/.profile";
	}

	return (CONST_SUCCESS, $EnvFile);
}

sub discover_ProcMon
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	my $ProcMonValues = ();

	$ProcMonValues -> { Java } { APPSERV }	= "jc";
	$ProcMonValues -> { Java } { ENQUEUE  }	= "en ms";
	$ProcMonValues -> { Java } { SMDAgent }	= "jc";
	$ProcMonValues -> { Java } { ENQREP }	= "er";

	$ProcMonValues -> { ABAP } { APPSERV }	= "dw";
	$ProcMonValues -> { ABAP } { ENQUEUE  }	= "en ms";
	$ProcMonValues -> { ABAP } { SMDAgent }	= "jc";
	$ProcMonValues -> { ABAP } { ENQREP }	= "er";

	$ProcMonValues -> { AddIn } { APPSERV }   = "dw jstart";
	$ProcMonValues -> { AddIn } { ENQUEUE  }   = "en ms";
	$ProcMonValues -> { AddIn } { SMDAgent }  = "jc";
	$ProcMonValues -> { AddIn } { ENQREP }  = "er";

	$ProcMonValues -> { 'AddIn-Embed-ms' } { APPSERV }	 = "dw";
	$ProcMonValues -> { 'AddIn-Embed-ms' } { ENQUEUE }	= "en ms";
	$ProcMonValues -> { 'AddIn-Embed-ms' } { SMDAgent }  = "jc";

	$ProcMonValues -> { 'ABAP-Embed-ms' } { APPSERV }	= "dw";
	$ProcMonValues -> { 'ABAP-Embed-ms' } { ENQUEUE	}	= "en ms";
	$ProcMonValues -> { 'ABAP-Embed-ms' } { SMDAgent }  = "jc";

	my $SystemType = $self -> discover_SystemType($ins,$sapSys);
	my $InstType = $self -> discover_InstType($ins,$sapSys);

	my $ProcMon = $ProcMonValues -> { $SystemType } { $InstType };

	return (CONST_SUCCESS, $ProcMon);
}

sub discover_SAPSID
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	return ( CONST_SUCCESS, $sapSys );
}

sub discover_SAPHome
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	my $SAPSID = $sapSys ;
	my $SAPInst = $ins ;

	my $SAPHome = "/usr/sap/$SAPSID/$SAPInst/exe";
	return (CONST_SUCCESS, $SAPHome);
}

sub discover_SAPAdmin
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	my $SAPAdmin = $sapSys ;

	$SAPAdmin =~ tr/A-Z/a-z/;
	$SAPAdmin .= "adm";

	return (CONST_SUCCESS, $SAPAdmin);
}

sub discover_InstProfile
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	# try to discover InstProfile
	my $SAPSID = $sapSys ;
	my $SAPInst = $ins  ;

	my ($CmdRet, $instProfile) = Utils::runCmd("ls -1 /sapmnt/$SAPSID/profile/${SAPSID}_${SAPInst}_* 2>&1");
	if ($CmdRet != CONST_SUCCESS) {
		die "Error while discovering instProfile";
	}

	my @lines = split(/\n/, $instProfile);

	if ( defined $lines[0] and $lines[0] ne "" ) {
		return (CONST_SUCCESS, Utils::strTrim($lines[0]));
	}

	return (CONST_FAILURE, "");
}

sub discover_IsLocalInst
{
	my $self = shift;
	my $ins = shift;
	my $sapSys = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called";
	Utils::vxlog("INFO", $logStr);

	# try to discover InstProfile
	my $SAPSID = $sapSys ;
	my $SAPInst = $ins  ;

	my $ExeDir = "/usr/sap/$SAPSID/$SAPInst/exe";

	if (not -d $ExeDir ) {

		my $logStr =  "The instance is not installed on this system.";
		Utils::vxlog("INFO", $logStr);

		return (CONST_SUCCESS, "The instance is not installed on this system.");
	}

	return (CONST_SUCCESS, "Yes");
}

#---------------------------------------------
# Overriding the createCommitSGDetails method
#---------------------------------------------
sub createCommitSGDetails
{
	my $self = shift;
	my $params = shift;

	my $sMyName = ( caller 0 )[3];
	#-------------------
	# Announce ourselves
	#-------------------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	# Initialize return parameters

	my ( $ret, $msg, $xml, $insxml, $insResName) = (CONST_SUCCESS, "", "", "", "");
	my ( $hrInsResXMLs, $hrInsResNames ) = ();

	my $SGName = "";
	
	my $clusDiscoveryInputRef   = ();
	$clusDiscoveryInputRef->{Resources}{Types}          = [ "SAPWebAS71" ] ;
	my $instExistingList = ();
	my $SGDependency = ();
	my $instSidResourceNameHash = ();
	($instExistingList,$SGDependency,$instSidResourceNameHash) = $self->getExistingConfigurations();
	Utils::vxlog("INFO", "in $sMyName Printing details ...SGDependency");
	Utils::vxlog("INFO", Dumper( $SGDependency ));
	Utils::vxlog("INFO", "in $sMyName Printing details ...instExistingList");
	Utils::vxlog("INFO", Dumper( $instExistingList ));

	my $tree = Utils::convertXMLTextToHash( $params );

	# Debug - check
	Utils::vxlog("INFO", Dumper( $tree ));

	$xml = "<xml>";


		my @inst = ();
	if ( defined ($tree->{Instance})) {

		@inst = Utils::convertHashValuesToArray( $tree->{Instance} );

		foreach my $ins (@inst) {
			my $Name = $ins -> { Name };
			my $SAPSID	= $ins -> { SAPSID } { Value };
			my $InstName = $ins -> { InstName } { Value };
		        $insResName = "SAP_${SAPSID}-${Name}_res" ;
                        Utils::vxlog("INFO", "Resource Name:$insResName"); 
                        my $SidInstanceKey = $InstName . " ($SAPSID)";
                        $instSidResourceNameHash-> {$SidInstanceKey } = $insResName ;
		}

		foreach my $ins (@inst) {
			my $enqResName = "";
			my $Name = $ins -> { Name };
			my $SAPSID	= $ins -> { SAPSID } { Value };
			my $InstName = $ins -> { InstName } { Value };
			my $InstType = $ins -> { InstType } { Value };
			my $SAPAdmin = $ins -> { SAPAdmin } { Value };
			my $SAPHome  = $ins -> { SAPHome  } { Value };
			my $ProcMon  = $ins -> { ProcMon  } { Value };
			my $EnvFile  = $ins -> { EnvFile  } { Value };
			my $InstProfile = $ins -> { InstProfile  } { Value };
			my $SidInstanceKey = $InstName . " ($SAPSID)";
			$insResName = $instSidResourceNameHash-> {$SidInstanceKey };

                        Utils::vxlog("INFO", "Resource Name:$insResName"); 


		      if (( defined ($ins->{MountPointList})) and
				( defined ($ins->{MountPointList}{MountPoint})) ) {
        				my @sMountPoints = Utils::convertHashValuesToArray ($ins ->{MountPointList}{MountPoint});
        				foreach my $sMountPoint (@sMountPoints) {
							$self->setMountDependency($Name, $sMountPoint->{Value});
        				}
      			}

			Utils::vxlog("INFO", "Adding resource for [$Name]");
			$insxml = "<Resource>";
			$insxml .= "<Name>$insResName</Name>";
			$insxml .= "<Type>SAPWebAS71</Type>";

			$insxml .= "<Attributes>";
			$insxml .= "<InstType><global>$InstType</global></InstType>";
			$insxml .= "<EnvFile><global>$EnvFile</global></EnvFile>";
			$insxml .= "<ProcMon><global>$ProcMon</global></ProcMon>";
			$insxml .= "<InstProfile><global>$InstProfile</global></InstProfile>";
			$insxml .= "<SAPAdmin><global>$SAPAdmin</global></SAPAdmin>";

			if ( $InstType eq "ENQREP" ) {
				$enqResName = $self->getEnqResName($Name,$SAPSID,$InstProfile,\@inst,$instExistingList,$instSidResourceNameHash);
				 if ( $enqResName ne "" ) {
					$insxml .= "<EnqSrvResName><global>$enqResName</global></EnqSrvResName>";
				 }
			}

			$insxml .= "</Attributes>";
			$insxml .= "</Resource>";
		        $hrInsResXMLs -> { $Name } = $insxml;
                        $hrInsResNames -> { $Name } = $insResName;
		        $self->addInstance($Name);

		}
		Utils::vxlog("INFO","All the instances are added.");

	}else {
                $ret = APPCONTROL_SAP_DISCOVER_INVALIDINSTANCENAME;
                $msg = "Instance tag is missing in commit input XML.";
                goto CommitEND;
        }

	my $tempXml = ();

	my $sSystemNICSGxml = ""; 
	($ret, $sSystemNICSGxml) = $self->getNetworkXML($tree);
        $xml .= $sSystemNICSGxml; 

	my $sgXML = $self->createSGXML ($tree, $hrInsResXMLs, $hrInsResNames, "SAP");

        chomp $sgXML;	
	$tempXml = "<xml>";
  	$tempXml .= $sgXML;
	$tempXml .= "</xml>";

	Utils::vxlog("INFO","Printing XML before adding PreOnline attribute");
	Utils::vxlog("INFO","$tempXml");

	my $SGtree = Utils::convertXMLTextToHash( $tempXml );
	Utils::vxlog("INFO", Dumper( $SGtree ));
	
	my @SGinstances = ();
	if ( defined ($SGtree->{ServiceGroup})) {
	    @SGinstances = Utils::convertHashValuesToArray( $SGtree->{ServiceGroup} );
	    foreach my $ins (@SGinstances) {
		my $SGName = $ins -> { Name };
		my $Resources = $ins -> { Resources };
		my @ResInstances = Utils::convertHashValuesToArray( $Resources -> { Resource });
		
		my $countOfSAPResources = 0 ;
		my $tempSGName = "SAP";
		my $tempDisplayName = "SAP Instances:";
		

                foreach my $insRes (@ResInstances) {
			my $ResName = $insRes -> { Name };
			my $ResType = $insRes -> { Type };
			if($ResType eq "SAPWebAS71" )
			{
				my $SAPSIDVal = "";
				my $InstNameVal = "";
				$countOfSAPResources++;
				if ( $ResName =~ m/SAP_([a-zA-Z0-9]+)-([a-zA-Z0-9]+)_res*/ ) {
					$SAPSIDVal = $1;
					$InstNameVal = $2;
				}
				$tempSGName .= "_".${SAPSIDVal}."-".${InstNameVal};	
				if($countOfSAPResources > 1 )
				{
					$tempDisplayName .= ",";
				}
				$tempDisplayName .= $InstNameVal;
				
			}
		}
		if($countOfSAPResources > 0 )
		{
			$tempSGName .= "_SG";
			$tempSGName =  $self->GetUniqueGroupName("$tempSGName"); 
			Utils::vxlog("INFO", "Group Name:$tempSGName"); 

			$SGName = $tempSGName;
			my $SGAttributes = $ins -> { Attributes };
			my $globalNameTag = ();
			$globalNameTag -> {"global"  } = "$SGName" ;
			$ins->  {"Name" } = $SGName  ;

			my $globalUserAssocTag = ();
			my $xmlUserAssocTag ="<xml>";
			$xmlUserAssocTag .= "<global>Type</global>";
			$xmlUserAssocTag .= "<global>SAP</global>";
			$xmlUserAssocTag .= "<global>Name</global>";
			$xmlUserAssocTag .= "<global>${tempDisplayName}</global>";
			$xmlUserAssocTag .= "</xml>";
			
			$globalUserAssocTag = Utils::convertXMLTextToHash( $xmlUserAssocTag );
			$SGAttributes->  {"UserAssoc" } = $globalUserAssocTag  ;

		}

                foreach my $insRes (@ResInstances) {
			my $ResName = $insRes -> { Name };
			my $ResType = $insRes -> { Type };
			if($ResType eq "SAPWebAS71" )
			{
				my $Attributes = $insRes -> { Attributes };
				if ( defined ($Attributes->{InstType})){

					my $InstType = $Attributes->{InstType} ;
					my $InstTypeVal = $InstType->{global}; 
					my $InstProfile = $Attributes->{InstProfile} ;
					my $InstProfileVal = $InstProfile->{global}; 

				
					my $SAPSIDVal = "";
					my $InstNameVal = "";
			                if ( $ResName =~ m/SAP_([a-zA-Z0-9]+)-([a-zA-Z0-9]+)_res*/ ) {
						$SAPSIDVal = $1;
						$InstNameVal = $2;
					}
	
					Utils::vxlog("INFO","$SGName -  $ResName - $InstTypeVal - $InstNameVal - $SAPSIDVal");
					my $SidInstanceKey = $InstNameVal . " ($SAPSIDVal)";
					$SGDependency-> {$SidInstanceKey } = $SGName ;


					if($InstTypeVal eq "ENQREP" ) {
						my $SGAttributes = $ins -> { Attributes };
						my $globalPreOnlineTag = ();
						$globalPreOnlineTag -> {"global"  } = "1" ;
						$SGAttributes->  {"PreOnline" } = $globalPreOnlineTag  ;
						my $globalTriggerPathTag = ();
						$globalTriggerPathTag -> {"global"  } = ".sapwebas71" ;
						$SGAttributes->  {"TriggerPath" } = $globalTriggerPathTag  ;
					}
					if($InstTypeVal eq "ENQUEUE" ) {
						if( CONST_SUCCESS == $self->isReplicationEnabled($InstNameVal,$InstProfileVal) ) 
						{
							my $SGAttributes = $ins -> { Attributes };
							my $globalPreOnlineTag = ();
							$globalPreOnlineTag -> {"global"  } = "1" ;
							$SGAttributes->  {"PreOnline" } = $globalPreOnlineTag  ;
							my $globalTriggerPathTag = ();
							$globalTriggerPathTag -> {"global"  } = ".sapwebas71" ;
							$SGAttributes->  {"TriggerPath" } = $globalTriggerPathTag  ;
						}
					}
				}
			}
		}
	    }
	}


	Utils::vxlog("INFO", Dumper( $SGtree ));
	our %XMLOUT_SIMPLE_OPTIONS = (
	    ContentKey => 'text',
	    KeepRoot   => 1,
	    KeyAttr    => []
	);	
	my $xs = new XML::Simple ();
	my $SGtreeXml = $xs->XMLout($SGtree,%XMLOUT_SIMPLE_OPTIONS);

	Utils::vxlog("INFO","Printing XML after adding PreOnline attribute");
	Utils::vxlog("INFO","$SGtreeXml");
	$xml .= $SGtreeXml ;

	my $ServiceGroupDependencyXml = $self->getSGDependency ($SGDependency);
	Utils::vxlog("INFO","ServiceGroup Dependency is -> [$ServiceGroupDependencyXml]");
	$xml .= $ServiceGroupDependencyXml ;

	($ret, $msg) = $self ->checkAndAddApplicationResTypes();

CommitEND:
	$xml .= "</xml>";
	Utils::vxlog("INFO", $xml);
	my $retRef = ();

	$retRef -> { ret } = $ret;
	$retRef -> { msg } = $msg;
	$retRef -> { xml } = $xml;

	return $retRef;
}

sub getSGDependency 
{
        my $self = shift;
        my $SGDependency = shift;

        my $sMyName = ( caller 0 )[3];
        #---------
        # Announce
        #---------
        my $logStr = "Subroutine $sMyName is called.";
        Utils::vxlog("INFO", $logStr);
	my $ChildIns = "";
	my $ChildSapSys = "";
	my $parentChildDependency = ();
	my $ServiceGroupDependencyXml = "<ServiceGroupDependency>";	

	foreach my $sKey ( keys %{$SGDependency} ) {
		
		
                if ( $sKey =~ m/([a-zA-Z0-9]+) \((.*)\)/ ) {
                        $ChildIns = $1;
                        $ChildSapSys = $2;
                }
		Utils::vxlog("INFO", "Child Instance Name  - [$ChildIns]");
		Utils::vxlog("INFO", "Child SAP SID - [$ChildSapSys]");



		if( $ChildIns =~ /^ASCS./i) {
			foreach my $sParentKey ( keys %{$SGDependency} ) {
				my $ParentName = "";
				my $ParentSAPSID = "";
				if ( $sParentKey =~ m/([a-zA-Z0-9]+) \((.*)\)/ ) {
					$ParentName = $1;
					$ParentSAPSID = $2;

				}
				Utils::vxlog("INFO", "Parent Instance Name  - [$ParentName]");
				Utils::vxlog("INFO", "Parent SAP SID - [$ParentSAPSID]");

				if( ($ParentName =~ /^DVEBMGS.|^D./i) && ($ParentSAPSID eq  $ChildSapSys)) { 
					my $parentGrp = $SGDependency->{ $sParentKey };
					my $childGrp  = $SGDependency->{ $sKey };
					
					if( $parentGrp ne $childGrp  ) {
						if( not grep /$parentGrp/, keys %{$parentChildDependency} )
						{
							push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
						}
						else
						{
							my @arrTempChildgrp = Utils::convertHashValuesToArray( $parentChildDependency->{$parentGrp}) ;
						
							if( not grep /$childGrp/, @arrTempChildgrp)
							{
								push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
							}

						}
					}
				}
			}

		}
		elsif( $ChildIns =~ /^SCS./i) {
			foreach my $sParentKey ( keys %{$SGDependency} ) {
				my $ParentName = "";
				my $ParentSAPSID = "";
				if ( $sParentKey =~ m/([a-zA-Z0-9]+) \((.*)\)/ ) {
					$ParentName = $1;
					$ParentSAPSID = $2;
				}
				Utils::vxlog("INFO", "Parent Instance Name  - [$ParentName]");
				Utils::vxlog("INFO", "Parent SAP SID - [$ParentSAPSID]");
				if( ($ParentName =~ /^J./i) && ($ParentSAPSID eq  $ChildSapSys)) { 
					my $parentGrp = $SGDependency->{ $sParentKey };
					my $childGrp  = $SGDependency->{ $sKey };
					if(  $parentGrp ne $childGrp  ) {
						if( not grep /$parentGrp/, keys %{$parentChildDependency} )
						{
							push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
						}
						else
						{
							my @arrTempChildgrp = Utils::convertHashValuesToArray( $parentChildDependency->{$parentGrp} );
							if( not grep /$childGrp/, @arrTempChildgrp)
							{
								push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
							}
						}
					}
				}
			}

		}
		elsif( $ChildIns =~ /^DVEBMGS.|^D.|^J./i) {
			foreach my $sParentKey ( keys %{$SGDependency} ) {
				my $ParentName = "";
				my $ParentSAPSID = "";
				if ( $sParentKey =~ m/([a-zA-Z0-9]+) \((.*)\)/ ) {
					$ParentName = $1;
					$ParentSAPSID = $2;
				}
				Utils::vxlog("INFO", "Parent Instance Name  - [$ParentName]");
				Utils::vxlog("INFO", "Parent SAP SID - [$ParentSAPSID]");
				if( ($ParentName =~ /^SMDA./i) && ($ParentSAPSID eq  $ChildSapSys)) { 
					my $parentGrp = $SGDependency->{ $sParentKey };
					my $childGrp  = $SGDependency->{ $sKey };
					if( $parentGrp ne $childGrp  ) {
						if( not grep /$parentGrp/, keys %{$parentChildDependency} )
						{
							push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
						}
						else
						{
							my @arrTempChildgrp = Utils::convertHashValuesToArray( $parentChildDependency->{$parentGrp}) ;
							if( not grep /$childGrp/, @arrTempChildgrp)
							{
								push ( @{$parentChildDependency->{ $parentGrp }} , $childGrp);
							}
						}
					}
				}
			}

		}
	 }	
	 my $vcsobj = new VCS();
	 my $ret = CONST_SUCCESS;
	 my $msg = "";
	 my $existingParentChildDependency = ();


	 ($ret, $msg, my $clusRef )     = $vcsobj -> getClusterDetails();
	 if($ret ne CONST_SUCCESS)
	 {
		$ServiceGroupDependencyXml = "";
		goto GETSGDEPENDENCY;
	 }

	 my $grpDepRef      = $clusRef -> { GrpDepRef };
	 Utils::vxlog("INFO", "Printing existing group dependencies.");
	 Utils::vxlog("INFO", Dumper( $grpDepRef ));
	 if ( defined ($grpDepRef)) {
		  foreach my $grpDep ( @$grpDepRef ) {
			  my $parentGrp = ${grpDep} -> { Parent };
			  my $childGrp = ${grpDep} -> { Child };
			  push ( @{$existingParentChildDependency->{ $parentGrp }} , $childGrp);
                }
	 }

	 foreach my $parentGrp ( keys %{$parentChildDependency} ) 	
	 {	
		 my @temparrChildgrp =  Utils::convertHashValuesToArray( $existingParentChildDependency->{ $parentGrp } );
		 my @arrChildgrp = Utils::convertHashValuesToArray( $parentChildDependency->{$parentGrp} ) ;
		 if ( scalar (@arrChildgrp) > 0 ) {
			 foreach my $childGrp ( @arrChildgrp ) {
				 my $alreadyDepExists = 0 ;
				 if ( scalar (@temparrChildgrp) > 0 ) {
					 foreach my $tempChildgrp ( @temparrChildgrp ) {
						 if ( $tempChildgrp eq $childGrp ) 
						 {
							$alreadyDepExists = 1 ;
							last;
						 }
					 }
				 }
				 if($alreadyDepExists eq 0 )
				 {
					 $ServiceGroupDependencyXml .= "<Dependency>";
					 $ServiceGroupDependencyXml .= "<Parent>$parentGrp</Parent>";
					 $ServiceGroupDependencyXml .= "<Child>$childGrp</Child>";
					 $ServiceGroupDependencyXml .= "<Relationship>online global soft</Relationship>";
					 $ServiceGroupDependencyXml .= "</Dependency>";
				}
			 }
		}
	 }

	 if ( $ServiceGroupDependencyXml ne "<ServiceGroupDependency>"  ) {
	    $ServiceGroupDependencyXml .= "</ServiceGroupDependency>";
	 }
	 else
	 {
	    $ServiceGroupDependencyXml = "";
	 }
	 chomp $ServiceGroupDependencyXml ;
	 
GETSGDEPENDENCY:
	return $ServiceGroupDependencyXml;
}


sub getEnqResName
{
        my $self = shift;
        my $Name = shift;
        my $sapSys = shift;
        my $instProfile = shift;

	my $insdataRef = shift;
	my $instExistingList = shift ;
	my $instSidResourceNameHash = shift ;

        my $sMyName = ( caller 0 )[3];
	my $instanceNumber = "";
	my $enqResName = "";
        #---------
        # Announce
        #---------
        my $logStr = "Subroutine $sMyName is called.";
        Utils::vxlog("INFO", $logStr);

	my $CmdRet = "awk '/^SCSID/' $instProfile | awk '{print \$NF;}'";

	my ($ret, $lsOp)  = Utils::runCmd ($CmdRet);

	if ($ret != CONST_SUCCESS) {
		my $logStr = "[$CmdRet] could not be executed successfully. Error occured while finding Enqueue resource name.";
		Utils::vxlog("INFO", $logStr);
		return "" ;
	}

        my @lines = split(/\n/, $lsOp);

        if ( defined $lines[0] and $lines[0] ne "" ) {
                $instanceNumber = Utils::strTrim($lines[0]);
		my $logStr = "Instance number is [$instanceNumber] .";
		Utils::vxlog("INFO", $logStr);
	}
	my $enqFound = 0 ;
	if ( $instanceNumber ne "" ) {
		foreach my $data ( @$insdataRef ) {
			my $insName = $data -> { Name };
			my $insSAPSID = $data -> { SAPSID } { Value };

			my $logStr = "insName is [$insName] and SAPSID is [$insSAPSID].";
			Utils::vxlog("INFO", $logStr);
			if( ($insName =~ /${instanceNumber}$/i) && ($insSAPSID eq  $sapSys)) { 
				my $SidInstanceKey = $insName . " ($sapSys)";
				if(defined $instSidResourceNameHash-> {$SidInstanceKey } )
				{
				  $enqResName = $instSidResourceNameHash-> {$SidInstanceKey };
				}
				my $logStr = "enqResName is [$enqResName].";
				Utils::vxlog("INFO", $logStr);
				$enqFound = 1;
				last;
			}
		}
		if( ! $enqFound)
		{
			foreach my $data ( @$instExistingList ) {
				my $insName = $data -> { InstName } ;
				my $insSAPSID = $data -> { SAPSID } ;

				my $logStr = "insName is [$insName] and SAPSID is [$insSAPSID].";
				Utils::vxlog("INFO", $logStr);
				if( ($insName =~ /${instanceNumber}$/i) && ($insSAPSID eq  $sapSys)) { 
					my $SidInstanceKey = $insName . " ($sapSys)";
					if(defined $instSidResourceNameHash-> {$SidInstanceKey } )
					{
					  $enqResName = $instSidResourceNameHash-> {$SidInstanceKey };
					}
					my $logStr = "enqResName is [$enqResName].";
					Utils::vxlog("INFO", $logStr);
					last;
				}
			}
		}
	}

	
	return $enqResName;
}
sub checkAndAddApplicationResTypes
{
  my $self = shift;

  my $sMyName = ( caller 0 )[3];
  #---------
  # Announce
  #---------
  my $logStr = "Subroutine $sMyName is called.";
  Utils::vxlog("INFO", $logStr);

  my $ret = CONST_SUCCESS;
  my $msg = "";

  my $vcsobj = new VCS();
  $vcsobj -> makeConfigWritable();

  # check if passed type is present or not
  my $CmdRet = $vcsobj -> isResTypeAvailable ( $SAPResType );
  if ( $CmdRet != CONST_SUCCESS ) {
    $CmdRet = $self -> addResType( $SAPResType );
    if ($CmdRet != CONST_SUCCESS) {
      $ret = APPCONTROL_SAP_COMMIT_RESADD_FAILURE;
      $msg = "Failed to add [$SAPResType] type. Verify that SAPWebAS71 agent is installed.";
      goto EXIT_LABEL;
    }
  }

  $vcsobj -> dumpAndMakeConfigReadOnly();

EXIT_LABEL:
  return ($ret, $msg);
}

sub addResType
{
  my $self = shift;
  my $resType = shift;

  my $sMyName = ( caller 0 )[3];
  #---------
  # Announce
  #---------
  my $logStr = "Subroutine $sMyName is called.";
  Utils::vxlog("INFO", $logStr);

  my $file = "/etc/VRTSagents/ha/conf/SAPWebAS71/SAPWebAS71Types.cmd";
  my ($CmdRet, $op) = Utils::runCmd( $file );
  if ( $CmdRet != CONST_SUCCESS ) {
    return $CmdRet;
  }

  return CONST_SUCCESS;
} 
sub getExistingConfigurations()
{
  my $self = shift;

  my $sMyName = ( caller 0 )[3];
  #---------
  # Announce
  #---------
  my $logStr = "Subroutine $sMyName is called.";
  Utils::vxlog("INFO", $logStr);
  my $vcsobj = new VCS();
  my $instList = ();
  my $SGDependency = ();
  my $instSidResourceNameHash = ();
  my $ret = CONST_SUCCESS; 
  my $msg = ""; 

  
  ($ret, $msg, my $clusRef )     = $vcsobj -> getClusterDetails();
  goto GETEXISTINGCONFIGURATIONSEND if($ret ne CONST_SUCCESS);

  my $resRef      = $clusRef -> { ResRef };
  Utils::vxlog("INFO", "Printing existing resource information.");
  Utils::vxlog("INFO", Dumper( $resRef ));

  if ( defined ($resRef)) { 
	  foreach my $resName ( sort keys %$resRef ) 
          { 
		  if ($resRef->{$resName}{Type}{global} =~ /^SAPWebAS71$/) { 
			  my $instProfile = $resRef->{$resName}{InstProfile}{global};
			  my $SGName = $resRef->{$resName}{Group}{global};
			  my $ResourceName = $resRef->{$resName}{ResourceName}{global};
			  my $InstType = $resRef->{$resName}{InstType}{global};
			  my $instanceName = "";
			  if ( $instProfile =~ m/\/sapmnt\/([A-Z]+|[A-Z]+[0-9]+)\/profile\/([A-Z]+|[A-Z]+[0-9]+)\_([A-Z]+[0-9]+)_/ ) {
				  my $instanceName = $3;
				  my $SAPSID = $2;
				  my $ins = ();
				  $ins -> { SAPSID }  = $SAPSID;
				  $ins -> { InstName }  = $instanceName;
			          push (@$instList, $ins);
				  my $SidInstanceKey = $instanceName . " ($SAPSID)";
				  $SGDependency-> {$SidInstanceKey } = $SGName ;
				  $instSidResourceNameHash-> {$SidInstanceKey } = $ResourceName ;
			  }
                  } 
          } 
  } 
  Utils::vxlog("INFO", "Printing details ...SGDependency");
  Utils::vxlog("INFO", Dumper( $SGDependency ));
  Utils::vxlog("INFO", "Printing details ...instList");
  Utils::vxlog("INFO", Dumper( $instList ));

GETEXISTINGCONFIGURATIONSEND:

  return($instList,$SGDependency,$instSidResourceNameHash);

       
}
sub getIPParentInstance
{
  my $self = shift;
  my $resRef = shift;
  my $sapResName= shift;
  if (defined $resRef-> { $sapResName } {InstProfile} {global}) {
      my $entry =  $resRef->{ $sapResName }{InstProfile}{global};
      my $instanceName = "";
      if ( $entry =~ m/\/sapmnt\/([A-Z]+|[A-Z]+[0-9]+)\/profile\/([A-Z]+|[A-Z]+[0-9]+)\_([A-Z]+[0-9]+)_/ ) {
	$instanceName = $3;
    return $instanceName;
      }
  }
    return "";
}

# This function would be called discoverExistingNetworkConfig to populate
# the instances to be shown on Network page
sub getDiscoveredInstances
{
  my $self = shift;
  my $resRef = shift;
  my @resList = ();
  foreach my $resName (keys %$resRef) {
    if ($resRef->{$resName}{Type}{global} !~ /^${SAPResType}$/) {
      next;
    }
    if (defined $resRef->{ $resName }{InstProfile}) {
      my $entry =  $resRef->{ $resName }{InstProfile}{global};
      my $instanceName = "";
      if ( $entry =~ m/\/sapmnt\/([A-Z]+|[A-Z]+[0-9]+)\/profile\/([A-Z]+|[A-Z]+[0-9]+)\_([A-Z]+[0-9]+)_/ ) {
	$instanceName = $3;
      push (@resList, $instanceName);
      }

    }
  }
  return \@resList;

}

sub isReplicationEnabled
{

        my $sMyName = ( caller 0 )[3];
        my $logStr = "Subroutine $sMyName is called.";
        Utils::vxlog("INFO", $logStr);
        #---------
        # Announce
        #---------		
        my $self = shift;
	my $instanceName = shift;
        my $instProfile = shift;
        $logStr = "Subroutine $sMyName is called with $instanceName and $instProfile";
        Utils::vxlog("INFO", $logStr);
        my $Cmd = "awk '/^enque\\/server\\/replication/' $instProfile | awk '{print \$NF;}'";

        my ($ret, $lsOp)  = Utils::runCmd ($Cmd);

        if ($ret != CONST_SUCCESS) {
		$logStr = "[$Cmd] could not be successfully executed. Error occured while determining if Enqueue replication is enabled.";
                Utils::vxlog("INFO", $logStr);
                return  CONST_FAILURE ;
        }

        my @lines = split(/\n/, $lsOp);

        if ( defined $lines[0] and $lines[0] ne "" ) {
                my $ReplicationEnabled = Utils::strTrim($lines[0]);
		if( $ReplicationEnabled eq "true" )
		{
			$logStr = "Replication is enabled for Enqueue Server [$instanceName].";
			Utils::vxlog("INFO", $logStr);
			return  CONST_SUCCESS ;					
		}
		else
		{
			$logStr = "Replication is not enabled for Enqueue Server [$instanceName].";
			Utils::vxlog("INFO", $logStr);
			return  CONST_FAILURE ;					
		}
        }
	else
	{
		$logStr = "[$instProfile] file does not contain information whether replication is enabled for Enqueue Server [$instanceName].";
		Utils::vxlog("INFO", $logStr);
		return  CONST_FAILURE ;
	}

}
# Shallow discovery is called by validate from the node where wizard is running.
# shallow discovery happens on all the other nodes where wizard is not running.
sub shallowDiscover
{
	my $self = shift;
	my $params = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName is called.";
	Utils::vxlog("INFO", $logStr);

	my $CmdRet = CONST_SUCCESS;
	my $_ownerUID = "";
	my $ret = CONST_SUCCESS;
	my $msg = "";
	my $xml = "<xml></xml>";

	my $packageName = "" ;

	$ret = PlatUtils::isAcclibPackageInstalled("SAP");

	if ( $ret != PlatUtils::PACKAGE_INSTALLED ) {
		$msg = "Package [VRTSacclib] is not installed for SAP. Install the package and then retry.";
		Utils::vxlog("INFO", $msg);
		return (APPCONTROL_SAP_DISCOVER_ACCLIBNOTINSTALLED, $msg, "");
	}
	else
	{
		$logStr = "Package [VRTSacclib] is installed for SAP.";
		Utils::vxlog("INFO", $logStr);
	}  

	($ret,$packageName) = PlatUtils::isAgentPackageInstalled("SAP");

	if ( $ret != PlatUtils::PACKAGE_INSTALLED ) {
	$msg = "Package [$packageName] is not installed for SAP. Install the package and then retry.";
		Utils::vxlog("INFO", $msg);
		return (APPCONTROL_SAP_DISCOVER_SAPAGENTNOTINSTALLED, $msg, "");
	}
	else
	{
		$logStr = "Package [$packageName] is installed for SAP.";
		Utils::vxlog("INFO", $logStr);
	}
                                                                                                     

	return ($ret, $msg, $xml);
}

# This function is called to get description for service groups
# for displaying on UI
sub getSGDescription
{
  my $self = shift;
  my $sg_name = shift;
  my $resources = shift;
  my $msg = "Returning description for service group $sg_name : ";
  my $sg_desc = "";
  my $instances = "";
  
  foreach my $res (@$resources) {
    my $res_xml = $self -> getResourceXML ( $res );
    my $res_details = Utils::convertXMLTextToHash( $res_xml );

    if ( defined ($res_details->{Type}) && $res_details->{Type} eq "SAPWebAS71" &&
      defined $res_details->{Attributes}{InstProfile}{global} ) {
      my $instProfile = $res_details->{Attributes}{InstProfile}{global};
      my $instanceName = "";
      my $sidName  = "";
      if ( $instProfile =~ m/\/sapmnt\/([a-zA-Z0-9]+)\/profile\/([a-zA-Z0-9]+)\_([A-Z]+[0-9]+)_/ ) {
        $instanceName = $3;
        $sidName = $1;
      }

      $instances .= "SAP System - " . $sidName . " and SAP Instance - " . $instanceName . ", ";
    }
   }
   if ( $instances ne "" ) {
      $instances =~ s/(.*)\, /$1/;
      $sg_desc = "This service group contains resources that relate to the following SAP instance(s): " . $instances . ".";
   }
   $msg .= $sg_desc;
   Utils::vxlog("DEBUG", $msg);
   return $sg_desc;
}


1; # so that perl doen't complain!
