
#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Id: Utils.pm,v 1.24 2014/07/21 06:19:58 sgupta2 Exp $
#

package Utils;

use base qw(Exporter);
our @EXPORT = qw(
	runCmd
	strTrim
	vxlog
	parseArguments
	getDateTime
	getOSInfo
	loadAndCreateModuleObject
	convertXMLTextToHash
	convertXMLFileToHash
	setLogConfig
	backupFile
	acquire_lock
	release_lock
	 );

use strict;
use warnings;

use Config;
use Data::Dumper;
use POSIX;
use Fcntl qw(:DEFAULT :flock);
use File::Copy;
use Net::Domain qw(hostname);

use Constants;
use PlatUtils;
use Logger;
use CmdUtils;
use Encode;

# Log size is considered in MBs
my $logFileSize = 5;
my $logDir		= getLogDir();
my $logfileName = "healthview";

my $logger		   = ();
my $cmdutils  = ();


#---------------
# Misc routines
#---------------
sub strTrim
{
	my $str = shift;

	if (!defined $str) {
		return undef;
	}

	$str =~ s/\s+$//;
	$str =~ s/^\s+//;

	return $str;
}


sub parseArguments
{
	my $query = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my @paramNames = $query->param;
	my %params = ();

	foreach my $p ( @paramNames ) {
		my $val = $query->param( $p );
		$params { $p } = $val;
	}

	return \%params;
}

sub convertXMLTextToHash
{
	my $params = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $tree = {}; # empty hash;

	if ( !defined $params ) {
		goto END;;
	}

	$params = Encode::encode('UTF-8',$params); # 2836964: XMLin fails for some non english chars.
	my $tmpFileName = PlatUtils::dumpToTmpFile( $params );
	$tree = convertXMLFileToHash( $tmpFileName );
	unlink $tmpFileName;

END:
	return $tree;
}


sub convertXMLFileToHash
{
	my $tmpFileName = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $tree = {}; # empty hash;

	my $simpleParser = XML::Simple -> new();
	$XML::Simple::PREFERRED_PARSER='XML::Parser' if ($^O =~ m/win/i) ;
	
	eval
	{
		$tree = $simpleParser -> XMLin ( $tmpFileName, SuppressEmpty => '' );  
	};
  	 
	if($@)
	{    
		Utils::vxlog("ERROR",$@);
		# This is a fix for incident 3027238,3030185 on windows. 
		# XMLin fails to parse XML with PREFERRED_PARSER='XML::Parser'.
		if($^O =~ m/win/i) 
		{
			Utils::vxlog("INFO","Setting preferred parser to XML::SAX::PurePerl and then parse again");
			$XML::Simple::PREFERRED_PARSER='XML::SAX::PurePerl';
			$tree = $simpleParser -> XMLin ( $tmpFileName, SuppressEmpty => '' );
		}
	}

END:
	return $tree;
}


sub convertHashToXML
{
	my $href = shift;
	my $rootTag = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	$rootTag = "xml" if ( ! defined $rootTag or "$rootTag" eq "");

	my $xml = ""; # empty xml
	our %XMLOUT_SIMPLE_OPTIONS = (
		ContentKey => 'text',
		KeepRoot   => 1,
		KeyAttr    => [],
		NoAttr     => 1,
		RootName   => $rootTag,
	);

	my $simpleParser = XML::Simple -> new();
	$XML::Simple::PREFERRED_PARSER='XML::Parser' if ($^O =~ m/win/i) ;

	$xml = $simpleParser -> XMLout ( $href, %XMLOUT_SIMPLE_OPTIONS);

END:
	return $xml;
}


sub getCmdResponseXML
{
	my $ID = shift;
	my $ReturnCode = shift;
	my $Message = shift;
	my $xmlstr = shift;

	my $xml = "<Cmd>\n";
	$xml .= "<ID>$ID</ID>\n";
	$xml .= "<ReturnCode>" . $ReturnCode . "</ReturnCode>\n";
	$xml .= "<Message>"    . $Message    . "</Message>\n";
	$xml .= "$xmlstr\n";
	$xml .= "</Cmd>\n";

	return $xml;
}


sub parseCmdResponseXML
{
	my $respXML = shift;

	my $ret = CONST_SUCCESS;
	my $msg = "";
	my $dbg = "";

	my $retTree = Utils::convertXMLTextToHash($respXML);

	if ( $retTree and $retTree -> { ReturnCode } and
			( $retTree -> { ReturnCode } != CONST_SUCCESS ) ) {
		$ret = $retTree -> { ReturnCode };
		$msg = $retTree -> { Message };
		$dbg = $retTree -> { Debug } if ( $retTree -> { Debug } );
	}

	return ($ret, $msg, $dbg);
}


sub getOSInfo
{
	my $href = ();
	my $host = "";

	$href -> { OS		} = $Config { osname };
	$href -> { OSCAPTION	} = PlatUtils::getOSName();
	$href -> { ARCH		} = PlatUtils::getOSArch();
	$href -> { PADV		} = PlatUtils::getOSPADV();
	$href -> { BITS		} = PlatUtils::getOSBits();
	$href -> { DATE		} = strftime '%Y-%m-%d %H:%M:%S', localtime;
	$href -> { DOMAIN	} = PlatUtils::getDomainName();
	
	if ( $^O =~ m/win/i )
	{
		chomp($host = `hostname 2> NUL`);
	}
	else
	{
		$host = hostname();
	}
	$href -> { COMPUTERNAME } = $host;

	#
	#Incident 2829565
	#When OS is part of cluster & Cluster's system list is returned in CAPS.
	#UI has larger impact of changing to case-insensitive comparison, So for windows OS Name is returned always in CAPS.
	#
	$href -> { COMPUTERNAME		   } = uc $href -> { COMPUTERNAME } if ($^O =~ m/win/i) ;
	$href -> { VIRTUALIZATION_PLATFORM } = PlatUtils::getVirtualizationPlatform();
	return $href;
}


sub getAppObjFactory
{
	my $module = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	Utils::vxlog("INFO", "Loading module [$module]");
	require "$module.pm";
	import $module;

	my $App = new $module();
	return $App;
}


#-------------------------------
# Command / File Run subroutines
#-------------------------------
sub runCmd
{
	return $cmdutils -> runCmd ( @_ );
}


#-----------------------------
# Logging related subriutines
#-----------------------------
sub vxlog
{
	my ( $package, $filename, $line ) = caller;
	push (@_, $filename );
	push (@_, $line );
	return $logger -> vxlog(@_);
}

sub setDefaultLogConfig
{
	$logger = new Logger( $logfileName, $logDir, $logFileSize );
}

sub setLogConfig
{
	my $logFileL = shift;
	my $logFileSizeL = shift;

	$logger = new Logger( $logFileL, $logDir, $logFileSizeL );
}


#----------------------------
# locking premitives ...
#----------------------------
# to get lock on a file
#
# input :
#	1. lock file name
#	2. flag to indicate the lock type
#		  0 read lock
#		  1 write lock
#
# output :
#	1. returns file descriptor
#
sub acquire_lock
{
	my($arg, $ret_val, $file);
	$ret_val = 0;
	$file = $_[0];

	# if file not present, first create it.
	if ( not -f $file ) {
		open FILE, ">$file";
		close FILE;
	}

	if($_[1] eq 1)
	{
		open (FILE, ">$file"); #write lock
		flock( FILE, 2 );
	}
	else
	{
		open(FILE, "<$file"); #read lock
		flock( FILE, 1 );
	}

	return \*FILE;
}

#
# routine to release lock
#
# input : file descriptor
#
sub release_lock
{
	close($_[0]);
}


#---------------------------------------------
# Encrypt the password
#---------------------------------------------
sub encrypt_password
{
	my $passwd = shift;


	my $ret = CONST_SUCCESS;
	my $hVCSPaths = PlatUtils::getVCSPaths();
	my $vcsencrypt = $hVCSPaths -> { vcsencrypt };

	if ( $passwd ne "" )
	{
		my $cmd = "";
        if ( $^O =~ /win/i ) {
			$passwd =~ s/"/\"/g;
			$cmd = "$vcsencrypt -agent \"$passwd\"";
		} else {
			$passwd =~ s/'/'"'"'/g;
			$cmd = "$vcsencrypt -agent \'$passwd\'";
		}
		my ($CmdRet, $op) = runCmd($cmd, {doNotLogCmd => 1, doNotLogOutput => 0});
		if ($CmdRet != CONST_SUCCESS )
		{
			vxlog("ERROR", "Failed to encrypt the password.");
			$ret = CONST_FAILURE;
		}
		else
		{
			$passwd = $op;
		}
	}

	return ($ret, $passwd);
}


#-----------------------------
# Command utils ...
#-----------------------------
sub setCmdUtil
{
	$cmdutils = new CmdUtils ( $logger );
}


sub convertHashValuesToArray
{
	my @hashValueArray;
	my $inputHashValue = shift;
	if (ref $inputHashValue eq "ARRAY") 
	{
		push @hashValueArray, @$inputHashValue;
	}
	elsif (defined $inputHashValue)
	{
		push @hashValueArray, $inputHashValue;
	}
	else
	{
		Utils::vxlog("ERROR", "NULL hash value provided in getArrayFromHashValues().");
	}
	return @hashValueArray;
}


sub getUniqueArray
{
	my $inputArray = shift;
	my %arrayElemHash;
	@$inputArray = grep { !$arrayElemHash{$_}++ } @$inputArray;
 }

#------------------------------------
# Reading and saving guest config xml
#------------------------------------
sub readGuestConfig
{
        my $sMyName = ( caller 0 )[3];
        #---------
        # Announce
        #---------
        my $logStr = "Subroutine $sMyName called";
        vxlog("INFO", $logStr);

        my $href = ();

        my $GuestConfigFile = "";
        if ( $^O =~ /win/i ) {
                $GuestConfigFile = PlatUtils::getVCSHome() . "\\portal\\world\\GuestConfig.xml";
        } else {
                $GuestConfigFile = "/opt/VRTSvcs/portal/world/GuestConfig.xml";
        }

        $href = Utils::convertXMLFileToHash( $GuestConfigFile );

END:
        return $href;
}


sub backupFile
{
	my $file = shift;

	my $sMyName = ( caller 0 )[3];
	#---------
	# Announce
	#---------
	my $logStr = "Subroutine $sMyName called";
	Utils::vxlog("INFO", $logStr);

	my $ret = CONST_SUCCESS;
	my $op  = "";
	my $cmdRet = 0;

	if ( -e $file) {
		my $date = PlatUtils::getCurrentDate();
		my $newLocation = $file . "_" . $date;
		my $mvCmd = "mv $file $newLocation";

		($ret, $op, $cmdRet) = Utils::runCmd($mvCmd);

		if ( $ret != CONST_SUCCESS ) {
			Utils:vxlog ("ERROR", "Failed to move $file.");
		}
	}

	return ($ret, $op);
}


# A function that reads a PID from a file and checks whether that PID 
# is running or not
sub readAndValidateSavedPID
{
	my $filename = shift;

	open (PIDFILE, "<", $filename) or do {
		Utils::vxlog("ERROR", "Cannot write to file $filename ($!)");
		return CONST_FAILURE; # report failure
	};

	my $pid = <PIDFILE>;
	close (PIDFILE);
	chomp ($pid); # TODO: Do we really need to chomp?

	if ( $pid and $pid =~ /\d+/ ) {
		return $pid if kill (0, $pid);
	}
	return 0; # pid does not exist
}


# A function to kill a given PID
sub killPID
{
	my $pid = shift;

	if ( $pid and $pid =~ /\d+/ ) {
		if ( kill (0, $pid) ) { 
			# process is running, kill it
			kill (15, $pid); # send SIGTERM
				sleep (1);

			if ( kill (0, $pid) ) {  # test
				kill (9, $pid); # send SIGKILL
					sleep (1);

				if ( kill (0, $pid) ) {  # test
					# cannot kill $pid
					Utils::vxlog("ERROR", "Cannot kill process with PID $pid");
					return CONST_FAILURE;
				}
			}
		}
	}

	return CONST_SUCCESS; # pid successfully killed
}


# A function to perform some operation on each of the specified nodes
# Takes 3 inputs:
#   - a reference to a hash that contains the username and password for each of the nodes in the following format:
#       $href -> { <nodename> }{ Username };
#       $href -> { <nodename> }{ Password };
#   - name of the script to be called on the remote node
#   - string to be passed to "params" argument of the script
# Returns (success/failure, msg)
sub RemoteOperation
{
	my $hNewNodeInfo = shift;
	my $scriptName   = shift;
	my $params       = shift;

	my $subName = ( caller 0 )[3];
	Utils::vxlog("INFO", "Subroutine $subName called");

	my $ret = CONST_SUCCESS;
	my $fret = CONST_SUCCESS;
	my $msg = "";
	my $dbg = "";

	my $hVCSPaths = PlatUtils::getVCSPaths();
	my $xprtlc = $hVCSPaths -> { xprtlc };
	my $paramsArgs = "";

	if ( $params ) {
		$paramsArgs = "-d params=$params";
	}

	foreach my $node ( keys %$hNewNodeInfo ) {
		my $nodeName = $node;
		my $userName = $hNewNodeInfo -> { $node }{ Username };
		my $passwd   = $hNewNodeInfo -> { $node }{ Password };


		my $targetUrl = "https://$nodeName:5634/vcs/admin/$scriptName";
		my $args = "";
		if ( $^O =~ m/win/i ) {
			$passwd =~ s/"/\"/g;
			$args = "-u \"$userName\" -p \"$passwd\" $paramsArgs";
		} else {
			$passwd =~ s/'/'"'"'/g;
			$args = "-u \'$userName\' -p \'$passwd\' $paramsArgs";
		
		}
		my $proxyCmd = $xprtlc . " -l $targetUrl $args";

		($fret, $msg, $ret) = Utils::runCmd($proxyCmd, { doNotLogCmd => 1 });
		if ($fret != CONST_SUCCESS) {
			Utils::vxlog("ERROR", "Error occurred on $nodeName - [$msg]");
			last;
		}
		else {
			($ret, $msg, $dbg) = parseCmdResponseXML ( $msg );
			if ( $ret != CONST_SUCCESS ) {
				Utils::vxlog("ERROR", "Error occurred on $nodeName - $msg [$dbg]" );
				last;
			}
		}
	}

	return ( $ret, $msg );
}


#------------------------
# Main logic starts here
#------------------------
setDefaultLogConfig();
setCmdUtil();


#finally ...
1; # so that perl compiler doesn't complain!
