#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
package CPIC;
use strict;

sub install {
    my $cpic=shift;
    my ($phase,$rtn);

    $rtn=$cpic->preinstall();
    return 0 if (!$rtn);

    $phase=$cpic->setup_install_phase();
    $phase->execute_tasks();

    return $cpic->postinstall();
}

# Return 1, continue install.
# Return 0, skip whole install phase.
sub preinstall {
    my $cpic=shift;
    my ($rel,$pkg,$patch,$prod);

    # If prod not specified, skip the whole install phase.
    return 0 if (!$cpic->{prod});

    $cpic->preinstall_messages() unless (Cfg::opt(qw(hotfixupgrade patchupgrade upgrade)));

    # Calculate package and patch list for installation.
    $rel=$cpic->rel;
    $cpic->{installpkgs}=$cpic->verify_list('installpkgs');
    $cpic->{installpatches}=$cpic->verify_list('installpatches');
    $cpic->{installpatchpkgs}||=[];
    $rel->{batchinstallpkgs}=0;
    $rel->{batchinstallpatches}=0;

    for my $pkgi (@{$cpic->{installpkgs}}) {
        $pkg=$cpic->pkg($pkgi);
        if ($pkg->{batchinstall}) {
            $rel->{batchinstallpkgs}=1;
        }
    }

    for my $patchi (@{$cpic->{installpatches}}) {
        $patch = $cpic->patch($patchi);
        if ($patch->{batchinstall}) {
            $rel->{batchinstallpatches}=1;
        }
    }

    for my $sys (@{$cpic->{systems}}) {
        for my $pkgi (@{$sys->{installpkgslast}}) {
            push(@{$cpic->{installpkgslast}},$pkgi) unless EDRu::inarr($pkgi,@{$cpic->{installpkgslast}});
        }
    }
    #fix for e2567015
    $cpic->{installpkgslast}=$rel->adjust_pkg_sequence($cpic->{installpkgslast});

    $prod=$cpic->prod;
    $prod->preinstall_tasks if($prod->can('preinstall_tasks'));

    # Show install UI form
    $cpic->show_ui_form('installing');

    return 1;
}

sub setup_install_phase {
    my $cpic=shift;
    my ($phase,$phase_name,$phase_desc,$task,$task_name,%task_params,$prod,$pkg,$patch,$rel,$pdfr,$pdfrs,$id);

    $rel=$cpic->rel;
    $pdfr=Msg::get('pdfr');
    $pdfrs=Msg::get('pdfrs');

    $prod=$cpic->prod;
    $phase_name=$cpic->{install_phase_name} || 'CPIC_install';
    $phase_desc=$cpic->{install_phase_desc} || Msg::new("Installing $prod->{abbr}", 30, 1263, "$prod->{abbr}");
    $phase=Phase->new($phase_name);
    $phase->set_description($phase_desc);
    $phase->initialize_tasks();

    # Perform preinstall tasks
    $id=10;
    $task_name = "product_preinstall";
    %task_params=(
        'sequence_id' => $id,
        'description' => Msg::new("Performing $prod->{abbr} preinstall tasks", 30, 1264, "$prod->{abbr}"),
        'description_sys' => Msg::new("Performing $prod->{abbr} preinstall tasks on #{SYS}", 30, 1265, "$prod->{abbr}"),
        'if_per_system' => 1,
        'handler' => \&CPIC::prod_preinstall_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::prod_preinstall_errors,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    # install packages batchly
    $id+=100;
    if ($rel->{batchinstallpkgs}) {
        $task_name = "install_pkgs";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Installing $prod->{abbr} $pdfrs", 30, 1266, "$prod->{abbr}", "$pdfrs"),
            'description_sys' => Msg::new("Installing $prod->{abbr} $pdfrs on #{SYS}", 30, 1267, "$prod->{abbr}", "$pdfrs"),
            'if_per_system' => 1,
            'handler' => \&CPIC::pkgs_install_sys,
            'handler_object_arg' => $cpic,
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);

        # Skip the task on some systems for serial mode
        for my $sys (@{$cpic->{systems}}) {
            if (!$sys->{installpkgs}) {
                $task->set_skip_sys($sys,1);
            }
        }
    }

    # install packages
    $id+=100;
    for my $pkgi (@{$cpic->{installpkgs}}) {
        $pkg=$cpic->pkg($pkgi);
        next if ($pkg->{installpkgslast} || $pkg->{batchinstall});

        $task_name = "install_pkg_$pkgi";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Installing $pkg->{pkg} $pdfr", 30, 1266, "$pkg->{pkg}", "$pdfr"),
            'description_sys' => Msg::new("Installing $pkg->{pkg} $pdfr on #{SYS}", 30, 1267, "$pkg->{pkg}", "$pdfr"),
            'if_per_system' => 1,
            'handler' => \&CPIC::pkg_install_sys,
            'handler_object_arg' => $cpic,
            'handler_args' => [ $pkgi ],
            'rc_handler' => \&CPIC::pkg_install_rc_check_sys,
            'rc_handler_object_arg' => $cpic,
            'rc_handler_args' => [ $pkgi ],
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);

        # Skip the task on some systems for serial mode
        for my $sys (@{$cpic->{systems}}) {
            if (!EDRu::inarr($pkgi, @{$sys->{installpkgs}})) {
                $task->set_skip_sys($sys,1);
            }
        }
        $id++;
    }

    # install patches batchly
    $id+=100;
    # Only consider single Padv for now
    if ($rel->{batchinstallpatches}) {
        $task_name = "install_pkg_patches";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Installing $prod->{abbr} patches", 30, 1268, "$prod->{abbr}"),
            'description_sys' => Msg::new("Installing $prod->{abbr} patches on #{SYS}", 30, 1269, "$prod->{abbr}"),
            'if_per_system' => 1,
            'handler' => \&CPIC::patches_install_sys,
            'handler_object_arg' => $cpic,
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);

        # Skip the task on some systems for serial mode
        for my $sys (@{$cpic->{systems}}) {
            if (!$sys->{installpatches}) {
                $task->set_skip_sys($sys,1);
            }
        }
    }

    # install patches
    $id+=100;
    if ($#{$cpic->{padvs}}) {
        # multi-padv must display "Installing $pkg patches" and install
        # the appropriate patch per the padv of the system
        for my $pkgi (@{$cpic->{installpatchpkgs}}) {
            $pkg=$cpic->pkg($pkgi);
            next if ($pkg->{installpatcheslast});

            $task_name = "install_pkg_patch_$pkgi";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Installing $pkg->{pkg} patch", 30, 1270, "$pkg->{pkg}"),
                'description_sys' => Msg::new("Installing $pkg->{pkg} patch on #{SYS}", 30, 1271, "$pkg->{pkg}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::multipadv_patch_install_sys,
                'handler_object_arg' => $cpic,
                'handler_args' => [ $pkgi ],
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);

            # Skip the task on some systems for serial mode
            for my $sys (@{$cpic->{systems}}) {
                my $need_install = 0;
                for my $patchi (@{$pkg->patches_sys($sys)}) {
                    if (EDRu::inarr($patchi, @{$sys->{installpatches}})) {
                        $need_install = 1;
                        last;
                    }
                }
                $task->set_skip_sys($sys, 1) if ($need_install == 0);
            }
            $id++;
        }
    } else {
        # single-padv can explicitely display "Installing $patch patch"
        for my $patchi (@{$cpic->{installpatches}}) {
            $patch=$cpic->patch($patchi);
            $pkg=$patch->pkg($patch->{pkg});
            next if ($pkg->{installpatcheslast} || $patch->{batchinstall});

            $task_name = "install_patch_$patchi";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Installing $patch->{patch_vers} patch", 30, 1270, "$patch->{patch_vers}"),
                'description_sys' => Msg::new("Installing $patch->{patch_vers} patch on #{SYS}", 30, 1271, "$patch->{patch_vers}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::patch_install_sys,
                'handler_object_arg' => $cpic,
                'handler_args' => [ $patchi ],
                'rc_handler' => \&CPIC::patch_install_rc_check_sys,
                'rc_handler_object_arg' => $cpic,
                'rc_handler_args' => [ $patchi ],
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);

            # Skip the task on some systems for serial mode
            for my $sys (@{$cpic->{systems}}) {
                if (!EDRu::inarr($patchi, @{$sys->{installpatches}})) {
                    $task->set_skip_sys($sys,1);
                }
            }
            $id++;
        }
    }

    # install packages round II
    $id+=100;
    for my $pkgi (@{$cpic->{installpkgslast}}) {
        $pkg=$cpic->pkg($pkgi);
        next if ($pkg->{batchinstall});

        $task_name = "install_pkg_$pkgi";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Installing $pkg->{pkg} $pdfr", 30, 1266, "$pkg->{pkg}", "$pdfr"),
            'description_sys' => Msg::new("Installing $pkg->{pkg} $pdfr on #{SYS}", 30, 1267, "$pkg->{pkg}", "$pdfr"),
            'if_per_system' => 1,
            'handler' => \&CPIC::pkg_install_sys,
            'handler_object_arg' => $cpic,
            'handler_args' => [ $pkgi ],
            'rc_handler' => \&CPIC::pkg_install_rc_check_sys,
            'rc_handler_object_arg' => $cpic,
            'rc_handler_args' => [ $pkgi ],
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);

        # Skip the task on some systems for serial mode
        for my $sys (@{$cpic->{systems}}) {
            if (!EDRu::inarr($pkgi, @{$sys->{installpkgs}})) {
                $task->set_skip_sys($sys,1);
            }
        }
        $id++;
    }

    # install patches round II
    $id+=100;
    if ($#{$cpic->{padvs}}) {
        # multi-padv must display "Installing $pkg patches" and install
        # the appropriate patch per the padv of the system
        for my $pkgi (@{$cpic->{installpatchpkgslast}}) {
            $pkg=$cpic->pkg($pkgi);
            next if (!$pkg->{installpatcheslast} || $patch->{batchinstall});
            $task_name = "install_pkg_patch_$pkgi";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Installing $pkg->{pkg} patch", 30, 1270, "$pkg->{pkg}"),
                'description_sys' => Msg::new("Installing $pkg->{pkg} patch on #{SYS}", 30, 1271, "$pkg->{pkg}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::multipadv_patch_install_sys,
                'handler_object_arg' => $cpic,
                'handler_args' => [ $pkgi ],
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);

            # Skip the task on some systems for serial mode
            for my $sys (@{$cpic->{systems}}) {
                my $need_install = 0;
                for my $patchi (@{$pkg->patches_sys($sys)}) {
                    if (EDRu::inarr($patchi, @{$sys->{installpatches}})) {
                        $need_install = 1;
                        last;
                    }
                }
                $task->set_skip_sys($sys, 1) if ($need_install == 0);
            }
            $id++;
        }
    } else {
        for my $patchi (@{$cpic->{installpatcheslast}}) {
            $patch=$cpic->patch($patchi);
            $pkg=$patch->pkg($patch->{pkg});
            next if (!$pkg->{installpatcheslast} || $patch->{batchinstall});
            $task_name = "install_patch_$patchi";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Installing $patch->{patch_vers} patch", 30, 1270, "$patch->{patch_vers}"),
                'description_sys' => Msg::new("Installing $patch->{patch_vers} patch on #{SYS}", 30, 1271, "$patch->{patch_vers}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::patch_install_sys,
                'handler_object_arg' => $cpic,
                'handler_args' => [ $patchi ],
                'rc_handler' => \&CPIC::patch_install_rc_check_sys,
                'rc_handler_object_arg' => $cpic,
                'rc_handler_args' => [ $patchi ],
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);

            # Skip the task on some systems for serial mode
            for my $sys (@{$cpic->{systems}}) {
                if (!EDRu::inarr($patchi, @{$sys->{installpatches}})) {
                    $task->set_skip_sys($sys,1);
                }
            }
            $id++;
        }
    }

    # Perform postremove tasks
    $id+=100;
    $task_name = "product_postinstall";
    %task_params=(
        'sequence_id' => $id,
        'description' => Msg::new("Performing $prod->{abbr} postinstall tasks", 30, 1272, "$prod->{abbr}"),
        'description_sys' => Msg::new("Performing $prod->{abbr} postinstall tasks on #{SYS}", 30, 1273, "$prod->{abbr}"),
        'if_per_system' => 1,
        'handler' => \&CPIC::prod_postinstall_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::prod_postinstall_errors,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    return $phase;
}

sub postinstall {
    my $cpic=shift;
    my ($prod,$msg,$failmsg);

    $prod=$cpic->prod;
    $msg=Msg::new("$prod->{name} Install completed successfully", 30, 1274, "$prod->{name}");
    $failmsg=Msg::new("$prod->{name} Install did not complete successfully", 30, 1275, "$prod->{name}");
    $msg->display_completion($cpic->failures('installfailmsg'), $msg, $failmsg);

    $cpic->update_summary_install();
    $cpic->ask_install_continue();

    return 1;
}

# displays CLI messages
sub preinstall_messages {
    my $cpic = shift;
    my $prod=$cpic->prod;
    $prod->preinstall_messages();
    return;
}

# call /opt/VRTS/install/bin/TRAINVERS/add_install_scripts on $sys
sub add_install_scripts_sys {
    my ($cpic,$sys) = @_;
    my ($script,$rootpath);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $rootpath=Cfg::opt('rootpath');
    $script=$rootpath.$cpic->{installpath}.'/bin/'.$cpic->{release}.'/add_install_scripts';
    if ($sys->exists($script)) {
        $sys->cmd("CALLED_BY=CPI $script");
    } else {
        Msg::log("$script not found on $sys->{sys}");
    }
    return;
}

# this whole thing is probably going away, but saving to compare the two
# upper copy works, this is merge from final 5.1SP1
#
# copy the CPI libraries to /opt/VRTSperl/lib/site_perl/TRAINVERS
# 5.0 copies base libraries in place
# install scripts are copied in completion
sub copy_cpi_sys {
    my ($cpic,$sys) = @_;
    my ($rootpath,@allprods,$rel,$vxgettext_dir,$bmcdomaindir,$scripts,$installscript_prod,$installscript_name,$prods,$cpidir,$installscript,$prod,$localsys);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $cpic=Obj::cpic() unless (ref($cpic) eq 'CPIC');
    return 1 if (Cfg::opt('makeresponsefile'));
    return '' if ($cpic->{prod} && $sys->prod->{nocpionsys});
    $localsys=$cpic->localsys;
    $rel=$sys->rel;
    $rel->copy_cpi_sys($sys);  #for extras
    if ($cpic->{prod}) {
        $sys->prod->copy_cpi_sys($sys);  #for product binaries and other scripts
    }

    # condition for -copyinstallscripts option (product not defined yet)
    $prods = ($cpic->{prod}) ? [ $cpic->{prod} ] : $rel->{prods};

    $rootpath=Cfg::opt('rootpath');
    $cpidir=$rootpath.$cpic->{perlpath}.'/lib/site_perl/'.$cpic->{release};

    # copy CPI libraries
    $bmcdomaindir=$cpic->{mediapath}.'/scripts';
    for my $dir ('CPIC','CPIP','EDR') {
        next unless (-e "$bmcdomaindir/$dir");
        $sys->mkdir("$cpidir/$dir");
        $localsys->copy_to_sys($sys, "$bmcdomaindir/$dir/*", "$cpidir/$dir");
    }
    if ($rel->{type} eq "H") {
        my $msg = Msg::new("copy modules for $rel->{hotfixname} to $cpidir/CPIC", 30, 1276, "$rel->{hotfixname}", "$cpidir");
        $msg->log;
        $sys->mkdir("$cpidir/CPIC");            # okay if it already exists
        $localsys->copy_to_sys($sys, "$bmcdomaindir/*.pm", "$cpidir/CPIC");
    }

    # copy CPI message catalogs
    $vxgettext_dir = "$bmcdomaindir/bin";
    # Make sure this image includes messages and a way to decode them.
    if (-x $vxgettext_dir && -d "$bmcdomaindir/messages") {
        $localsys->copy_to_sys($sys, "$vxgettext_dir", "$cpidir");
        $sys->chmod("$cpidir/bin", "+x");
        $sys->rm("$cpidir/messages");
        $sys->mkdir("$cpidir/messages");
        $localsys->copy_to_sys($sys, "$bmcdomaindir/messages/*", "$cpidir/messages");
        # FIX: need to sort out unique directories per train as each family
        # will have its own bmcmap file
    }

    # copy CPI installer scripts
    $sys->mkdir("$rootpath$cpic->{installpath}");
    $sys->createfile("$rootpath$cpic->{installpath}/.cpi5");

    if (Cfg::opt('upgrade') && (defined $rel->{obsolete_scripts})){
        $scripts=join(' ', @{$rel->{obsolete_scripts}});
        unless ($scripts =~ /^\s*$/m) {
            $sys->cmd("cd $rootpath$cpic->{installpath}; _cmd_rmr $scripts");
        }
    }
    if ($rel->{pkgs}) {
        for my $prodname (@$prods) {
            $prod=$sys->prod($prodname);
            @allprods=($prodname);
            push(@allprods,@{$prod->{subprods}}) if ($cpic->{prod} && $prod->{subprods});
            for my $prodx (@allprods) {
                $prod=$sys->prod($prodx);

                # copy product binaries and other scripts
                $prod->copy_cpi_sys($sys);
                next if ($prod->{nocpionsys});

                # copy other modules
                for my $mod (@{$prod->{cpimodules}}) {
                    $localsys->copy_to_sys($sys, "$bmcdomaindir/$mod", $cpidir);
                }

                # copy task scripts
                $installscript_prod = $prod->{installscript_prod} || $prodx;
                $installscript_name = lc($prod->{installscript_name} || $prod->{prod});
                $installscript_name=~s/\s+//m;
                $installscript = "$rootpath$cpic->{installpath}/install" . $installscript_name;
                $cpic->create_task_script_sys($sys,'configure',$installscript_prod,$installscript) if (EDRu::inarr('configure',@{$rel->{args_task}}));
                $installscript = "$rootpath$cpic->{installpath}/uninstall" . $installscript_name;
                $cpic->create_task_script_sys($sys,'uninstall',$installscript_prod,$installscript) if (EDRu::inarr('uninstall',@{$rel->{args_task}}));
            }
        }
    } else {
        unless (Padv::plat($cpic->{padv}) eq "Linux") {
            $installscript_prod = '';
            $installscript_name = $rel->{hotfixname};
            $installscript = "$rootpath/opt/VRTS/install/uninstall" . $installscript_name;
            $cpic->create_task_script_sys($sys,'rollback',$installscript_prod,$installscript);
        }
    }

    $sys->chown_root("$rootpath$cpic->{installpath}");
    return 1;
}

sub create_task_script_sys {
    my ($cpic,$sys,$task,$prodname,$filename) = @_;
    my ($script,$rel,$edr,$padvs);
    $rel=$sys->rel;
    $edr=Obj::edr();
    $padvs=join(' ',@{$edr->{padvs}});
    $script=<<'SCRIPT';
#!/bin/sh

#
# Copyright (c) 2014 Symantec Corporation. All rights reserved.
# Symantec and the Symantec Logo are trademarks or registered trademarks
# of Symantec Corporation or its affiliates in the U.S. and other countries.
# Other names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be "commercial
# computer software" and "commercial computer software documentation"
# as defined in FAR Sections 12.212 and DFARS Section 227.7202.
#

PDIR=/opt/VRTSperl
SDIR=/opt/VRTSperl/lib/site_perl/RELEASE
PERLVERS=`$PDIR/bin/perl -v | grep This.is | awk '{print $4}' | sed 's/v//'`
PERLFOR=`$PDIR/bin/perl -v | grep This.is | awk '{print $7}'`
IOPT="-I$PDIR/lib/$PERLVERS/$PERLFOR -I$PDIR/lib/$PERLVERS -I$PDIR/lib/site_perl/$PERLVERS/$PERLFOR -I$PDIR/lib/site_perl/$PERLVERS -I$PDIR/lib/site_perl"
IOPT="-I$SDIR/EDR -I$SDIR/CPIC -I$SDIR/CPIP $IOPT"

if [ -z "$PDIR" ]; then
        echo "Cannot find perl to execute $0"
        exit 1
fi

eval 'exec $PDIR/bin/perl -x -S $IOPT $0 ${1+"$@"}'
        if 0;
#!$PDIR/bin/perl

use RELEASE;
use PATCHNAME;

CPIC::main( "RELEASE",
            "PADV",
            "",
            "TASK",
            "PROD");
SCRIPT

    $script=~s/RELEASE/$cpic->{release}/mg;
    if ($rel->{type} eq "H") {
        $script=~s/PATCHNAME/$rel->{hotfixname}/mg;
    } else {
        $script=~s/'use PATCHNAME;'//mg;
    }
    $script=~s/PADV/$padvs/mg;
    $script=~s/TASK/$task/mg;
    $script=~s/PROD/$prodname/mg;

    $sys->cmd("_cmd_rmr $filename");
    $sys->writefile($script, $filename);
    $sys->chmod($filename, '+x');
    return '' if(EDR::cmdexit());
    return 1;
}


# push the copy_pkg_patch_sys on the upper range of threads
sub copy_pkg_patch_subs {
    my ($cpic,$thr) = @_;
    my ($sys,$tid,$pkgi);
    $tid=$cpic->nsystems;
    for my $sys (@{$cpic->{systems}}) {
        $tid++;
        $thr->mq_add_sub($tid, 'CPIC', 'copy_pkg_patch_sys', $sys->{sys});
    }
    return;
}

# copy the packages and patches in parallel with running the installs
# each routine copies three packages and then waits for the install to
# remove the package it has completed installing before installing the next
sub copy_pkg_patch_sys {
    my ($cpic,$sys) = @_;
    my ($n,@copypp,$patch,$pkgi,$pkg,$patchi);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if (Cfg::opt('makeresponsefile'));;

    # create the list of everything we need to copy, packages
    for my $pkgi (@{$cpic->{installpkgs}}) {
        push(@copypp, $sys->pkg($pkgi))
            if (EDRu::inarr($pkgi, @{$sys->{installpkgs}}) && !(EDRu::inarr($pkgi, @{$sys->{installpkgslast}})));
    }

    # then patches
    if ($#{$cpic->{padvs}}) {
        for my $pkgi (@{$cpic->{installpatchpkgs}}) {
            $pkg=$sys->pkg($pkgi);
            for my $patchi (@{$pkg->patches_sys($sys)}) {
                push(@copypp, $sys->patch($patchi))
                    if (EDRu::inarr($patchi, @{$sys->{installpatches}}) && !(EDRu::inarr($patchi, @{$sys->{installpatcheslast}})));
            }
        }
    } else {
        for my $patchi (@{$cpic->{installpatches}}) {
            push(@copypp, $sys->patch($patchi))
                if (EDRu::inarr($patchi, @{$sys->{installpatches}}) && !(EDRu::inarr($patchi, @{$sys->{installpatcheslast}})));
        }
    }

    # packages for install round II should be last in copypp array
    for my $pkgi (@{$cpic->{installpkgslast}}) {
        push(@copypp, $sys->pkg($pkgi)) if (EDRu::inarr($pkgi, @{$sys->{installpkgslast}}));
    }

    # then patches in round II
    if ($#{$cpic->{padvs}}) {
        for my $pkgi (@{$cpic->{installpatchpkgslast}}) {
            $pkg=$sys->pkg($pkgi);
            for my $patchi (@{$pkg->patches_sys($sys)}) {
                push(@copypp, $sys->patch($patchi))
                    if (EDRu::inarr($patchi, @{$sys->{installpatcheslast}}));
            }
        }
    } else {
        for my $patchi (@{$cpic->{installpatcheslast}}) {
            push(@copypp, $sys->patch($patchi))
                if (EDRu::inarr($patchi, @{$sys->{installpatcheslast}}));
        }
    }
    # now do the copies, copying no more than 3 ahead of the install
    # may be a pkg or a patch, but methods and references are identical
    for my $n (0..$#copypp) {
        if ($n>=4) {
            $pkg=$copypp[$n-4];
            EDRu::wait_for_file_rm($pkg->copy_mark_sys($sys));
        }
        $pkg=$copypp[$n];
        $pkg->copy_sys($sys);
    }
    return '';
}

sub multipadv_patch_install_sys {
    my ($cpic,$sys,$pkgi) = @_;
    my $pkg=$sys->pkg($pkgi);
    my ($fails,$msg);
    $fails=0;
    for my $patch (@{$pkg->patches_sys($sys)}) {
        next if (!EDRu::inarr($patch, @{$sys->{installpatches}}));
        if ($cpic->patch_install_sys($sys, $patch)) {
            Msg::log("$patch installed successfully on $sys->{sys}");
            $sys->set_value('patchinstallpass', 'push', $patch);
            EDRu::appendfile($patch, EDR::tmpdir()."/patch_install_pass.$sys->{sys}")
                if (Cfg::opt('bait'));
        } else {
            $msg=Msg::new("$patch failed to install on $sys->{sys}", 30, 1277, "$patch", "$sys->{sys}");
            $msg->log;
            $sys->set_value('installfailmsg', 'push', $msg->{msg});
            $sys->set_value('patchinstallfail', 'push', $patch);
            EDRu::appendfile($patch, EDR::tmpdir()."/patch_install_fail.$sys->{sys}")
                if (Cfg::opt('bait'));
            $fails++;
        }
    }
    return 0 if ($fails);
    return 1;
}

# can't use $thr->mq_read_rtns() due to $pkg->$patch dereference
sub read_multipadv_patch_install_returns {
    my ($cpic,$thr,$pkgi) = @_;
    my ($sys,$ttid,$patch,$tid,$rtn,$pkg,$patchi);
    for my $sys (@{$cpic->{systems}}) {
        $tid++;
        $pkg=$sys->pkg($pkgi);
        for my $patchi (@{$pkg->patches_sys($sys)}) {
            $patch=$sys->patch($patchi);
            $ttid=(($tid-1)%$thr->{nthreads})+1;
            $rtn=shift(@{$thr->{$ttid}{rtns}});
            next unless (defined $rtn);
            if ($rtn) {
                patch_install_passed_sys($sys, $patch);
            } else {
                patch_install_failed_sys($sys, $patch);
            }
        }
    }
    return;
}

sub pkgs_install_sys {
    my ($cpic,$sys)= @_;
    my ($pkgs,$pkgis,$pkg,$pkg_file,$pkg_file_r,$tmpdir,$localsys);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');

    $pkgs='';
    $pkgis='';
    $pkg_file='';
    for my $pkgi (@{$cpic->{installpkgs}}) {
        next if (!EDRu::inarr($pkgi, @{$sys->{installpkgs}}));
        $pkg=$sys->pkg($pkgi);
        next if ($pkg->{installpkgslast} || !$pkg->{batchinstall});
        $pkg->preinstall_sys($sys);
        $pkgs.=" $pkg->{pkg}";
        $pkgis.=" $pkgi";
        $pkg_file||=$pkg->{file};
    }
    for my $pkgi (@{$sys->{installpkgslast}}) {
        $pkg=$sys->pkg($pkgi);
        next if (!$pkg->{batchinstall});
        $pkg->preinstall_sys($sys);
        $pkgs.=" $pkg->{pkg}";
        $pkgis.=" $pkgi";
        $pkg_file||=$pkg->{file};
    }
    return 1 unless ($pkgs);

    unless (Cfg::opt('makeresponsefile')) {
        $localsys=$cpic->localsys;
        if (!$sys->{islocal} && $pkg_file) {
            $tmpdir=EDR::tmpdir();
            $pkg_file_r="$tmpdir/" . EDRu::basename($pkg_file);
            if (!$sys->exists($pkg_file_r)) {
                $localsys->copy_to_sys($sys,$pkg_file,$tmpdir);
            }
            $pkg_file=$pkg_file_r;
        }
        $sys->install_pkgs($pkg_file,$pkgs);

        my $iof=EDRu::outputfile('install', $sys->{sys}, "pkgs");
        $sys->copy_to_sys($localsys,$iof) unless ($sys->{islocal});
        for my $pkgi (@{$cpic->{installpkgs}}) {
            next if (!EDRu::inarr($pkgi, @{$sys->{installpkgs}}));
            $pkg=$sys->pkg($pkgi);
            next if ($pkg->{installpkgslast} || !$pkg->{batchinstall});
            $pkg->postinstall_sys($sys);
        }
        for my $pkgi (@{$sys->{installpkgslast}}) {
            $pkg=$sys->pkg($pkgi);
            next if (!$pkg->{batchinstall});
            $pkg->postinstall_sys($sys);
        }
    }
    return 1 if ($cpic->pkgs_install_verify_sys($sys,$pkgis));
    return 0;
}

sub pkgs_install_verify_sys {
    my ($cpic,$sys,$pkgis)= @_;
    my ($fails,$pkg,$pdfr,$msg);
    $pkgis=~s/^\s*//;
    $pkgis=~s/\s*$//;
    $fails=0;
    $pdfr=Msg::get('pdfr');
    for my $pkgi (split(/\s+/,$pkgis)) {
        $pkg=$sys->pkg($pkgi);
        if (Cfg::opt('makeresponsefile') || $pkg->install_success_sys($sys)) {
            $msg=Msg::new("$pkg->{pkg} $pdfr installed successfully on $sys->{sys}", 30, 1278, "$pkg->{pkg}", "$pdfr", "$sys->{sys}");
            $msg->log;
            $sys->set_value('pkginstallpass', 'push', $pkg->{pkg});
            $sys->set_value("pkgvers,$pkg->{pkg}",$pkg->{vers});
            EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_install_pass.$sys->{sys}")
                if (Cfg::opt('bait'));
        } else {
            $msg=Msg::new("$pkg->{pkg} $pdfr failed to install on $sys->{sys}", 30, 1279, "$pkg->{pkg}", "$pdfr", "$sys->{sys}");
            $msg->log;
            $sys->set_value('installfailmsg', 'push', $msg->{msg});
            $sys->set_value('pkginstallfail', 'push', $pkg->{pkg});
            EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_install_fail.$sys->{sys}")
                if (Cfg::opt('bait'));
            $fails++;
        }
    }
    return 0 if ($fails);
    return 1;
}

sub patches_install_sys {
    my ($cpic,$sys)= @_;
    my ($patches,$patchis,$patch,$pkg,$patch_file,$patch_file_r,$tmpdir,$localsys);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');

    $patches='';
    $patchis='';
    $patch_file='';
    for my $patchi (@{$cpic->{installpatches}}) {
        next if (!EDRu::inarr($patchi, @{$sys->{installpatches}}));
        $patch=$sys->patch($patchi);
        $pkg=$patch->pkg($patch->{pkg});
        next if ($pkg->{installpatcheslast} || !$patch->{batchinstall});
        #e3389127:remove driver before install patch
        $patch->preinstall_sys($sys);
        $patches.=" $patch->{patchname}";
        $patchis.=" $patchi";
        $patch_file||=$patch->{file};
    }
    for my $patchi (@{$sys->{installpatcheslast}}) {
        $patch=$sys->patch($patchi);
        next if (!$patch->{batchinstall});
        #e3389127:remove driver before install patch
        $patch->preinstall_sys($sys);
        $patches.=" $patch->{patchname}";
        $patchis.=" $patchi";
        $patch_file||=$patch->{file};
    }
    return 1 unless ($patches);

    unless (Cfg::opt('makeresponsefile')) {
        $localsys=$cpic->localsys;
        if (!$sys->{islocal} && $patch_file) {
            $tmpdir=EDR::tmpdir();
            $patch_file_r="$tmpdir/" . EDRu::basename($patch_file);
            if (!$sys->exists($patch_file_r)) {
                $localsys->copy_to_sys($sys,$patch_file,$tmpdir);
            }
            $patch_file=$patch_file_r;
        }
        # Since on Solaris 11, patches and pkgs are using exact same mechanism
        $sys->install_patches($patch_file,$patches);

        my $iof=EDRu::outputfile('install', $sys->{sys}, "patches");
        $sys->copy_to_sys($localsys,$iof) unless ($sys->{islocal});
    }
    for my $patchi (@{$cpic->{installpatches}}) {
        next if (!EDRu::inarr($patchi, @{$sys->{installpatches}}));
        $patch=$sys->patch($patchi);
        $pkg=$patch->pkg($patch->{pkg});
        next if ($pkg->{installpatcheslast} || !$patch->{batchinstall});
        $patch->postinstall_sys($sys);
    }
    for my $patchi (@{$sys->{installpatcheslast}}) {
        $patch=$sys->patch($patchi);
        next if (!$patch->{batchinstall});
        $patch->postinstall_sys($sys);
    }

    return 1 if ($cpic->patches_install_verify_sys($sys,$patchis));
    return 0;
}

sub patches_install_verify_sys {
    my ($cpic,$sys,$patchis)= @_;
    my ($fails,$patch,$pdfr,$msg);
    $patchis=~s/^\s*//;
    $patchis=~s/\s*$//;
    $fails=0;
    $pdfr=Msg::get('pdfr');
    for my $patchi (split(/\s+/,$patchis)) {
        $patch=$sys->patch($patchi);
        if (Cfg::opt('makeresponsefile') || $patch->install_success_sys($sys)) {
            $msg=Msg::new("$patch->{patch} $pdfr installed successfully on $sys->{sys}", 30, 1278, "$patch->{patch}", "$pdfr", "$sys->{sys}");
            $msg->log;
            $sys->set_value('patchinstallpass', 'push', $patch->{patch});
            $sys->set_value("patchvers,$patch->{patch}",$patch->{vers});
            EDRu::appendfile($patch->{patch}, EDR::tmpdir()."/patch_install_pass.$sys->{sys}")
                if (Cfg::opt('bait'));
        } else {
            $msg=Msg::new("$patch->{patch} $pdfr failed to install on $sys->{sys}", 30, 1279, "$patch->{patch}", "$pdfr", "$sys->{sys}");
            $msg->log;
            $sys->set_value('installfailmsg', 'push', $msg->{msg});
            $sys->set_value('patchinstallfail', 'push', $patch->{patch});
            EDRu::appendfile($patch->{patch}, EDR::tmpdir()."/patch_install_fail.$sys->{sys}")
                if (Cfg::opt('bait'));
            $fails++;
        }
    }
    return 0 if ($fails);
    return 1;
}

# used when one thread runs copy, install, remove routine
sub pkg_install_sys {
    my ($cpic,$sys,$pkg,$copythr)= @_;
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    return 1 if (Cfg::opt('makeresponsefile'));;
    # pkg install is conditional
    return if (!EDRu::inarr($pkg->{pkgi}, @{$sys->{installpkgs}}));
    if ($pkg->{mpok}) {
        Msg::log("Not installing $pkg->{pkg} on $sys->{sys} as it is missing from the release media");
        return;
    }

    $pkg->preinstall_sys($sys);
    ($copythr) ? $pkg->wait_for_copy_sys($sys) :
                 $pkg->copy_sys($sys);
    $pkg->install_sys($sys);
    $pkg->remove_sys($sys);
    my $iof=EDRu::outputfile('install', $sys->{sys}, $pkg->{pkg});
    $sys->copy_to_sys($cpic->localsys,$iof) unless ($sys->{islocal});
    $pkg->postinstall_sys($sys);

    return 0 if ($cpic->false_failure);
    return 1 if ($pkg->install_success_sys($sys));
    return 0;
}

sub patch_install_sys {
    my ($cpic,$sys,$patch,$copythr)= @_;
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $patch=$sys->patch($patch) unless (ref($patch));
    return 1 if (Cfg::opt('makeresponsefile'));;
    # patch install is conditional
    return if (!EDRu::inarr($patch->{patch}, @{$sys->{installpatches}}));
    if ($patch->{mpok}) {
        Msg::log("Not installing $patch->{patch_vers} on $sys->{sys} as it is missing from the release media");
        return;
    }

    $patch->preinstall_sys($sys);
    ($copythr) ? $patch->wait_for_copy_sys($sys) :
                 $patch->copy_sys($sys);
    $patch->install_sys($sys);
    $patch->remove_sys($sys);
    my $iof=EDRu::outputfile('install', $sys->{sys}, "patch.$patch->{patch_vers}");
    $sys->copy_to_sys($cpic->localsys,$iof) unless ($sys->{islocal});
    $patch->postinstall_sys($sys);

    return 1 if ($patch->install_success_sys($sys));
    return 0;
}

sub pkg_install_failed_sys {
    my ($sys,$pkg)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    push(@{$sys->{pkginstallfail}}, $pkg->{pkg});
    my $pdfr=Msg::get('pdfr');
    my $io=EDRu::readfile(EDRu::outputfile('install', $sys->{sys}, $pkg->{pkg}));
    my $msg=Msg::new("$pkg->{pkg} $pdfr failed to install on $sys->{sys}", 30, 1279, "$pkg->{pkg}", "$pdfr", "$sys->{sys}");
    $msg->log;
    Msg::log("$io");
    push(@{$sys->{installfailmsg}}, $msg);
    EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_install_fail.$sys->{sys}")
        if (Cfg::opt('bait'));
    return '';
}

sub pkg_install_passed_sys {
    my ($sys,$pkg)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    # pkg install is conditional
    return '' if (!EDRu::inarr($pkg->{pkgi}, @{$sys->{installpkgs}}));
    push(@{$sys->{pkginstallpass}}, $pkg->{pkg});
    $sys->{pkgvers}{$pkg->{pkg}}=$pkg->{vers};
    my $pdfr=Msg::get('pdfr');
    Msg::log("$pkg->{pkg} $pdfr installed successfully on $sys->{sys}");
    EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_install_pass.$sys->{sys}")
        if (Cfg::opt('bait'));
    return '';
}

sub pkg_install_rc_check_sys {
    my ($cpic,$sys,$pkg,$rc)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    if ($rc) {
        pkg_install_passed_sys($sys, $pkg);
    } elsif (defined $rc) {
        pkg_install_failed_sys($sys, $pkg);
    }
    return $rc;
}

sub patch_install_failed_sys {
    my ($sys,$patch)= @_;
    $patch=$sys->patch($patch) unless (ref($patch));
    $sys->set_value('patchinstallfail', 'push', $patch->{patch_vers});
    my $io=EDRu::readfile(EDRu::outputfile('install', $sys->{sys}, "patch.$patch->{patch_vers}"));
    my $msg=Msg::new("$patch->{patch_vers} patch failed to install on $sys->{sys}", 30, 1280, "$patch->{patch_vers}", "$sys->{sys}");
    $msg->log;
    Msg::log("$io");
    push(@{$sys->{installfailmsg}}, $msg);
    EDRu::appendfile($patch->{patch_vers}, EDR::tmpdir()."/patch_install_fail.$sys->{sys}")
        if (Cfg::opt('bait'));
    return '';
}

sub patch_install_passed_sys {
    my ($sys,$patch)= @_;
    $patch=$sys->patch($patch) unless (ref($patch));
    # patch install is conditional
    return '' if (!EDRu::inarr($patch->{patch}, @{$sys->{installpatches}}));
    $sys->set_value('patchinstallpass', 'push', $patch->{patch_vers});
    Msg::log("$patch->{patch_vers} patch installed successfully on $sys->{sys}");
    EDRu::appendfile($patch->{patch_vers},
        EDR::tmpdir()."/patch_install_pass.$sys->{sys}") if (Cfg::opt('bait'));
    return '';
}

sub patch_install_rc_check_sys {
    my ($cpic,$sys,$patch,$rc)= @_;
    $patch=$sys->patch($patch) unless (ref($patch));
    if ($rc) {
        patch_install_passed_sys($sys, $patch);
    } elsif (defined $rc) {
        patch_install_failed_sys($sys, $patch);
    }
    return $rc;
}

sub prod_preinstall_sys {
    my ($cpic,$sys) = @_;
    my ($tmpsub,$prod,$mr);
    return 1 if (Cfg::opt('makeresponsefile'));;
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');

    $sys->rel->preinstall_sys($sys);
    $sys->rel->preinstall_patch_sys($sys);

    $prod=$sys->prod;
    $tmpsub = upgrade_sub('preinstall_sys');
    if ($prod->can($tmpsub)) {
        $prod->$tmpsub($sys);
    }
    if ($mr = $cpic->{mrobj}) {
        $mr->preinstall_sys($sys);
    }
    for my $hotfix(@{$cpic->{hotfixobj}}) {
        $hotfix->preinstall_sys($sys);
    }

    return 1;
}

sub prod_preinstall_errors {
    my ($cpic) = @_;
    my $prod=$cpic->prod;
    $prod->preinstall_errors($cpic) if ($cpic->sys_error);
    EDR::reset_errors_warnings();
    return 1;
}

sub prod_postinstall_sys {
    my ($cpic,$sys) = @_;
    my ($tmpsub,$prod,$rootpath,$scripts);
    return 1 if (Cfg::opt('makeresponsefile'));
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $prod=$sys->prod;
    $tmpsub = upgrade_sub('postinstall_sys');
    if ($prod->can($tmpsub)) {
        $prod->$tmpsub($sys);
    }

    for my $hotfix(@{$cpic->{hotfixobj}}) {
        $hotfix->postinstall_sys($sys);
    }

    if ($cpic->rel->{cpipkg}) {
        $rootpath=Cfg::opt('rootpath');
        if (Cfg::opt('upgrade') && (defined $cpic->rel->{obsolete_scripts})){
            $scripts=join(' ', @{$cpic->rel->{obsolete_scripts}});
            unless ($scripts =~ /^\s*$/m) {
                $sys->cmd("cd $rootpath$cpic->{installpath}; _cmd_rmr $scripts");
            }
        }
        if (Cfg::opt('upgrade') && !$sys->{partial_upgrade}){
            $scripts = $rootpath.$cpic->{installpath}.'/uninstallrp';
            if ($sys->exists($scripts)) {
                $sys->cmd("_cmd_rmr $scripts");
            }
        }
        $cpic->add_install_scripts_sys($sys);
        if (defined $prod->{obsolete_scripts}) {
            $scripts=join(' ', @{$prod->{obsolete_scripts}});
            unless ($scripts =~ /^\s*$/m) {
                $sys->cmd("cd $rootpath$cpic->{installpath}; _cmd_rmr $scripts");
            }
        }
    } else {
        $cpic->copy_cpi_sys($sys) if ($cpic->rel->{type} ne 'H');
    }

    $sys->rel->postinstall_sys($sys);
    return 1;
}

sub prod_postinstall_errors {
    my ($cpic) = @_;
    my $prod=$cpic->prod;
    $prod->postinstall_errors($cpic) if ($cpic->sys_error);
    EDR::reset_errors_warnings();
    return 1;
}

sub update_summary_install {
    my $cpic=shift;
    my ($msg,$pdfrs,$pkgset);

    $pdfrs=Msg::get('pdfrs');
    if ($cpic->{task} eq 'install') {
        if (Cfg::opt('installminpkgs')) {
            $pkgset = Msg::new("The Minimum package set was selected for installation", 30, 1281);
        } elsif (Cfg::opt('installrecpkgs')) {
            $pkgset = Msg::new("The Recommended package set was selected for installation", 30, 1282);
        } elsif (Cfg::opt('installallpkgs')) {
            $pkgset = Msg::new("The All package set was selected for installation", 30, 1283);
        }
    }
    for my $sys (@{$cpic->{systems}}) {
        if ($cpic->{task} eq 'install' && $pkgset) {
            $msg=Msg::new_obj("$pkgset->{msg}");
            $cpic->update_summary($msg);
        }
        if ($#{$sys->{pkginstallpass}}>=0) {
            $msg=Msg::new("The following $pdfrs were installed on $sys->{sys}:", 30, 1284, "$pdfrs", "$sys->{sys}");
            $cpic->update_summary($msg, @{$sys->{pkginstallpass}});
        }
        if ($#{$sys->{pkginstallfail}}>=0) {
            $msg=Msg::new("The following $pdfrs failed to install on $sys->{sys}:", 30, 1285, "$pdfrs", "$sys->{sys}");
            $cpic->update_summary($msg, @{$sys->{pkginstallfail}});
        }
        if ($#{$sys->{patchinstallpass}}>=0) {
          $msg=Msg::new("The following patches were installed on $sys->{sys}:", 30, 1286, "$sys->{sys}");
            $cpic->update_summary($msg, @{$sys->{patchinstallpass}});
        }
        if ($#{$sys->{patchinstallfail}}>=0) {
            $msg=Msg::new("The following patches failed to install on $sys->{sys}:", 30, 1287, "$sys->{sys}");
            $cpic->update_summary($msg, @{$sys->{patchinstallfail}});
        }
    }
    return '';
}

#this function could be called in the "sub install" function as an enhancement, its related incident number is 2663858 and it#could be evaluated in the future.
sub ask_install_continue {
    my $cpic = shift;
    my ($msg, $msg_advice, $pdfrs, $ayn, $pkginstallfail_flag, $patchinstallfail_flag);
    my $edr = Obj::edr();
    for my $sys (@{$cpic->{systems}}) {
        if ($#{$sys->{pkginstallfail}}>=0) {
            $pkginstallfail_flag = 1;
        }
        if ($#{$sys->{patchinstallfail}}>=0) {
            $patchinstallfail_flag = 1;
        }
    }
    if ($pkginstallfail_flag == 1 || $patchinstallfail_flag == 1){
        $pdfrs=Msg::get('pdfrs');
        $msg = Msg::new("$pdfrs/patches failed to install. Do you want to exit $cpic->{script}?", 30, 1288, "$pdfrs", "$cpic->{script}");
        $ayn = $msg->ayny;
        Msg::n();
        if ($ayn eq 'Y') {
            $msg_advice = Msg::new("Refer to the installer logs for details and correct the settings. Continue with the installer again, or try to install the packages/patches manually.", 30, 1289);
            $msg_advice->log;
            $msg_advice->print;
            Msg::n();
            $cpic->edr_completion();
        }
    }
    return '';
}

1;
