#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
package CPIC;
use strict;

# CPIC::verify_systems has 6 steps
#    STEP                         RUN
# 1. installed_prod_sys, prod_version_sys   always
# 2. prod_prerequisite_sys        install_check_task
# 3. rel_platvers_sys             always
# 4. fsspace_sys                  install_check_task && !nospacecheck
# 5. vxlicrep_sys                 install_check_task
# 6. prod_precheck_sys            always, prod sub has appropriate task filters

# we can define this per-check uniquely, but currently is the same for the 3 optional tasks
sub install_check_task { return Cfg::opt(qw(precheck install patchupgrade hotfixupgrade upgrade)); }

sub systems {
    my $cpic=shift;
    my ($cfg,$default_sys_names,$edr,$rel,$web,$phase,$verify_passed,$clean_error,$rtn);

    $cfg = Obj::cfg();
    $edr=Obj::edr();
    $web=Obj::web();
    $rel=$cpic->rel();
    $clean_error=0;
    $cpic->discover_common_padv() if ($cpic->{padv} eq 'Common');

    # always requires input
    Cfg::unset_opt('silent');
    while (!$verify_passed) {
        $default_sys_names = $cpic->prod_default_systemnames();
        if (!$cfg->{systems} && Obj::webui()) {
            if ($clean_error > 0) {
                EDR::reset_prodinfo_sys();
                EDR::reset_pkginfo_sys();
                EDR::reset_errors_warnings();
            }
            $clean_error=1;
            my $msg = Msg::new("Enter system names separated by spaces and click Next", 30, 1398);
            $web->web_script_form('selectsystems',$msg);
        }
        return '' if (Cfg::opt('back'));
        $edr->cli_select_systems($default_sys_names);
        $edr->{savelog}=1; # from this point on log locations are displayed
        $cpic->{systems}=$edr->init_sys_objects();
        $cpic->check_memory_swap_for_localsys() if (!EDR::local_windows());
        if (Obj::webui()) {
            if (Cfg::opt("rolling_upgrade")) {
                $web->web_script_form('precheck', "Verifying systems");
            } else {
                $web->web_script_form('precheck');
            }
        }
        $cpic->cli_set_title();

        $cpic->{retry_verify_checks}=0;

        $phase=$cpic->setup_verify_systems_phase();
        $phase->execute_tasks();

        next if ($cpic->{retry_verify_checks});

        if (EDR::continue_checks()) {
            $cpic->verify_all_systems();
        }
        $verify_passed=$cpic->verify_system_checks;
        delete $cfg->{systems} unless ($verify_passed);
        # verify_system_checks may proceed with successful systems only
        $cpic->{systems}=$edr->{systems};
        $cpic->cli_set_title();
    }

    $cpic->precheck_report() if (Cfg::opt('precheck'));
    $cpic->handle_precheck_issues();
    $rel->read_licenses() if ((Cfg::opt(qw(install configure))) &&
                              (!$cpic->{eebi}));
    $cpic->update_summary_verify();
    EDR::reset_errors_warnings();
    return;
}

sub setup_verify_systems_phase {
    my ($cpic, $check_com_only) = @_;
    my ($edr,$rel,$phase,$desc,$task,$task_name,%task_params,$pdfrs,$id);
    my ($skip_prerequisite,$skip_fsspace,$skip_license);

    $cpic->{falsefailures}=0;
    $phase= Phase->new('CPIC_prechecks');

    # Directly return $phase if it is already created.
    return $phase if (defined $phase->{description});

    $desc = Msg::new("Verifying systems", 30, 1399);
    $phase->set_description($desc);
    $phase->set_fatal_error_key('stop_checks');
    $phase->initialize_tasks();

    $edr=Obj::edr();
    $rel=$cpic->rel();
    $pdfrs=Msg::get('pdfrs');

    # By default, skip product version/license checking,
    # skip patches checking, skip file system space checking
    $skip_prerequisite=1;
    $skip_fsspace=1;
    $skip_license=1;

    # if install/upgrade task specified, then run above checkings.
    if (install_check_task()) {
        $skip_prerequisite=0;
        $skip_fsspace=0;
        $skip_license=0;
    }
    if ($rel->rollback_check_task()) {
        $skip_fsspace=0;
    }
    $skip_fsspace=1 if (Cfg::opt('nospacecheck'));
    $skip_license=1 if (!$rel->can("read_licenses_sys"));

    $id=100;
    $task_name='edr_check_communication';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking system communication", 30, 1400),
        'description_sys' => Msg::new("Checking communication on #{SYS}", 30, 1401),
        'if_per_system' => 1,
        'handler' => \&EDR::transport_sys,
        'handler_object_arg' => $edr,
        'post_handler' => \&CPIC::post_transport,
        'post_handler_object_arg' => $cpic,
        'error_handler' => \&CPIC::transport_error_handler,
        'error_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    return $phase if ($check_com_only);

    $id+=100;
    $task_name='edr_check_release_compatibility';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking release compatibility", 30, 1402),
        'description_sys' => Msg::new("Checking release compatibility on #{SYS}", 30, 1403),
        'if_per_system' => 1,
        'handler' => \&EDR::supported_padv_sys,
        'handler_object_arg' => $edr,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_check_installed_product';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking installed product", 30, 1404),
        'description_sys' => Msg::new("Checking installed product on #{SYS}", 30, 1405),
        'if_per_system' => 1,
        'handler' => \&CPIC::installed_prod_info_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::post_installed_prod_info,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_check_prerequisite_packages_patches';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking prerequisite patches and $pdfrs", 30, 1406, "$pdfrs"),
        'description_sys' => Msg::new("Checking prerequisite patches and $pdfrs on #{SYS}", 30, 1407, "$pdfrs"),
        'if_per_system' => 1,
        'skip' => $skip_prerequisite,
        'handler' => \&CPIC::prod_prerequisite_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::post_prod_prerequisite,
        'post_handler_object_arg' => $cpic,
        'error_handler' => \&CPIC::prod_prerequisite_error_handler,
        'error_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_check_platform_version';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking platform version", 30, 1408),
        'description_sys' => Msg::new("Checking platform version on #{SYS}", 30, 1409),
        'if_per_system' => 1,
        'handler' => \&CPIC::rel_platvers_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::post_rel_platvers,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_check_filesystem_free_space';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking file system free space", 30, 1410),
        'description_sys' => Msg::new("Checking file system free space on #{SYS}", 30, 1411),
        'if_per_system' => 1,
        'skip' => $skip_fsspace,
        'handler' => \&CPIC::fsspace_sys,
        'handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_check_product_license';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Checking product licensing", 30, 1412),
        'description_sys' => Msg::new("Checking product licensing on #{SYS}", 30, 1413),
        'if_per_system' => 1,
        'skip' => $skip_license,
        'handler' => \&CPIC::prod_licenses_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::post_prod_licenses,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    $id+=100;
    $task_name='cpic_perform_product_prechecks';
    %task_params= (
        'sequence_id' => $id,
        'description' => Msg::new("Performing product prechecks", 30, 1414),
        'description_sys' => Msg::new("Performing product prechecks on #{SYS}", 30, 1415),
        'if_per_system' => 1,
        'pre_handler' => \&CPIC::pre_prod_precheck,
        'pre_handler_object_arg' => $cpic,
        'handler' => \&CPIC::prod_precheck_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::post_prod_precheck,
        'post_handler_object_arg' => $cpic,
        'error_handler' => \&CPIC::prod_precheck_error_handler,
        'error_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    return $phase;
}

# broken out to override, if necessary
sub verify_system_checks {
    my $cpic=shift;
    my $edr=Obj::edr();
    return 1 if (Cfg::opt(qw(rwe precheck)));
    return 1 if ($edr->verify_systems_errors());
    return '' if (Obj::webui());

    $cpic->rel->completion();
    $edr->exit_noexitfile();
    return;
}

sub create_system_padv_objects {
    my $cpic=shift;

    return if (Cfg::opt('serial'));

    # create padv objects for main thread for each system
    for my $sys (@{$cpic->{systems}}) {
        for my $padv ($sys->{padv}, $cpic->{padvisa}{$sys->{padv}}) {
            "Padv\::$padv"->new() if ($padv && (!Obj::created('Padv', $padv)));
        }
        my $transport=$sys->{transport};
        $sys->{transport_obj}="Transport\::$transport"->new() if ($transport);
    }
    return;
}

sub create_system_cpip_objects {
    my $cpic=shift;
    my ($padv,$rel);

    # create CPIP objects for main thread for each system
    for my $sys (@{$cpic->{systems}}) {
        next if ($sys->{stop_checks});
        $padv=$cpic->{padvisa}{$sys->{padv}};
        if ($padv && !Obj::created('Rel', $cpic->{release}, $padv)) {
            $rel=$cpic->create_cpip_objects($padv);
            $cpic->verify_media($rel);
        }
        $padv=$sys->{padv};
        if ($padv && !Obj::created('Rel', $cpic->{release}, $padv)) {
            $rel=$cpic->create_cpip_objects($padv);
        }
    }
    return;
}

sub create_system_objects {
    my $cpic=shift;

    # Stop prechecks if any system deny communication
    my $edr=Obj::edr();
    $edr->{stop_checks}=1 if (EDR::com_deny_sys());

    # re-create Padv objects in main thread if in thread mode
    $cpic->create_system_padv_objects() if (!Cfg::opt('serial'));
    $cpic->create_system_cpip_objects();
    return;
}

sub post_transport {
    my $cpic=shift;
    my ($edr,$hostname,$msg,%duplicated,$duplicated_sys);

    # Check whether any 2 systems are the same node
    $edr=Obj::edr();
    %duplicated=();
    for my $sys (@{$cpic->{systems}}) {
        $edr->{stop_checks}=1 if ($sys->{stop_checks});

        $hostname = $sys->{hostname};
        next unless (defined $hostname);

        $duplicated_sys=$duplicated{$hostname};
        if (defined $duplicated_sys) {
            $msg=Msg::new("$sys->{sys} and $duplicated_sys are the same node", 30, 1416, "$sys->{sys}", "$duplicated_sys");
            $sys->push_stop_checks_error($msg);
            $edr->{stop_checks}=1;
            next;
        } else {
            $duplicated{$hostname}=$sys->{sys};
        }
    }

    # Create system objects in main thread
    $cpic->create_system_objects();

    return 1;
}

# To setup transportation if not ready
sub transport_error_handler {
    my $cpic=shift;
    my ($edr,$rtn);

    if ((!EDR::local_windows()) &&
        (!Cfg::opt("responsefile")) &&
        (EDR::com_deny_sys())) {
        $edr=Obj::edr();
        $rtn=$edr->ask_com_setup();
        $cpic->{retry_verify_checks}=1 if ($rtn>0);
    }

    return 1;
}

sub check_memory_swap_for_localsys {
    my $cpic=shift;
    my ($localsys,$sys1,$edr,$msg,$nsystems,$nthreads,$free_tmppath_space,$minimum_tmppath_space);
    my ($memory_size,$small_cluster_memory,$large_cluster_memory);
    my ($swap_size,$minimum_swap_size);

    # The RAM/swap requirements for successfully installation based on the number of install nodes:
    #
    #  1. If the number install nodes is greater than 8 nodes,  it is recommended to have at least 2G RAM on the install driver node, otherwise 1G is recommended.
    #  2. If the number install nodes is greater than 8 nodes,  it is required to have minimal 1G of free swap space on the install driver node, otherwise, (<num_of_nodes> + 1) * 128 M of free swap space is required.
    #
    # For more informations about the requirements, you can refer to the following incidents:
    #  1846892:    minimal free swap space requirements
    #  2107764:    recommended RAM requirements for large cluster installation.

    $localsys=Obj::localsys();
    $edr=Obj::edr();
    $nsystems=$cpic->nsystems();
    $sys1=$cpic->{systems}->[0];

    # check if tmp path's space is enough on the local system to save temporary files for communication
    $minimum_tmppath_space=100;
    $free_tmppath_space=$cpic->volumespace_sys($localsys, $edr->{tmppath}, 1);
    if (!defined $free_tmppath_space) {
        $msg=Msg::new("Cannot determine free space available for $edr->{tmppath} on $localsys->{sys}", 30, 1417, "$edr->{tmppath}", "$localsys->{sys}");
        $msg->log;
    } elsif ($free_tmppath_space<$minimum_tmppath_space) {
        $msg=Msg::new("$minimum_tmppath_space KB is required in the $edr->{tmppath} volume and only $free_tmppath_space KB is available on $localsys->{sys}", 30, 1418, "$minimum_tmppath_space", "$edr->{tmppath}", "$free_tmppath_space", "$localsys->{sys}");
        $msg->die;
    }

    if ($nsystems>4 && !Cfg::opt('serial')) {
        $memory_size=$localsys->padv->memory_size_sys($localsys);
        if ($memory_size) {
            # Convert the size based on MegaByte.
            if ($memory_size=~/\s*(\d+)\s*[kK]/mx) {
                $memory_size=int($1/1024);
            } elsif ($memory_size=~/\s*(\d+)\s*[mM]/mx) {
                $memory_size=$1;
            } elsif ($memory_size=~/\s*(\d+)\s*[gG]/mx) {
                $memory_size=$1*1024;
            }

            # large cluster
            $large_cluster_memory=$edr->{large_cluster_memory} || 2048;
            if ($memory_size < $large_cluster_memory) {
                $msg = Msg::new("The available memory on the local system '$localsys->{sys}' is $memory_size MB. It is recommended to have $large_cluster_memory MB of memory, or use the '-serial' option to run the $cpic->{script} program.", 30, 1419, "$localsys->{sys}", "$memory_size", "$large_cluster_memory", "$cpic->{script}");
                $sys1->push_warning($msg);
            }

            $small_cluster_memory=$edr->{small_cluster_memory} || 1280;
            if ($memory_size < $small_cluster_memory) {
                # Set threadlimit to 4
                $cpic->{threadlimit}=4;
            }
        }
    }

    if (Cfg::opt('serial')) {
        $nthreads=1;
    } else {
        $nthreads=$cpic->{threadlimit};
        if ($nsystems<$nthreads) {
            $nthreads=$nsystems+1;
        }
    }

    # Set minimum swap size as $nthreads * 128M.
    $minimum_swap_size=$nthreads * 128;

    $swap_size=$localsys->padv->swap_size_sys($localsys);
    if ($swap_size) {
        # Convert the size based on MegaByte.
        if ($swap_size=~/\s*(\d+)\s*[kK]/mx) {
            $swap_size=int($1/1024);
        } elsif ($swap_size=~/\s*(\d+)\s*[mM]/mx) {
            $swap_size=$1;
        } elsif ($swap_size=~/\s*(\d+)\s*[gG]/mx) {
            $swap_size=$1*1024;
        }
        if ($swap_size < $minimum_swap_size) {
            $msg = Msg::new("The free swap space on local system '$localsys->{sys}' is $swap_size MB. It is recommended to setup minimal $minimum_swap_size MB of free swap space to successfully run the $cpic->{script} program.", 30, 1420, "$localsys->{sys}", "$swap_size", "$minimum_swap_size", "$cpic->{script}");
            $msg->die;
        }
    }

    return '';
}

sub verify_list {
    my ($cpic,$cpici,$sysi) = @_;
    my ($index,$sys,@list);
    $sysi||=$cpici;
    for my $index (@{$cpic->{$cpici}}) {
        for my $sys (@{$cpic->{systems}}) {
            if (EDRu::inarr($index,@{$sys->{$sysi}})) {
                push(@list, $index);
                last;
            }
        }
    }
    return \@list;
}

sub verify_pkg_version_sys {
    my $cpic = shift;
    my $sys = shift;
    my ($rel, $cv, $pvers, $msg, $pkg, $patch, $padv);
    $rel = $cpic->rel;
    for my $pkgname (keys %{$rel->{hf_packages}}) {
        $pkg = $sys->pkg($pkgname);
        $padv = $sys->padv;
        for my $patchi (@{$pkg->patches_sys($sys)}) {
            $patch = $sys->patch($patchi);
            #AIX patch check is essentially a pkg check
            if ($sys->aix()) {
                $pvers = $pkg->version_sys($sys);
                $cv = EDRu::compvers($patch->{vers}, $pvers);
                if ($cv == 2) {
                    $msg = Msg::new("The higher version of $pkgname($pvers) is already installed on $sys->{sys}", 30, 1421, "$pkgname", "$pvers", "$sys->{sys}");
                    $sys->push_error($msg);
                    last;
                }
                if ($cv ==0) {
                    $msg = Msg::new("$pkgname($pvers) is already installed on $sys->{sys}", 30, 1422, "$pkgname", "$pvers", "$sys->{sys}");
                    $sys->push_note($msg);
                }
            } else {
                $pvers = $padv->patch_version_sys($sys, $patch);
                $cv = EDRu::compvers($patch->{vers},$pvers, 4);
                if ($cv == 2) {
                    $msg = Msg::new("The higher version of $patch->{patchname}($pvers) is already installed on $sys->{sys}", 30, 1421, "$patch->{patchname}", "$pvers", "$sys->{sys}");
                    $sys->push_error($msg);
                    last;
                }
                if ($cv ==0) {
                    $msg = Msg::new("$patch->{patchname}($pvers) is already installed on $sys->{sys}", 30, 1422, "$patch->{patchname}", "$pvers", "$sys->{sys}");
                    $sys->push_note($msg);
                }
            }
        }
    }
    return 1;
}

sub verify_prod_version_sys {
    my $cpic = shift;
    my $sys = shift;
    my ($rel, $cv, $ivers, $msg, $prodname, $prod, $subProdIsHigher);
    $prod=$cpic->prod;
    $rel = $cpic->rel;
    $ivers=$prod->version_sys($sys);
    if (!$ivers && Cfg::opt(qw(upgrade patchupgrade))) {
        $msg=Msg::new("Cannot upgrade $prod->{abbr} to version $prod->{vers} on $sys->{sys} because it is not installed on the system", 30, 1423, "$prod->{abbr}", "$prod->{vers}", "$sys->{sys}");
        $sys->push_error($msg);
    }
    if($ivers){
        for my $subprodi(@{$sys->{upgradeprod}}){
            my $subprod=$sys->prod($subprodi);
            my $subprod_ivers=$subprod->version_sys($sys,1);
            if (EDRu::compvers($subprod_ivers,$ivers,3)==1){
                $subProdIsHigher = 1;
                last;
            }
        }
    }
    $cv=EDRu::compvers($ivers,$prod->{vers},4) if ($ivers);
    if (!($sys->{errors}) && $cv && Cfg::opt(qw(upgrade patchupgrade hotfixupgrade))) {
        $prodname= $sys->{dbed_prod} || $prod->{abbr};
        $prodname= $prod->{upgrade_from} if (defined $prod->{upgrade_from});
        $ivers = $prod->version_with_extra_mainpkgs_sys($sys);
        if (!defined $sys->{mix_partial_upgrade}) {
            if (Cfg::opt('upgrade_kernelpkgs')) {
                $msg=Msg::new("$prodname version $ivers kernel components will be upgraded to $prod->{abbr} version $prod->{vers} on $sys->{sys}", 30, 1424, "$prodname", "$ivers", "$prod->{abbr}", "$prod->{vers}", "$sys->{sys}");
            } elsif (Cfg::opt('upgrade_nonkernelpkgs')) {
                $msg=Msg::new("$prodname version $ivers non-kernel components will be upgraded to $prod->{abbr} version $prod->{vers} on $sys->{sys}", 30, 1425, "$prodname", "$ivers", "$prod->{abbr}", "$prod->{vers}", "$sys->{sys}");
            } elsif (!$subProdIsHigher) {
                $msg=Msg::new("$prodname version $ivers will be upgraded to $prod->{abbr} version $prod->{vers} on $sys->{sys}", 30, 1426, "$prodname", "$ivers", "$prod->{abbr}", "$prod->{vers}", "$sys->{sys}");
            }
        }
        $sys->push_note($msg) unless(EDRu::inarr($prodname,@{$rel->{ru_noupgrademsg_prod}}) &&
            (Cfg::opt('upgrade_nonkernelpkgs') || $rel->determine_rolling_upgrade($cpic->{prod})==2));
    }
    return 1;
}

sub verify_all_systems {
    my $cpic=shift;
    my ($ivers,$msg,$prod,$prodname,$rel,$cv);
    return 1 unless ($cpic->{prod});
    $prod=$cpic->prod;
    $rel=$cpic->rel;
    $cpic->cli_set_title();

    for my $sys (@{$cpic->{systems}}) {
        # Note: how should this handle some systems having stop_checks set?
        next if ($sys->{stop_checks});
        if ($rel->{type} eq 'H') {
            #hotfix only needs to check pkg version
            $cpic->verify_pkg_version_sys($sys);
        } else {
            $cpic->verify_prod_version_sys($sys);
        }
    }
    return 1;
}

sub sys_error {
    my ($cpic,$erronly_flag) = @_;
    for my $sys (@{$cpic->{systems}}) {
        #$erronly_flag ?
        #return 1 if ($sys->{errors}) :
        return 1 if ($sys->{errors} || $sys->{warnings});
    }
    return '';
}

sub prod_default_systemnames {
    my $cpic=shift;
    return ($cpic->{prod}) ? $cpic->prod->default_systemnames :
                             $cpic->rel->default_systemnames;
}

sub rel_platvers_sys {
    my($cpic,$sys)=@_;
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return '' if ($sys->{stop_checks});
    return $sys->rel->platvers_sys($sys);
}

sub post_rel_platvers {
    my $cpic=shift;
    return $cpic->rel->post_platvers();
}

sub prod_prerequisite_sys {
    my ($cpic,$sys) = @_;

    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if ($sys->{stop_checks});
    return 1 unless (Cfg::opt(qw(install patchupgrade hotfixupgrade upgrade configure start precheck postcheck)));

    my $rtn=$cpic->prod_prerequisite_ospatches_sys($sys);
    $rtn=$cpic->prod_prerequisite_ospkgs_sys($sys) && $rtn;
    $rtn=$cpic->prod_prerequisite_thirdpartypkgs_sys($sys) && $rtn;
    $rtn=$cpic->prod_prerequisite_oslibs_sys($sys) && $rtn;
    $rtn=$cpic->prod_prerequisite_hfpatches_sys($sys) && $rtn;
    return $rtn;
}

sub prod_prerequisite_ospatches_sys {
    my ($cpic,$sys) = @_;
    my ($rel,$prod,$pkg,$padv,$pdfrs,$minorvers,$fallback_minorvers);
    my ($patchindex,$patchi,$vers,$patchi2,$vers2,$i,$msg);
    my (%checked_patches,$patches,$patchobj,@ospatches,@missing_ospatches);

    $rel=$sys->rel;
    $prod=$sys->prod;
    $padv=$sys->padv;

    $pdfrs=Msg::get('pdfrs');
    $minorvers=$sys->{minorvers};
    $patchindex=$sys->patchindex();

    @ospatches=();

    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);
        $pkg->define_ospatches($padv,$sys);

        push(@ospatches, @{$pkg->{ospatches}{all}})
            if (defined($pkg->{ospatches}{all}));
        push(@ospatches, @{$pkg->{ospatches}{$patchindex}})
            if (defined($pkg->{ospatches}{$patchindex}));
        if ($minorvers) {
            $fallback_minorvers=$pkg->{minorvers_for_fallback};
            $fallback_minorvers||=$rel->{minorvers_for_fallback};

            if (defined $pkg->{ospatches}{$minorvers}) {
                push(@ospatches, @{$pkg->{ospatches}{$minorvers}});
            } elsif (defined $fallback_minorvers &&
                     defined $pkg->{ospatches}{$fallback_minorvers}) {
                # The ospatches are not defined for this minorvers platform,
                # try to check the fallback category
                push(@ospatches, @{$pkg->{ospatches}{$fallback_minorvers}});
            }
        }
    }

    # Add release level information
    if (defined $rel) {
        push(@ospatches, @{$rel->{ospatches}{all}})
            if (defined($rel->{ospatches}{all}));
        push(@ospatches, @{$rel->{ospatches}{$patchindex}})
            if (defined($rel->{ospatches}{$patchindex}));
        if ($minorvers) {
            $fallback_minorvers=$rel->{minorvers_for_fallback};

            if (defined $rel->{ospatches}{$minorvers}) {
                push(@ospatches, @{$rel->{ospatches}{$minorvers}});
            } elsif (defined $fallback_minorvers &&
                     defined $rel->{ospatches}{$fallback_minorvers}) {
                # The ospatches are not defined for this minorvers platform,
                # try to check the fallback category
                push(@ospatches, @{$rel->{ospatches}{$fallback_minorvers}});
            }
        }
    }

    return 1 unless (@ospatches);

    # To check OS patch dependencies.
    @missing_ospatches=();
    for my $ospatch (@ospatches) {
        next if (defined $checked_patches{$ospatch});
        $checked_patches{$ospatch}=1;

        ($patchi,$vers)=split(/-/m,$ospatch,2);
        $patchobj=Patch->new($patchi,$sys->{padv});
        $patchobj->{patch_vers}=$ospatch;
        push(@missing_ospatches,$ospatch) unless ($padv->ospatch_sys($sys,$patchobj));
    }

    # Code to remove lower versions of ospatches e.g remove A-A if A-B & A-A in the list (B>A)
    $i=0;
    for my $ospatch (@missing_ospatches) {
        ($patchi,$vers)=split(/-/m,$ospatch,2);
        for my $ospatch2 (@missing_ospatches) {
            next if ($ospatch eq $ospatch2);
            if ($ospatch2 =~ /$patchi/m) {
               ($patchi2,$vers2)=split(/-/m,$ospatch2,2);
               delete $missing_ospatches[$i] if (EDRu::compvers($vers2,$vers)==1);
            }
        }
        $i++;
    }

    #print the missing OS patches list
    if (@missing_ospatches) {
        $patches=join(' ',@missing_ospatches);
        if (Cfg::opt('ignorepatchreqs')) {
            $msg=Msg::new("The following required OS patches were not found on $sys->{sys}:\n\t$patches\nPackage installation may fail when required OS patches are not installed on the system", 30, 1427, "$sys->{sys}", "$patches");
            $sys->push_warning($msg);
        } else {
            $msg=Msg::new("The following required OS patches were not found on $sys->{sys}:\n\t$patches\nPackage installation may fail when required OS patches are not installed on the system. You can attempt to install $prod->{abbr} $pdfrs without the missing dependencies using the '-ignorepatchreqs' option.", 30, 1428, "$sys->{sys}", "$patches", "$prod->{abbr}", "$pdfrs");
            $sys->push_error($msg);
        }
        return 0;
    }

    return 1;
}

sub prod_prerequisite_ospkgs_sys {
    my ($cpic,$sys) = @_;
    my ($rel,$prod,$pkg,$pdfrs,$minorvers,$fallback_minorvers);
    my ($pkgi,$vers,$defvers,$msg,$pkgobj,$pkgs,%checked_pkgs,@ospkgs,@missing_ospkgs);

    $rel=$sys->rel;
    $prod=$sys->prod;

    $pdfrs=Msg::get('pdfrs');
    $minorvers=$sys->{minorvers};

    # check the existence of prod's prerequisite OS pkgs and version if any.
    @ospkgs=();

    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);
        push(@ospkgs, @{$pkg->{ospkgs}{all}})
            if (defined($pkg->{ospkgs}{all}));
        push(@ospkgs, @{$pkg->{ospkgs}{$sys->{platvers}}})
            if (defined($pkg->{ospkgs}{$sys->{platvers}}));
        if ($minorvers) {
            $fallback_minorvers=$pkg->{minorvers_for_fallback};
            $fallback_minorvers||=$rel->{minorvers_for_fallback};

            if (defined $pkg->{ospkgs}{$minorvers}) {
                push(@ospkgs, @{$pkg->{ospkgs}{$minorvers}})
            } elsif (defined $fallback_minorvers &&
                     defined $pkg->{ospkgs}{$fallback_minorvers}) {
                # The ospkgs are not defined for this minorvers platform,
                # try to check the fallback category
                push(@ospkgs, @{$pkg->{ospkgs}{$fallback_minorvers}})
            }
        }
    }

    # Add release level information
    if (defined $rel) {
        push(@ospkgs, @{$rel->{ospkgs}{all}})
            if (defined($rel->{ospkgs}{all}));
        push(@ospkgs, @{$rel->{ospkgs}{$sys->{platvers}}})
            if (defined($rel->{ospkgs}{$sys->{platvers}}));
        if ($minorvers) {
            $fallback_minorvers=$rel->{minorvers_for_fallback};

            if (defined $rel->{ospkgs}{$minorvers}) {
                push(@ospkgs, @{$rel->{ospkgs}{$minorvers}})
            } elsif (defined $fallback_minorvers &&
                     defined $rel->{ospkgs}{$fallback_minorvers}) {
                # The ospkgs are not defined for this minorvers platform,
                # try to check the fallback category
                push(@ospkgs, @{$rel->{ospkgs}{$fallback_minorvers}})
            }
        }
    }

    $sys->set_value('missing_ospkgs_for_product', undef);
    return 1 unless (@ospkgs);

    # To check OS package dependencies.
    @missing_ospkgs=();
    for my $ospkg (@ospkgs) {
        next if (defined $checked_pkgs{$ospkg});
        $checked_pkgs{$ospkg}=1;

        ($pkgi,$defvers)=split(/\s+/,$ospkg);
        $pkgobj=Pkg->new_pkg($pkgi, $prod->{padv});
        $vers=$pkgobj->version_sys($sys);
        if (!$vers) {
            push(@missing_ospkgs,$ospkg);
        } elsif ($defvers) {
            if (EDRu::compvers($vers,$defvers)==2) {
                $msg=Msg::new("The $pkgi version $vers is lower than $defvers on $sys->{sys}. Please upgrade the $pkgi version before you proceed.", 30, 1429, "$pkgi", "$vers", "$defvers", "$sys->{sys}", "$pkgi");
                $sys->push_warning($msg);
            }
        }
    }

    if (@missing_ospkgs) {
        $sys->set_value('missing_ospkgs_for_product', 'list', @missing_ospkgs);
        $pkgs=join(' ',@missing_ospkgs);
        $msg=Msg::new("The following required OS $pdfrs were not found on $sys->{sys}:\n\t$pkgs", 30, 1430, "$pdfrs", "$sys->{sys}", "$pkgs");
        $sys->push_error($msg);
        return 0;
    }

    return 1;
}

sub prod_prerequisite_thirdpartypkgs_sys {
    my ($cpic,$sys) = (@_);
    my ($rel,$prod,$pkg,$pdfrs,$minorvers,$fallback_minorvers);
    my ($pkgi,$vers,$defvers,$msg,$pkgobj,$pkgs,%checked_pkgs,@thirdpartypkgs,@missing_thirdpartypkgs);

    $rel=$sys->rel;
    $prod=$sys->prod;

    $pdfrs=Msg::get('pdfrs');
    $minorvers=$sys->{minorvers};

    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);
        push (@thirdpartypkgs, @{$pkg->{thirdpartypkgs}{all}}) 
            if (defined($pkg->{thirdpartypkgs}{all}));
        push (@thirdpartypkgs, @{$pkg->{thirdpartypkgs}{$sys->{platvers}}})
            if (defined($pkg->{thirdpartypkgs}{$sys->{platvers}}));
        if ($minorvers) {
            $fallback_minorvers=$pkg->{minorvers_for_fallback};
            $fallback_minorvers||=$rel->{minorvers_for_fallback};

            if (defined $pkg->{thirdpartypkgs}{$minorvers}) {
                push(@thirdpartypkgs, @{$pkg->{thirdpartypkgs}{$minorvers}})
            } elsif (defined $fallback_minorvers &&
                     defined $pkg->{thirdpartypkgs}{$fallback_minorvers}) {
                # The thirdpartypkgs are not defined for this minorvers platform,
                # try to check the fallback category
                push(@thirdpartypkgs, @{$pkg->{thirdpartypkgs}{$fallback_minorvers}})
            }
        }
    }

    # Add release level information
    if (defined $rel) {
        push (@thirdpartypkgs, @{$rel->{thirdpartypkgs}{all}}) 
            if (defined($rel->{thirdpartypkgs}{all}));
        push (@thirdpartypkgs, @{$rel->{thirdpartypkgs}{$sys->{platvers}}})
            if (defined($rel->{thirdpartypkgs}{$sys->{platvers}}));
        if ($minorvers) {
            $fallback_minorvers=$rel->{minorvers_for_fallback};

            if (defined $rel->{thirdpartypkgs}{$minorvers}) {
                push(@thirdpartypkgs, @{$rel->{thirdpartypkgs}{$minorvers}})
            } elsif (defined $fallback_minorvers &&
                     defined $rel->{thirdpartypkgs}{$fallback_minorvers}) {
                # The thirdpartypkgs are not defined for this minorvers platform,
                # try to check the fallback category
                push(@thirdpartypkgs, @{$rel->{thirdpartypkgs}{$fallback_minorvers}})
            }
        }
    }

    return 1 unless (@thirdpartypkgs);

    # Check third party package dependencies
    @missing_thirdpartypkgs=();
    for my $ospkg (@thirdpartypkgs) {
        next if (defined $checked_pkgs{$ospkg});
        $checked_pkgs{$ospkg}=1;

        ($pkgi,$defvers)=split(/\s+/,$ospkg);
        $pkgobj=Pkg->new_pkg($pkgi, $sys->{padv});
        $vers=$pkgobj->version_sys($sys);
        if (!$vers) {
            push(@missing_thirdpartypkgs,$ospkg);
        } elsif ($defvers) {
            if (EDRu::compvers($vers,$defvers)==2) {
                Msg::log("The $pkgi version $vers is lower than $defvers on $sys->{sys}. Need to upgrade the $pkgi version before proceeding");
                push(@missing_thirdpartypkgs,$ospkg);
            }
        }
    }

    # Print the missing thirdparty pkgs list
    $sys->set_value('missing_thirdpartypkgs_for_product', undef);
    if (@missing_thirdpartypkgs) {
        $sys->set_value('missing_thirdpartypkgs_for_product', 'list', @missing_thirdpartypkgs);
#        $pkgs=join(' ',@missing_thirdpartypkgs);
#        $msg=Msg::new("The following required third party $pdfrs were not found or have lower version on $sys->{sys}:\n\t$pkgs");
#        $sys->push_error($msg);
        return 0;
    }

    return 1;
}

sub prod_prerequisite_oslibs_sys {
    my ($cpic,$sys) = @_;
    my ($rel,$prod,$pkg,$padv,$pdfrs,$minorvers,$fallback_minorvers);
    my ($ospkg,$ospkgs,$msg,$rtn,$rtns,%all_oslibs,$oslibs,$uniq_oslibs,@oslibs,@missing_oslibs,@missing_oslibpkgs);

    $rel=$sys->rel;
    $prod=$sys->prod;
    $padv=$sys->padv;

    $pdfrs=Msg::get('pdfrs');
    $minorvers=$sys->{minorvers};

    %all_oslibs=();
    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);

        $oslibs=$pkg->{oslibs}{all};
        @all_oslibs{ keys %{$oslibs} } = values %{$oslibs} if (defined $oslibs);

        $oslibs=$pkg->{oslibs}{$sys->{platvers}};
        @all_oslibs{ keys %{$oslibs} } = values %{$oslibs} if (defined $oslibs);

        if ($minorvers) {
            $fallback_minorvers=$pkg->{minorvers_for_fallback};
            $fallback_minorvers||=$rel->{minorvers_for_fallback};

            if (defined $pkg->{oslibs}{$minorvers}) {
                $oslibs=$pkg->{oslibs}{$minorvers};
                @all_oslibs{ keys %{$oslibs} } = values %{$oslibs};
            } elsif (defined $fallback_minorvers &&
                     defined $pkg->{oslibs}{$fallback_minorvers}) {
                # The oslibs are not defined for this minorvers platform,
                # try to check the fallback category
                $oslibs=$pkg->{oslibs}{$fallback_minorvers};
                @all_oslibs{ keys %{$oslibs} } = values %{$oslibs};
            }
        }
    }

    # Add release level information
    if (defined $rel) {
        $oslibs=$rel->{oslibs}{all};
        @all_oslibs{ keys %{$oslibs} } = values %{$oslibs} if (defined $oslibs);

        $oslibs=$rel->{oslibs}{$sys->{platvers}};
        @all_oslibs{ keys %{$oslibs} } = values %{$oslibs} if (defined $oslibs);

        if ($minorvers) {
            $fallback_minorvers=$rel->{minorvers_for_fallback};

            if (defined $rel->{oslibs}{$minorvers}) {
                $oslibs=$rel->{oslibs}{$minorvers};
                @all_oslibs{ keys %{$oslibs} } = values %{$oslibs};
            } elsif (defined $fallback_minorvers &&
                     defined $rel->{oslibs}{$fallback_minorvers}) {
                # The oslibs are not defined for this minorvers platform,
                # try to check the fallback category
                $oslibs=$rel->{oslibs}{$fallback_minorvers};
                @all_oslibs{ keys %{$oslibs} } = values %{$oslibs};
            }
        }
    }

    @oslibs= keys %all_oslibs;
    return 1 unless (@oslibs);

    # To check library dependencies (mainly for Linux)
    @missing_oslibs=();

    $uniq_oslibs=EDRu::arruniq(@oslibs);
    if ($padv->can('oslibraries_sys')) {
        $rtns=$padv->oslibraries_sys($sys,@{$uniq_oslibs});
        for my $i (0..$#$rtns) {
            push(@missing_oslibs,$uniq_oslibs->[$i]) unless ($rtns->[$i]);
        }
    } elsif ($padv->can('oslibrary_sys')) {
        for my $oslib (@{$uniq_oslibs}) {
            $rtn=$padv->oslibrary_sys($sys,$oslib);
            push(@missing_oslibs,$oslib) unless ($rtn);
        }
    }

    if (@missing_oslibs) {
        @missing_oslibpkgs=();

        for my $oslib (@missing_oslibs) {
            $ospkg=$all_oslibs{$oslib};
            ($ospkg) = split(/\s+/,$ospkg);
            push (@missing_oslibpkgs,$ospkg) unless EDRu::inarr($ospkg,@missing_oslibpkgs);
        }

        if (@missing_oslibpkgs) {
            $sys->set_value('missing_ospkgs_for_product', 'push', @missing_oslibpkgs);
            $ospkgs=join(' ',@missing_oslibpkgs);
            $msg=Msg::new("The following required OS $pdfrs (or higher version) were not found on $sys->{sys}:\n\t$ospkgs", 30, 1431, "$pdfrs", "$sys->{sys}", "$ospkgs");
        } else {
            $oslibs=join(' ',@missing_oslibs);
            $msg=Msg::new("The following required OS libraries were not found on $sys->{sys}:\n\t$oslibs", 30, 1432, "$sys->{sys}", "$oslibs");
        }
        $sys->push_error($msg);
        return 0;
    }

    return 1;
}

sub prod_prerequisite_hfpatches_sys {
    my ($cpic,$sys) = @_;
    my ($rel,$prod,$pkg,$padv,$vers,$cv,$patch,$patches,$msg,@hfpatches);

    $prod=$sys->prod;
    $padv=$sys->padv;

    # Check the existence of HFs
    @hfpatches=();
    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);
        for my $patchi (@{$pkg->{privatehfpatches}{$sys->{platvers}}},@{$pkg->{privatehfpatches}{all}}) {
            ($patchi,$vers)=split(/-/m,$patchi,2);
            $patch=$sys->patch($patchi);
            $cv=EDRu::compvers($vers,$padv->patch_version_sys($sys, $patch),4);
            push(@hfpatches,"$patchi-$vers") unless ($cv);
        }
    }

    #print the installed HF patches list
    if (@hfpatches) {
        $rel=$sys->rel;
        $patches=join(' ',@hfpatches);
        $msg= Msg::new("The following patches are found on $sys->{sys}:\n\t$patches\nThese patches are not included in $rel->{titlevers}. Installing $prod->{abbr} $rel->{titlevers} will remove the functionality provided by these patches.", 30, 1433, "$sys->{sys}", "$patches", "$rel->{titlevers}", "$prod->{abbr}", "$rel->{titlevers}");
        $sys->push_warning($msg);
        return 0;
    }

    return 1;
}

sub post_prod_prerequisite {
    my $cpic=shift;
    return 1;
}

# To install missing OS packages
sub prod_prerequisite_error_handler {
    my $cpic=shift;
    my ($ret,$sys0,$native_method,$pdfrs,@missing_ospkgs_systems,$pkgs,$rtn,$msg,$webmsg);

    $sys0=$cpic->{systems}->[0];
    $native_method=$sys0->padv->{native_install_tool};

    # need not automatically install the missing OS packages if it is not supported.
    return 1 if (!$native_method);

    $pdfrs=Msg::get('pdfrs');
    @missing_ospkgs_systems =();
    $webmsg='';

    for my $sys(@{CPIC::get('systems')}) {
        if (defined $sys->{missing_thirdpartypkgs_for_product}) {
            # install third party packages
            Msg::n();
            $pkgs=join ' ', @{$sys->{missing_thirdpartypkgs_for_product}};
            $msg=Msg::new("The following required third party $pdfrs (or higher version) were not found on $sys->{sys}:\n\t$pkgs", 30, 1434, "$pdfrs", "$sys->{sys}", "$pkgs");
            $msg->print;
            $ret = $cpic->prod->install_thirdparty_pkgs_sys($sys,$sys->{missing_thirdpartypkgs_for_product});
            $rtn ||= $ret;
        }
        if (defined $sys->{missing_ospkgs_for_product}) {
            # get if any systems is missing OS required packages
            Msg::n();
            $pkgs=join ' ', @{$sys->{missing_ospkgs_for_product}};
            $msg=Msg::new("The following required OS $pdfrs (or higher version) were not found on $sys->{sys}:\n\t$pkgs", 30, 1431, "$pdfrs", "$sys->{sys}", "$pkgs");
            $msg->print;
            $webmsg .=$msg->{msg}."\\n";
            push @missing_ospkgs_systems, $sys;
        }
    }

    if (Obj::webui()) {
        my $web = Obj::web();
        $web->web_script_form('alert',$webmsg) if (@missing_ospkgs_systems);
    }

    # Try to install the missing OS packages with native install method
    $rtn=$cpic->install_missing_os_pkgs(\@missing_ospkgs_systems) if (@missing_ospkgs_systems);
    if ($rtn) {
        $cpic->{retry_verify_checks}=1;
        EDR::reset_errors_warnings();
    }

    return 1;
}

# return 1: means the missing OS packages are installed successfully
# return 0: means the missing OS packages are not installed successfully
sub install_missing_os_pkgs {
    my ($cpic,$systems) =@_;
    my ($sys0,$msg,$help,$menus,$backopt,$defaultopt,$option,$pdfrs,$native_method);

    $pdfrs=Msg::get('pdfrs');

    # Show messages that some required OS packages were not installed
    # CPI will provide 2 options to help user install those packages
    # or show guidance how to install thme manually.
    Msg::n();

    $sys0=$systems->[0];
    $native_method=$sys0->padv->{native_install_tool};
    if ($native_method) {
        if (Obj::webui()){
            my $web = Obj::web();
            $option = $web->web_script_form('os_pkg_missing',$pdfrs, $native_method);
        } else {
            $msg=Msg::new("The installer provides some guidance about how to install OS $pdfrs using native methods, like $native_method, or how to manually install the required OS $pdfrs.", 30, 1435, "$pdfrs", "$native_method", "$pdfrs");
            $msg->printn();

            # Show options about how to install those missing required OS packages.
            $msg = Msg::new("Install the missing required OS $pdfrs with $native_method, if $native_method is configured on the systems", 30, 1436, "$pdfrs", "$native_method", "$native_method");
            push(@{$menus},$msg->{msg});

            $msg = Msg::new("Install the missing required OS $pdfrs manually, (detailed steps are provided)", 30, 1437, "$pdfrs");
            push(@{$menus},$msg->{msg});

            $msg = Msg::new("Do not install the missing required OS $pdfrs", 30, 1438, "$pdfrs");
            push(@{$menus},$msg->{msg});

            $msg = Msg::new("How would you like to install the missing required OS $pdfrs?", 30, 1439, "$pdfrs");

            $backopt = 0;
            $defaultopt = 1;
            $help=Msg::new("Select the option for installing the missing required OS $pdfrs", 30, 1440, "$pdfrs");

            $option = $msg->menu($menus,$defaultopt,$help,$backopt);
            Msg::n();
        }

        if ($option eq '1') {
            return $cpic->install_missing_os_pkgs_with_native_tool($systems);
        } elsif ($option eq '2') {
            return $cpic->install_missing_os_pkgs_with_manual_guide($systems);
        } elsif ($option eq '3') {
            return 0;
        }
    } else {
        $msg=Msg::new("Some required OS $pdfrs were not found, the installer provides some guidance about how to install them manually.", 30, 1441, "$pdfrs");
        $msg->printn();

        return $cpic->install_missing_os_pkgs_with_manual_guide($systems);
    }

    return 1;
}

sub install_missing_os_pkgs_with_native_tool {
    my ($cpic,$systems)=@_;
    my ($msg,$pdfrs,$native_method,$feature,$missing_os_pkgs,$rtn,$webmsg);

    $pdfrs=Msg::get('pdfrs');

    $msg = Msg::new("The installation may take a few minutes, be patient.", 30, 1442);
    $msg->printn;

    $rtn=1;
    for my $sys (@{$cpic->{systems}}) {
        $missing_os_pkgs=$sys->{missing_ospkgs_for_product};
        next unless (defined $missing_os_pkgs && @{$missing_os_pkgs});

        $native_method=$sys->padv->{native_install_tool};
        $msg=Msg::new("Install the missing OS $pdfrs with $native_method on $sys->{hostname}", 30, 1443, "$pdfrs", "$native_method", "$sys->{hostname}");
        $msg->left();
        if ($cpic->install_missing_os_pkgs_with_native_tool_sys($sys,$missing_os_pkgs)) {
            Msg::right_done();
        } else {
            Msg::right_failed();
            $rtn=0;
        }
    }

    $webmsg='';
    for my $sys (@{$cpic->{systems}}) {
        $msg=$sys->{native_install_fail_message};
        next unless (defined $msg);

        Msg::n();
        $msg->error;
        $webmsg .=$msg->{msg}."\\n";
        undef $sys->{native_install_fail_message};
    }

    if ($webmsg && Obj::webui()){
        my $web = Obj::web();
        $web->web_script_form('alert',$webmsg);
    }

    # Pause if something wrong happen
    Msg::prtc();

    return $rtn;
}

sub install_missing_os_pkgs_with_native_tool_sys {
    my ($cpic,$sys,$missing_os_pkgs)=@_;
    my ($native_method,$native_install_handler,$native_install_cmd,$pkgs,$msg);

    return 1 unless (defined $missing_os_pkgs && @{$missing_os_pkgs});

    $native_install_handler=$sys->padv->{native_install_handler};
    $native_install_cmd=$sys->padv->{native_install_cmd};

    if (defined $native_install_handler) {
        return &$native_install_handler($sys->padv,$sys,$missing_os_pkgs);
    } elsif (defined $native_install_cmd) {
        $sys->cmd("$native_install_cmd $pkgs 2>&1");
        if (EDR::cmdexit()) {
            $native_method=$sys->padv->{native_install_tool};
            $pkgs=join(' ', @{$missing_os_pkgs});
            $msg=Msg::new("$native_method failed to install $pkgs on $sys->{sys}", 30, 1444, "$native_method", "$pkgs", "$sys->{sys}");
            $sys->{native_install_fail_message}=$msg;
            return 0;
        }
    }
    return 1;
}

sub install_missing_os_pkgs_with_manual_guide {
    my ($cpic,$systems)=@_;
    my ($sys0,$padv,$msg,$pdfrs,$native_method,$native_cmd,$package_cmd,$feature,$missing_os_pkgs,$pkgs,$webmsg);

    $pdfrs=Msg::get('pdfrs');

    $sys0=$systems->[0];
    $padv=$sys0->padv;
    $native_method=$padv->{native_install_tool};
    $native_cmd=$padv->{native_install_cmd};
    if ($native_method && $native_cmd) {
        $msg=Msg::new("If $native_method was configured on the systems, run the following command to install the OS $pdfrs manually:", 30, 1445, "$native_method", "$pdfrs");
        $msg->printn;
        $webmsg .= $msg->{msg}.'\\n';
        for my $sys (@{$cpic->{systems}}) {
            $missing_os_pkgs=$sys->{missing_ospkgs_for_product};
            next unless (defined $missing_os_pkgs && @{$missing_os_pkgs});

            $pkgs=join(' ', @{$missing_os_pkgs});
            $msg=Msg::new("    On $sys->{sys}, run: $native_cmd $pkgs", 30, 1446, "$sys->{sys}", "$native_cmd", "$pkgs");
            $msg->print;
            $webmsg .= $msg->{msg}.'\\n';
        }
    }

    $package_cmd=$padv->{package_install_cmd};
    if ($package_cmd) {
        Msg::n();
        $msg=Msg::new("Or change directory to the package repository, and run the following package installation command to install the OS $pdfrs manually:", 30, 1447, "$pdfrs");
        $msg->printn;
        $webmsg .= $msg->{msg}.'\\n';
        for my $sys (@{$cpic->{systems}}) {
            $missing_os_pkgs=$sys->{missing_ospkgs_for_product};
            next unless (defined $missing_os_pkgs && @{$missing_os_pkgs});

            $msg=Msg::new("    On $sys->{sys}, run:", 30, 1448, "$sys->{sys}");
            $msg->print;
            $webmsg .= $msg->{msg}.'\\n';
            for my $pkg (@{$missing_os_pkgs}) {
                my ($pkg_show,$ver)=split(/\s+/,$pkg);
                $msg="        $package_cmd ${pkg_show}\*";
                Msg::print($msg); 
                $webmsg .= $msg.'\\n';
            }
        }
    }
    if (Obj::webui()){
        my $web = Obj::web();
        $web->web_script_form('alert',$webmsg);
    }
    Msg::prtc();
    return 1;
}

sub installed_prod_sys {
    my ($cpic,$sys) = @_;
    my ($msg,$cfg,$prodi,$ivers,$prod,$installtime,$prior_rp,$rel);
    $cfg=Obj::cfg();
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $rel=$sys->rel;
    return '' if ($sys->{stop_checks});
    $sys->pkgs_patches();
    $rel->feature_usage_sys($sys);
    $rel->check_installed_pkgs_version_sys($sys);
    for my $prodi (@{$sys->rel->{upgradeprods}}) {
        $prod=$sys->prod($prodi);
        # set $force_flag = 1 when query product version
        $ivers=$prod->version_sys($sys,1);
        if ($ivers) {
            $installtime = $prod->installtime_sys($sys);
            $sys->set_value('upgradeprod', 'push', $prodi);
            $sys->set_value('upgradeprod_abbr', 'push', $prod->{abbr});
            $sys->set_value('prodvers', 'push', $ivers);
            $sys->set_value('prodinstalltimes', 'push', $installtime);
            # replace $cpic->prod if use installsfcfs to configure SFCFSHA product
            if (!$cpic->{prod} || ($prod->configure_alternate_prod() && (Cfg::opt(qw/start stop configure precheck postcheck/))) ) {
                $cpic->set_value('prod',$prodi) unless ((Cfg::opt(qw(patchupgrade hotfixupgrade rollback))) && (EDRu::compvers($ivers,$prod->{vers},2)));
            }
        }
    }
    $cpic->set_value('prod',${$sys->{upgradeprod}}[0]) if (!$cpic->{prod} && $sys->{upgradeprod});
    if ($rel->{type}=~/^M/) {
        $prior_rp = $rel->detect_prior_rp_sys($sys) if ($cpic->{prod});
        $sys->set_value('prior_rp',$prior_rp);
    }
    return 1;
}

sub prod_version_sys {
    my ($cpic,$sys)=@_;
    my ($prod,$prodname,$padv,$ivers,$msg,$opt,$cv,$installedprod,$rel,$noconfig_flag,$prior_rp_title,$basevers_flag,$basevers);

    # noconfig_flag is related with SF or VCS partial upgrade out of installed SFHA
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $rel=$sys->rel;
    return 1 if ($sys->{stop_checks});
    return 1 if ($rel->{lp});
    if ($cpic->{prod}) {
        $prod=$sys->prod;
        $prodname=$prod->{abbr};
        $ivers=$prod->version_sys($sys);
    } else {
        $prodname = 'Any product';
    }
    $noconfig_flag = ($sys->{noconfig_sf} || $sys->{noconfig_vcs});
    $cv=EDRu::compvers($ivers,$prod->{vers},4) if ($ivers);
    # task        check whether product installed            check whether current version
    # install     X                                          X
    # upgrade     V                                          X
    # patchupgrade     V                                     X
    # hotfixupgrade    V                                     X
    # uninstall   V                                          V
    # configure   V                                          V
    # start       V                                          V
    # stop        V                                          V
    # license     V                                          V
    # precheck    X                                          X
    if (Cfg::opt(qw(uninstall rollback configure start stop startstoptest postcheck license))) {
        if (!$ivers) {
            $msg=Msg::new("$prodname $prod->{vers} does not appear to be installed or licensed on $sys->{sys}", 30, 1449, "$prodname", "$prod->{vers}", "$sys->{sys}");
            if (Cfg::opt(qw(uninstall license))) {
                $sys->push_warning($msg);
            } else {
                unless (Cfg::opt('install') && Cfg::opt('configure')) {
                    $sys->push_stop_checks_error($msg);
                    return '';
                }
            }
        } elsif ((${$sys->{upgradeprod}}[0] ne "") && (Cfg::opt("rollback")) && $cv) {
          # No warning message for MR as customers many times have only partial MR installed
          $msg=Msg::new("$rel->{titlevers} does not appear to be completely installed on $sys->{sys}", 30, 1450, "$rel->{titlevers}", "$sys->{sys}");
          $sys->push_warning($msg);
        } elsif ((Cfg::opt("rollback")) && $sys->{prior_rp}) {
            $prior_rp_title = $rel->{prior_rp_title}{$sys->{prior_rp}};
            $msg=Msg::new("$sys->{sys} will be rolled back to $prior_rp_title release", 30, 1451, "$sys->{sys}", "$prior_rp_title");
            $sys->push_warning($msg);
        }
    # install/upgrade/patchupgrade/hotfixupgrade/precheck below :
    } elsif (!$ivers) {
        # not installed
        if (Cfg::opt(qw(patchupgrade hotfixupgrade)) && (${$sys->{upgradeprod}}[0] eq '')) {
            return 1 if (Cfg::opt(qw(upgrade patchupgrade hotfixupgrade)) && $sys->{skip_versioncheck} && $cpic->{prod});
            $msg=Msg::new("$prodname does not appear to be installed or licensed on $sys->{sys}", 30, 1452, "$prodname", "$sys->{sys}");
            $sys->push_stop_checks_error($msg);
            return '';
        } elsif (${$sys->{upgradeprod}}[0]) {
            # cross stack upgrade/install
            $installedprod=$sys->prod(${$sys->{upgradeprod}}[0]);
            $ivers=$installedprod->version_sys($sys);
            $cv=EDRu::compvers($ivers,$installedprod->{vers},4);
            if ($cv==2) {
                if (Cfg::opt('upgrade')) {
                    if (!$installedprod->upgradeable_version_sys($sys)) {
                        $msg=Msg::new("Upgrade from $installedprod->{abbr} $ivers to $prodname $prod->{vers} is not supported on $sys->{sys}", 30, 1453, "$installedprod->{abbr}", "$ivers", "$prodname", "$prod->{vers}", "$sys->{sys}");
                        $sys->push_error($msg);
                        return '';
                    }
                    if ($installedprod->zru_version_sys($sys)) {
                        $sys->set_value('zru_supported',1);
                    } else {
                        $sys->set_value('zru_supported',0);
                    }
                }
            } elsif ($cv==1) {
                # more recent version is installed
                $msg=Msg::new("A more recent version of $installedprod->{abbr}, $ivers, is already installed on $sys->{sys}", 30, 1454, "$installedprod->{abbr}", "$ivers", "$sys->{sys}");
                $sys->push_error($msg);
            }
        } elsif (Cfg::opt(qw(install))) {
            return 1;
        }
    } elsif ($cv==0) {
        $msg=Msg::new("$prodname version $ivers is already installed on $sys->{sys}", 30, 1455, "$prodname", "$ivers", "$sys->{sys}");
        # When product already installed and user is performing RU phase 2.
        # Now RU's precheck won't call $cpic->prod_version_sys
        # And the option 'upgrade_nonkernelpkgs' will be determined after user confirmed cluster info.
        if (Cfg::opt('upgrade_nonkernelpkgs')) {
            $sys->set_value('zru_supported',1);
            return 1;
        }
        if (Cfg::opt('upgradecv') && Cfg::opt('upgrade')) {
            $sys->push_warning($msg);
            # assume zero reboot supported for upgrade current version
            $sys->set_value('zru_supported',1);
            return 1;
        }
        if (!$noconfig_flag) {
            if (Cfg::opt(qw(patchupgrade hotfixupgrade))) {
            } elsif (Cfg::opt('upgrade')) {
                $sys->push_error($msg);
                return '';
            } else {
                $sys->push_warning($msg);
            }
        } else {
            if ($sys->{prod_to_upgrade} && $sys->{prod_to_upgrade}->zru_version_sys($sys)) {
                $sys->set_value('zru_supported',1);
            } else {
                $sys->set_value('zru_supported',0);
            }
        }
    } elsif (Cfg::opt(qw(patchupgrade hotfixupgrade)) && (!$rel->{pkgs})) {
        # patchonly release version check
        # Match if basevers
        for my $basevers (@{$rel->{basevers}}) {
            for my $subprodi(@{$sys->{upgradeprod}}){
                my $subprod=$sys->prod($subprodi);
                my $subprod_ivers=$subprod->version_sys($sys,1);
                if (!EDRu::compvers($subprod_ivers,$basevers,3)){
                    $basevers_flag=1;
                    last;
                }
            }
        }
        if (!$basevers_flag) {
            $basevers=join ', ',@{$rel->{basevers}};
            $msg=Msg::new("$prodname $basevers does not appear to be installed or licensed on $sys->{sys}", 30, 1449, "$prodname", "$basevers", "$sys->{sys}");
            $sys->push_stop_checks_error($msg);
        }
    } elsif ($cv==1) {
        # more recent version is installed
        $msg=Msg::new("A more recent version of $prodname, $ivers, is already installed on $sys->{sys}", 30, 1454, "$prodname", "$ivers", "$sys->{sys}");
        $sys->push_error($msg);
    } elsif ((Cfg::opt(qw(responsefile))) &&
        (!$prod->{responsefileupgradeok})) {
        # prior version installed, can't upgrade w responsefile
        # questions being asked is almost inevitable
        # installonly may need to run start/stop/config routines to upgrade
        $opt = (Cfg::opt('responsefile')) ? '-responsefile' : '-install';
        $msg=Msg::new("$prodname version $ivers is installed on $sys->{sys}.  Upgrading $prodname is not supported using the $opt option", 30, 1456, "$prodname", "$ivers", "$sys->{sys}", "$prodname", "$opt");
        $sys->push_error($msg);
        return '';
        # Code to check if upgrade from current version is supported or not is yet to be added (UGVERS)
    } else {
        return '' if (!$sys->rel->upgradevers_sys($sys));
        if ($prod->zru_version_sys($sys)) {
            $sys->set_value('zru_supported',1);
        } else {
            $sys->set_value('zru_supported',0);
        }
    }
    return 1;
}

sub installed_prod_info_sys {
    my ($cpic,$sys) = @_;
    return $cpic->installed_prod_sys($sys);
}

sub post_installed_prod_info {
    my $cpic=shift;
    if (install_check_task() || Cfg::opt('uninstall')) {
        $cpic->rel->upgrade_prod_mix_check();
    }
    for my $sys (@{$cpic->{systems}}) {
        $cpic->prod_version_sys($sys);
    }
    return 1;
}

sub fsspace_sys {
    my ($cpic,$sys) = @_;
    my ($rootpath,$dfk,$msg,$max,@localvol,$reqspace,$plat,$n,$patch,@dfk,$free,$pdfr,@rsp,@vol,@space,$pkgi,$itv,$pkg,$patchi,@dc,$tmpdir,$prod,$tmpspace);
    my ($freeM,$maxM,$reqspaceM,$tmpspaceM);
    my $rel=Obj::rel();
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if ($sys->{stop_checks});
    if (Cfg::opt('rootpath')) {
        $rootpath=Cfg::opt('rootpath');
        $rootpath=~s/\/+/\//mg;
        $rootpath=~s/\/$//m;
    }
    @vol=("$rootpath/opt", "$rootpath/usr", "$rootpath/var", "$rootpath/");
    $cpic->all_volumespace_sys($sys,@vol);
    $plat=EDRu::plat($sys->{padv});
    $prod=$sys->prod;
    @localvol=@vol;

    # log free fs space
    Msg::log('Filesystem available:');
    Msg::log("System               $rootpath/opt       $rootpath/usr       $rootpath/var         $rootpath/");
    $msg=sprintf ("%-16s %10s %10s %10s %10s\n", $sys->{sys},
                   $sys->{space}{"$rootpath/opt"}, $sys->{space}{"$rootpath/usr"},
                   $sys->{space}{"$rootpath/var"}, $sys->{space}{"$rootpath/"});
    Msg::log($msg);

    @space=(0,0,0,0);
    $max=0;
    if ($rel->rollback_check_task()) {
        return $rel->rollback_fsspace_sys($sys);
    }

    # accumulate the required space
    Msg::log('Required filesystem space:');
    $pdfr=Msg::get('pdfr');
    $msg=sprintf('%-16s %9s %9s %9s %9s', $pdfr, $vol[0],
        $vol[1], $vol[2], $vol[3]);
    Msg::log($msg);
    # make a list of package patches, upsell pkgs are not included yet.
    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$prod->pkg($pkgi);
        next if (Cfg::opt('upgrade_nonkernelpkgs') && (!$pkg->{unkernelpkg}));
        # mock value for pkg space
        $pkg->{space}||=[100,100,100,100];
        unless ((scalar(@{$pkg->{space}})==4) || (Cfg::opt(qw(patchupgrade hotfixupgrade)))) {
            $msg=Msg::new("SPACE for $pkg->{pkg} is not defined on $plat", 30, 1457, "$pkg->{pkg}", "$plat");
            $sys->push_warning($msg);
            next;
        }
        unless (Cfg::opt(qw(patchupgrade hotfixupgrade))) {
            # do not count pkg space if upsell and the pkg is already installed
            next if (($sys->pkgvers($pkg)) && (!Cfg::opt('upgrade')));
            # mock value for pkg size
            $pkg->{size}||=10;
            $max=$pkg->{size} if ($max<$pkg->{size});
            # do not count pkg space if upgrade and the pkg is already installed
            unless (($sys->pkgvers($pkg)) && (Cfg::opt('upgrade'))) {
                @space=@{$pkg->{space}};
                $msg=sprintf('%-16s %9d %9d %9d %9d', $pkg->{pkg}, $space[0],
                    $space[1], $space[2], $space[3]);
                Msg::log($msg);
                for my $n (0..3) {
                    $rsp[$n]+=$space[$n];
                }
            }
        }
        for my $patchi (@{$pkg->patches_sys($sys)}) {
            $patch=$sys->patch($patchi);
            $patch->{space}||=$pkg->{space};
            # do not count patch space is it is already installed
            next if (($sys->linux() && $sys->{pkgvers}{"$patch->{patchname}"}) || $patch->installed_sys($sys));
            unless (scalar(@{$patch->{space}})==4) {
                $msg=Msg::new("SPACE for $patch->{patchname} is not defined on $plat", 30, 1457, "$patch->{patchname}", "$plat");
                $sys->push_warning($msg);
                next;
            }
            # mock patch size
            $patch->{size}||=5;
            $max=$patch->{size} if ($max<$patch->{size});
            @space=@{$patch->{space}};
            $msg=sprintf('%-16s %9d %9d %9d %9d', $patchi, $space[0],
                $space[1], $space[2], $space[3]);
            Msg::log($msg);
            for my $n (0..3) {
                $rsp[$n]+=$space[$n];
            }
        }
    }
    $msg=sprintf('%-16s %9d %9d %9d %9d', 'KB total', $rsp[0],
        $rsp[1], $rsp[2], $rsp[3]);
    Msg::log($msg);

    # now determine file system space requirement and add
    # largest package/packages size*1.1 for copying
    $max = ($cpic->nsystems>$cpic->{threadlimit}/2) ? int($max*1.1) : int($max*3);
    $tmpdir=EDR::tmpdir();
    $dfk=$sys->cmd("_cmd_dfk $tmpdir");
    @dfk=split(/\n/,$dfk);
    $dfk=pop(@dfk);
    @dfk=split(/\s+/m,$dfk);
    # check if tmppath is set on local, mounted or rootpath file system
    $itv=EDRu::arrpos("$dfk[5]", @localvol);
    if ($itv<0) {
        $itv=EDRu::arrpos("$dfk[5]", @vol) if ($rootpath ne '');
        if ($itv<0) {
            # tmppath is set on mounted file system
            push (@vol, "$dfk[5]");
            push (@rsp, 0);
            $rsp[-1]+=$max;
        } else {
            # tmppath is set on rootpath file system
            $rsp[$itv]+=$max;
        }
        Msg::log("$dfk[5] requires $max KB as the copied $pdfr and patch destination");
    } else {
        Msg::log("$localvol[$itv] requires $max KB as the copied $pdfr and patch destination");
        if (Cfg::opt('rootpath') && ($rootpath ne '')) {
            # tmppath is set on local file system and pkgs/patches are installed to rootpath
            # check local free space for gzipping/untarring
            $free=$cpic->volumespace_sys($sys, $localvol[$itv]);
            if ((!defined $free) && ($itv != 3)) {
                Msg::log("$sys->{sys} does not have a unique $localvol[$itv] volume");
                $itv=3;
                $free=$cpic->volumespace_sys($sys, $localvol[$itv]);
            }
            if ((!defined $free) && ($itv == 3)) {
                $msg=Msg::new("Cannot determine free space available in root file system on $sys->{sys}", 30, 1458, "$sys->{sys}");
                $sys->push_error($msg);
                return ''
            }
            if ($free<$max) {
                Msg::log("$max KB is required in the $localvol[$itv] volume and only $free KB is available on $sys->{sys}");
                $freeM = sprintf (".0f" , $free / 1024);
                $maxM = sprintf (".0f" , $max / 1024);
                $msg=Msg::new("$maxM MB is required in the $localvol[$itv] volume and only $freeM MB is available on $sys->{sys}", 30, 1459, "$maxM", "$localvol[$itv]", "$freeM", "$sys->{sys}");
                $sys->push_error($msg);
            }
        } else {
            $rsp[$itv]+=$max;
        }
    }

    # now check against free space in each volume
    for my $n (0..$#vol) {
        $reqspace=$rsp[$n];
        next if (!$reqspace);
        $free=$cpic->volumespace_sys($sys, $vol[$n]);
        $cpic->{spacewarning}||=0;
        if (!defined $free) {
            if ($n==3) {
                $msg=Msg::new("Cannot determine free space available in root ($rootpath) file system on $sys->{sys}", 30, 1460, "$rootpath", "$sys->{sys}");
                $sys->push_error($msg);
                return '';
            }
            $rsp[3]+=$reqspace;
            Msg::log("$sys->{sys} does not have a unique $vol[$n] volume");
            Msg::log("$vol[$n] space of $reqspace KB will be added to $vol[3]=$rsp[3]");
        } elsif ($free < $reqspace) {
            $tmpspace = $reqspace - $free;
            Msg::log("$reqspace KB is required in the $vol[$n] volume and only $free KB is available on $sys->{sys}. $tmpspace KB needs to be freed up in the $vol[$n] volume.");
            $reqspaceM = sprintf("%.0f", $reqspace / 1024);
            $freeM = sprintf ("%.0f", $free / 1024);
            $tmpspaceM = sprintf ("%.0f", $tmpspace / 1024);
            $msg=Msg::new("$reqspaceM MB is required in the $vol[$n] volume and only $freeM MB is available on $sys->{sys}. $tmpspaceM MB needs to be freed up in the $vol[$n] volume.", 30, 1461, "$reqspaceM", "$vol[$n]", "$freeM", "$sys->{sys}", "$tmpspaceM", "$vol[$n]");
            $sys->push_error($msg);
        } elsif (($free-$reqspace)<$cpic->{spacewarning}) {
            push(@dc,$vol[$n],$free-$reqspace);
        } else {
            Msg::log("$sys->{sys} has $vol[$n]");
        }
    }
    return 1 if (Cfg::opt(qw(responsefile precheck)));
    # limited space warnings
    while (@dc) {
        $n=shift(@dc);
        $free=shift(@dc);
        $free=sprintf('%d', $free);
        $msg=Msg::new("Installing $prod->{name} will leave only approximately $free KB of free space remaining in the $n volume on $sys->{sys}", 30, 1462, "$prod->{name}", "$free", "$n", "$sys->{sys}");
        $sys->push_warning($msg);
    }
    return 1;
}

# determin the space needed for copying pkgs/patches with copy thread.
sub max_pkgssize_sys {
    my ($cpic,$sys)= @_;
    my (@pkgs,$pkg,$prod,$tmpsize,$maxsize);
    $prod=$sys->prod;
    for my $pkgi (@{$prod->get_pkgs}) {
        $pkg=$sys->pkg($pkgi);
        push (@pkgs, $pkg);
        for my $patchi (@{$pkg->patches_sys($sys)}) {
            push (@pkgs,$sys->patch($patchi));
        }
    }
    for my $i (0..($#pkgs-2)) {
        $tmpsize=int($pkgs[$i]->{size}+$pkgs[$i+1]->{size}+$pkgs[$i+2]->{size});
        $maxsize=$tmpsize if ($maxsize<$tmpsize);
    }
    return int($maxsize*1.1);
}

sub all_volumespace_sys {
    my ($cpic,$sys,@vol) = @_;
    for my $vol (@vol) {
        $sys->{space}{$vol}=$cpic->volumespace_sys($sys,$vol) || 0;
    }
    return 1;
}

# if allowdir set, even if $vol is not a mount point,
# volumespace_sys will return the space of the volume that $vol belong to.
sub volumespace_sys {
    my ($cpic,$sys,$vol,$allowdir) = @_;
    my ($dfk,@dfk,$mount);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    Msg::debug("Checking $vol volume on $sys->{sys}");
    @dfk=split(/\n/, $sys->cmd("_cmd_dfk $vol 2> /dev/null"));
    $dfk=pop(@dfk);
    return unless $dfk;
    @dfk=split(/\s+/m,$dfk);
    Msg::debug("no $vol volume on $sys->{sys}") unless ($dfk[3]);
    $mount=$dfk[5];
    if($allowdir || $dfk[5] =~ /^$mount(\/|)$/) {
        # if allow $vol is not a mount point
        # or $vol exactly matched
        return $dfk[3];
    }
    return;
}

sub pre_prod_precheck {
    my ($cpic)= @_;
    return 0 unless (EDR::no_stop_checks());

    return $cpic->prod->precheck_task();
}

sub prod_precheck_sys {
    my ($cpic,$sys)= @_;
    my ($rel,$prod,$superprod,$msg,$mr);

    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if ($sys->{stop_checks});

    # release common check for product
    # Check if other installer instances running on the system and the system is locked or not
    $rel=$cpic->rel;
    $rel->check_multilock_sys($sys);

    # product common prerequisite check
    $prod=$sys->prod;
    $prod->basic_precheck_sys($sys);
    $prod->rel->basic_precheck_sys($sys);

    (Cfg::opt('uninstall')) ? $prod->uninstall_precheck_sys($sys) :
        (Cfg::opt('patchupgrade')) ? $prod->patchupgrade_precheck_sys($sys) :
        (Cfg::opt('hotfixupgrade')) ? $prod->hotfixupgrade_precheck_sys($sys) :
        (Cfg::opt('upgrade')) ? $prod->upgrade_precheck_sys($sys) :
        (Cfg::opt('rollback')) ? $prod->rollback_precheck_sys($sys) :
        (Cfg::opt('configure')) ? $prod->configure_precheck_sys($sys) :
        (Cfg::opt('start')) ? $prod->start_precheck_sys($sys) :
        (Cfg::opt('stop')) ? $prod->stop_precheck_sys($sys) :
        (Cfg::opt('postcheck')) ? $prod->postcheck_precheck_sys($sys) :
                             $prod->install_precheck_sys($sys);

    if (Cfg::opt(qw(uninstall configure stop)) && $prod->{superprods}) {
        for my $superprodi (@{$prod->{superprods}}) {
            $superprod = $prod->prod($superprodi,$sys->{padv},1);
            if ( $superprod && $superprod->version_sys($sys) ) {
                if ( Cfg::opt('uninstall')) {
                    $msg=Msg::new("$superprod->{name} is installed on $sys->{sys}. Processes or modules of $prod->{name} may fail to stop and packages of $prod->{name} may fail to uninstall", 30, 1463, "$superprod->{name}", "$sys->{sys}", "$prod->{name}", "$prod->{name}");
                } else {
                    $msg=Msg::new("$superprod->{name} is installed on $sys->{sys}. Processes or modules of $prod->{name} may fail to stop", 30, 1464, "$superprod->{name}", "$sys->{sys}", "$prod->{name}");
                }
                $sys->push_warning($msg);
                last;
            }
        }
    }
    # may go elsewhere, but for now
    if (Cfg::opt(qw(install upgrade patchupgrade hotfixupgrade))) {
        if ($mr = $cpic->{mrobj}) {
            $mr->precheck_sys($sys);
        }
        for my $hotfix(@{$cpic->{hotfixobj}}) {
            $hotfix->precheck_sys($sys);
        }
    }

    return 1;
}

sub fake_installed_packages {
    my $cpic=shift;
    my ($pkg,$pkgs,$prod,$rel);
    $prod=$cpic->prod;
    $rel=$cpic->rel;
    $pkgs=$prod->allpkgs;
    for my $pkgi (@$pkgs) {
        $pkg=$cpic->pkg($pkgi);
        for my $sys (@{$cpic->{systems}}) {
            $sys->{pkgvers}{$pkg->{pkg}}=$rel->{vers};
        }
    }
    return;
}

sub post_prod_precheck {
    my($cpic)= @_;
    $cpic->prod->post_precheck_task();
    return 1;
}

# To handle fatal ERRORs which found when run product prechecks
sub prod_precheck_error_handler {
    my $cpic=shift;
    $cpic->prod->handle_precheck_fatal_errors();
    return 1;
}

# the common padv can be defined per release or per-product
sub discover_common_padv {
    my $cpic=shift;
    (defined($cpic->rel->{pkgpaths})) ? $cpic->discover_rel_padv :
                                        $cpic->discover_prod_padv;
    return;
}

# for release-based settings, existence of the correct directory is enough
# this is primarily used by EEB's
sub discover_rel_padv {
    my $cpic=shift;
    my ($rel,$edr,$found,$padvisa);
    $rel=$cpic->rel;
    $edr=Obj::edr();
    for my $padvs(sort keys(%{$rel->{pkgpaths}})) {
        if (!-d "$cpic->{mediapath}/$rel->{pkgpaths}{$padvs}") {
            Msg::log("No $rel->{eebn} $padvs support: no $cpic->{mediapath}/$rel->{pkgpaths}{$padvs}");
            next;
        }
        Msg::log("$rel->{eebn} $padvs support: $cpic->{mediapath}/$rel->{pkgpaths}{$padvs} exists");
        ($padvisa,undef)=split(/\s+/, $padvs, 2);
        for my $padv(split(/\s+/, $padvs)) {
            $padv=~s/\+//;
            $padvisa=~s/\+//;
            Padv::set_padv($edr, $padv, $padvisa);
            Padv::set_padv($cpic, $padv, $padvisa);
            $cpic->{pkgpath}{$padv}=$rel->{pkgpaths}{$padvs};
            $found++;
        }
    }
    EDR::die("Did not find any padv directories") if (!$found);
    return;
}

# any to many padv for NetBackup differs between releases
# setting all potential padvs for all products and letting
# mainpkg discovery dictate exactly what to support
# should find Client for all padvs and Server/OpsCenter for
# one on release images, but this allows the potential for
# larger images with more options
sub discover_prod_padv {
    my $cpic=shift;
    my (@files,$edr,$found,$padvisa,$prod,$rel);
    $edr=Obj::edr();
    $rel=$cpic->rel;
    $prod=$cpic->prod;
    EDR::die("\%\$prod->{pkgpaths} undefined") unless (defined($prod->{pkgpaths}));
    EDR::die("\@\$prod->{mainpkgs} undefined") unless (defined($prod->{mainpkgs}));
    for my $padvs (sort keys(%{$prod->{pkgpaths}})) {
        if (!-d "$cpic->{mediapath}/$prod->{pkgpaths}{$padvs}") {
            Msg::log("No $prod->{prod} $padvs support: no $cpic->{mediapath}/$prod->{pkgpaths}{$padvs}");
            next;
        }
        for my $pkg (@{$prod->{mainpkgs}}) {
            @files = glob("$cpic->{mediapath}/$prod->{pkgpaths}{$padvs}/$pkg*");
            if (@files) {
                Msg::log("$prod->{prod} $padvs support: found $files[0]");
                ($padvisa,undef)=split(/\s+/m, $padvs, 2);
                for my $padv (split(/\s+/m, $padvs)) {
                    $padv=~s/\+//m;
                    $padvisa=~s/\+//m;
                    Padv::set_padv($edr, $padv, $padvisa);
                    Padv::set_padv($cpic, $padv, $padvisa);
                    $cpic->{pkgpath}{$padv}=$prod->{pkgpaths}{$padvs};
                    $found++;
                }
                last;
            }
        }
        Msg::log("No $prod->{prod} $padvs support: no mainpkg in $cpic->{mediapath}/$prod->{pkgpath}{$padvs}") unless (@files);
        # What about patchpath?
    }
    EDR::die("Did not find any padv directories") if (!$found);
    return;
}

sub prod_licenses_sys {
    my ($cpic,$sys) = @_;

    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $sys->rel->read_licenses_sys($sys);

    return 1;
}

sub post_prod_licenses {
    my ($cpic) = @_;
    return 1;
}

# To handle WARNINGs which found when run product prechecks, like time sync issue
sub handle_precheck_issues {
    my ($cpic) = @_;
    if ($cpic->{prod}) {
        $cpic->rel->add_multilock();
        $cpic->prod->handle_precheck_issues();
    }
    return 1;
}

sub update_summary_verify {
    my $cpic=shift;

    my $edr=Obj::edr();
    my $msg=Msg::new("The following checks and actions were performed during system verification on systems:", 30, 1465);
    $edr->update_check_action_summary($msg);

    return 1;
}

1;
